/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.ui.view;

import java.util.Vector;
import java.util.Date;

import com.funambol.util.Log;
import javax.microedition.lcdui.Canvas;


/**
 * This class implements a Canvas with the ability to listen to touch events and
 * recognize some basic gestures (such as multiple taps and basic scrolls).
 * The class is abstract as the handleTouchEvent must be implemented to handle
 * events.
 */
abstract class TouchCanvas extends Canvas {

    private int  lastPressedX = 0;
    private int  lastPressedY = 0;
    private int  lastDraggedX = -1;
    private int  lastDraggedY = -1;

    private boolean isKeyboardOnScreen = false;

    private TappingThread tt = null;

    protected abstract void handleFTouchEvent(FTouchEvent event);

    private class TappingThread extends Thread {

        private boolean newTapReceived = false;
        private int     tapCount       = 0;
        private FTouchEvent event       = new FTouchEvent(FTouchEvent.CLICK);

        public TappingThread() {
        }

        public void run() {

            do {
                try {
                    newTapReceived = false;
                    Thread.sleep(600);
                } catch (InterruptedException ioe) {
                }
            } while (newTapReceived);
            tt = null;
            // Handle the tapping event
            event.setTapCount(tapCount);
            handleFTouchEvent(event);
        }

        public void newTap(int pressedX, int pressedY, int releasedX, int releasedY) {
            tapCount++;
            event.addPressCoordinates(pressedX, pressedY);
            event.addReleaseCoordinates(releasedX, releasedY);
            // Reset the thread waiting
            newTapReceived = true;
        }

    }

    public TouchCanvas() {
        super();
    }

    public void pointerPressed (int x, int y) {
        lastPressedX = x;
        lastPressedY = y;
        lastDraggedX = -1;
        lastDraggedY = -1;
        FTouchEvent event = new FTouchEvent(FTouchEvent.PRESSED);
        event.addPressCoordinates(x, y);
        handleFTouchEvent(event);
    }

    public void pointerReleased (int x, int y) {
        // Detect basic movements (vertical and horizontal)
        int deltaY = Math.abs(y - lastPressedY);
        int deltaX = Math.abs(x - lastPressedX);

        int minXDist = 5;
        int minYDist = 5;
        float minYGrad = 2.0f;
        float minXGrad = 2.0f;

        // Horizontal and vertical scrolls are detected based on the
        // gradient of the movement
        if (deltaY > minYDist && ((float)deltaY/(float)deltaX) > minYGrad) {
            // This is a vertical movement
            if (y - lastPressedY > 0) {
                // Scroll up
                FTouchEvent event = new FTouchEvent(FTouchEvent.SCROLL_UP);
                event.addPressCoordinates(lastPressedX, lastPressedY);
                event.addReleaseCoordinates(x, y);
                handleFTouchEvent(event);
            } else {
                // Scroll down
                FTouchEvent event = new FTouchEvent(FTouchEvent.SCROLL_DOWN);
                event.addPressCoordinates(lastPressedX, lastPressedY);
                event.addReleaseCoordinates(x, y);
                handleFTouchEvent(event);
            }
        } else if (deltaX > minXDist && ((float)deltaX/(float)deltaY) > minXGrad) {
            // This is an horizontal movement
            if (x - lastPressedX > 0) {
                // Scroll right
                FTouchEvent event = new FTouchEvent(FTouchEvent.SCROLL_RIGHT);
                event.addPressCoordinates(lastPressedX, lastPressedY);
                event.addReleaseCoordinates(x, y);
                handleFTouchEvent(event);
            } else {
                // Scroll left
                FTouchEvent event = new FTouchEvent(FTouchEvent.SCROLL_LEFT);
                event.addPressCoordinates(lastPressedX, lastPressedY);
                event.addReleaseCoordinates(x, y);
                handleFTouchEvent(event);
            }
        } else {
            // TODO: Detect multi tapping
            // We start a thread and if no more events happen within short time,
            // then we assume tapping is over and reset it
            if (tt == null) {
                tt = new TappingThread();
                tt.newTap(lastPressedX, lastPressedY, x, y);
                tt.start();
            } else {
                tt.newTap(lastPressedX, lastPressedY, x, y);
            }
        }
    }

    public void pointerDragged(int x, int y) {
        // Detect scrolling movement while in progress
        if (lastDraggedX == -1) {
            lastDraggedX = lastPressedX;
        }
        if (lastDraggedY == -1) {
            lastDraggedY = lastPressedY;
        }
        int deltaY = y - lastDraggedY;
        int deltaX = x - lastDraggedX;
        if (y - lastDraggedY > 3 && Math.abs(deltaX) < 5) {
            // Scroll up
            FTouchEvent event = new FTouchEvent(FTouchEvent.DRAGGING_UP);
            event.addPressCoordinates(lastPressedX, lastPressedY);
            event.addDraggingCoordinates(x, y);
            handleFTouchEvent(event);
        } else if (lastDraggedY - y > 3 && Math.abs(deltaX) < 5) {
            // Scroll down
            FTouchEvent event = new FTouchEvent(FTouchEvent.DRAGGING_DOWN);
            event.addPressCoordinates(lastPressedX, lastPressedY);
            event.addDraggingCoordinates(x, y);
            handleFTouchEvent(event);
        } else if (x - lastDraggedX > 3 && Math.abs(deltaY) < 5) {
            // Scroll right
            FTouchEvent event = new FTouchEvent(FTouchEvent.DRAGGING_RIGHT);
            event.addPressCoordinates(lastPressedX, lastPressedY);
            event.addDraggingCoordinates(x, y);
            handleFTouchEvent(event);
        } else if (lastDraggedX - x > 3 && Math.abs(deltaY) < 5) {
            // Scroll left
            FTouchEvent event = new FTouchEvent(FTouchEvent.DRAGGING_LEFT);
            event.addPressCoordinates(lastPressedX, lastPressedY);
            event.addDraggingCoordinates(x, y);
            handleFTouchEvent(event);
        }
        lastDraggedX = x;
        lastDraggedY = y;
    }

    protected void showKeyboard(boolean enableKeyboard) {
        isKeyboardOnScreen = enableKeyboard;
    }

    protected boolean isKeyboardOnScreen() {
        return isKeyboardOnScreen;
    }

}
