/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.ui.view;

import net.rim.device.api.lcdui.BlackBerryCanvas;
import net.rim.device.api.lcdui.control.VirtualKeyboardControl;
import net.rim.device.api.ui.TouchEvent;
import net.rim.device.api.ui.TouchGesture;

import java.util.Timer;
import java.util.TimerTask;

/**
 * This class implements a BlackBerryCanvas with the ability to listen to touch
 * events specific of BlackBerry OS4.7 with touch screen support. It recognizes
 * some basic gestures.
 * The class is abstract as the handleFTouchEvent must be implemented to handle
 * events.
 */
abstract class TouchCanvas extends BlackBerryCanvas {

    private int  lastDownX = 0;
    private int  lastDownY = 0;

    private final int FAST_SCROLL_COUNT = 15;

    private boolean isKeyboardOnScreen = false;

    VirtualKeyboardControl vk = null;

    protected abstract void handleFTouchEvent(FTouchEvent event);

    public TouchCanvas() {
        super();
    }

    public void touchEvent(TouchEvent message) {
        
        FTouchEvent event;
        switch(message.getEvent()) {
            case TouchEvent.CLICK:
                // A click event has been detected
                // First of all select the item
                event = new FTouchEvent(FTouchEvent.PRESSED);
                event.addPressCoordinates(message.getGlobalX(1), message.getGlobalY(1));
                handleFTouchEvent(event);
                // Then open the item
                event = new FTouchEvent(FTouchEvent.CLICK);
                event.addPressCoordinates(message.getGlobalX(1), message.getGlobalY(1));
                //Simulate double click when the click touch event is detected
                event.setTapCount(2);
                handleFTouchEvent(event);
                break;
            case TouchEvent.DOWN:
                // Keep track of the last touch down position
                lastDownX = message.getGlobalX(1);
                lastDownY = message.getGlobalY(1);
                break;
            case TouchEvent.MOVE:
                int pointSize = message.getMovePointsSize();
                int[] x = new int[pointSize];
                int[] y = new int[pointSize];
                int[] time = new int[pointSize];
                message.getMovePoints(1, x, y, time);
                if(pointSize > 1) {
                    lastDownY = x[pointSize-2];
                    lastDownY = y[pointSize-2];
                }
                int dx = x[pointSize-1] - lastDownX;
                int dy = y[pointSize-1] - lastDownY;
                // Scroll up
                if (dy > 3 && Math.abs(dx) < 20) {
                    event = new FTouchEvent(FTouchEvent.DRAGGING_UP);
                    event.addPressCoordinates(lastDownX, lastDownY);
                    event.addDraggingCoordinates(x[pointSize-1], y[pointSize-1]);
                    handleFTouchEvent(event);
                }
                // Scroll down
                else if (dy < -3 && Math.abs(dx) < 20) {
                    event = new FTouchEvent(FTouchEvent.DRAGGING_DOWN);
                    event.addPressCoordinates(lastDownX, lastDownY);
                    event.addDraggingCoordinates(x[pointSize-1], y[pointSize-1]);
                    handleFTouchEvent(event);
                }
                break;
            case TouchEvent.GESTURE:
                // A set of basic touch events have been detected
                TouchGesture gesture = message.getGesture();
                // Quick touch and release gesture on the touchscreen
                if(gesture.getEvent() == TouchGesture.TAP) {
                    event = new FTouchEvent(FTouchEvent.PRESSED);
                    event.addPressCoordinates(message.getGlobalX(1), message.getGlobalY(1));
                    handleFTouchEvent(event);
                }
                // Quick motion gesture across the touchscreen
                else if(gesture.getEvent() == TouchGesture.SWIPE) {
                    // Scroll left motion
                    if(gesture.getSwipeDirection() == TouchGesture.SWIPE_WEST) {
                        event = new FTouchEvent(FTouchEvent.SCROLL_RIGHT);
                        handleFTouchEvent(event);
                    }
                    // Scroll right motion
                    else if(gesture.getSwipeDirection() == TouchGesture.SWIPE_EAST) {
                        event = new FTouchEvent(FTouchEvent.SCROLL_LEFT);
                        handleFTouchEvent(event);
                    }
                    // Scroll up motion
                    if(gesture.getSwipeDirection() == TouchGesture.SWIPE_NORTH) {
                        event = new FTouchEvent(FTouchEvent.DRAGGING_DOWN);
                        handleMultipleFTouchEvents(event);
                    }
                    // Scroll down motion
                    if(gesture.getSwipeDirection() == TouchGesture.SWIPE_SOUTH) {
                        event = new FTouchEvent(FTouchEvent.DRAGGING_UP);
                        handleMultipleFTouchEvents(event);
                    }
                }
                break;
        }
    }

    /**
     * Handle multiple FTouchEvents
     * @param event
     * @param dt
     */
    private void handleMultipleFTouchEvents(FTouchEvent event) {
        Timer timer = new Timer();
        TouchTask task;
        for(int i=0; i<FAST_SCROLL_COUNT; i++) {
            task = new TouchTask(event);
            int base = (i*i*i)/8;
            timer.schedule(task, base );
        }
    }

    public class TouchTask extends TimerTask {
        private FTouchEvent event;
        public TouchTask(FTouchEvent e) {
            super();
            event = e;
        }
        public void run() {
            handleFTouchEvent(event);
        }
    }

    protected void showKeyboard(boolean enableKeyboard) {
        vk = (VirtualKeyboardControl) getControl("net.rim.device.api.lcdui.control.VirtualKeyboardControl");
        if (vk!=null) {
            if (enableKeyboard) {
                vk.setKeyboardVisibility(VirtualKeyboardControl.KEYBOARD_SHOW);
            } else {
                vk.setKeyboardVisibility(VirtualKeyboardControl.KEYBOARD_HIDE);
            }
        }
        isKeyboardOnScreen = enableKeyboard;
    }

    protected boolean isKeyboardOnScreen() {
        return isKeyboardOnScreen;
    }
}

