/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 *
 *
 */
package com.funambol.mailclient.ui.utils;

import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;

import com.funambol.platform.FileAdapter;
import com.funambol.util.Log;


/**
 * This class builds the thumbnail for a given image. This implementation
 * supports two different methods to compute thumbnails.
 * <UL>
 * <LI> Nokia's hidden directories with thumbnails
 * <LI> Pure java resizing algorithm
 * </UL>
 * The first approach is very fast and gentle on memory consumption, but it
 * works only on Nokia's devices and somehow the "Gallery" application must have
 * created the thumbnails.
 * The second approach is more general and it works on any device, but it
 * needs to expand the image in memory. Given the amount of heapsize
 * that MIDlets are given, it is frequent that the algorithm is not able to load
 * the image in memory to shrink it.
 */
public class ThumbnailGenerator {

    private String path;
    private String name;
    private String thumbDir;
    private boolean thumbAvailable;
    private boolean rotate = false;

    public ThumbnailGenerator(String path, String name) {
        this.path = path;
        this.name = name;
    }

    public Image createThumbnail(int width, int height) throws IOException {
        // We have different ways of creating the thumbnail. We choose the best
        // one here

        Image img = null;
        boolean nokia;
        String platform = System.getProperty("microedition.platform");
        Log.trace("Creating thumbnail on " + platform);
        if (platform != null && platform.indexOf("Nokia") != -1) {
            nokia = true;
        } else {
            nokia = false;
        }
        if (nokia) {
            if (platform.indexOf("N80") != -1) {
                rotate = true;
            }
            //If this fails, shall we revert to the default method?
            img = createThumbnailFromThumbDB();
        }

        if (img == null) {
            try {
                img = createThumbnailByResizing(width, height);
            } catch (OutOfMemoryError ooe) {
                // This case is frequent enough to be handled here
                throw new IOException(ooe.toString());
            }
        }

        return img;
    }

    private Image createThumbnailByResizing(int width, int height) throws IOException {

        String fileName = "file:///" + path + name;

        FileAdapter fa = new FileAdapter(fileName, true);
        InputStream is = fa.openInputStream();
        Image origImage = Image.createImage(is);
        Image thumbnail = resize(origImage, width, height, 0);
        is.close();
        fa.close();
        return thumbnail;
    }

    private final Image resize(Image image, int expectedWidth,
                                            int expectedHeight, int padding)
    {
        final int sourceWidth = image.getWidth();
        final int sourceHeight = image.getHeight();
        int thumbWidth = -1;
        int thumbHeight = -1;

        // big width
        if(sourceWidth >= sourceHeight) {
            thumbWidth = expectedWidth - padding;
            thumbHeight = thumbWidth * sourceHeight / sourceWidth;
            // fits to height ?
            if(thumbHeight > (expectedHeight - padding)) {
                thumbHeight = expectedHeight - padding;
                thumbWidth = thumbHeight * sourceWidth / sourceHeight;
            }
        } else {
            // big height
            thumbHeight = expectedHeight - padding;
            thumbWidth = thumbHeight * sourceWidth / sourceHeight;
            // fits to width ?
            if(thumbWidth > (expectedWidth - padding)) {
                thumbWidth = expectedWidth - padding;
                thumbHeight = thumbWidth * sourceHeight / sourceWidth;
            }
        }

        thumbHeight = (sourceHeight < thumbHeight) ? sourceHeight : thumbHeight;
        thumbWidth = (sourceWidth < thumbWidth) ? sourceWidth : thumbWidth;

        return getThumbnail(image, thumbWidth, thumbHeight);
    }

    private final Image getThumbnail(Image image, int thumbWidth, int thumbHeight) {
        int x, y, pos, tmp, z = 0;
        final int sourceWidth = image.getWidth();
        final int sourceHeight = image.getHeight();

        // integer ratio ..
        final int ratio = sourceWidth / thumbWidth;

        // buffer where we read in data from image source
        final int[] in = new int[sourceWidth];

        // buffer of output thumbnail image
        final int[] out = new int[thumbWidth*thumbHeight];

        final int[] cols = new int[thumbWidth];

        // pre-calculate columns we need to access from source image
        for (x = 0,pos = 0; x < thumbWidth; x++) {
            cols[x] = pos;

            // increase the value without fraction calculation
            pos += ratio;
            tmp = pos + (thumbWidth - x) * ratio;
            if(tmp > sourceWidth) {
                pos--;
            } else if(tmp < sourceWidth - ratio) {
                pos++;
            }
        }

        // read through the rows ..
        for (y = 0, pos = 0, z = 0; y < thumbHeight; y++) {

            // read a single row ..
            image.getRGB(in, 0, sourceWidth, 0, pos, sourceWidth, 1);

            for (x = 0; x < thumbWidth; x++, z++) {
                // write this row to thumbnail
                out[z] = in[cols[x]];
            }

            pos += ratio;
            tmp = pos + (thumbHeight - y) * ratio;
            if(tmp > sourceHeight) {
                pos--;
            } else if(tmp < sourceHeight - ratio) {
                pos++;
            }
        }
        return Image.createRGBImage(out, thumbWidth, thumbHeight, false);
    }

    private String searchThumbnail(String baseDir, String name, String searchName) throws IOException {

        if (!name.endsWith("/")) {
            // This is a file
            if (name.indexOf(searchName) != -1) {
                // Found it!
                return baseDir + name;
            }
        } else {
            String thumbDir = baseDir + name;
            FileAdapter dirConn = new FileAdapter(thumbDir, true);
            Enumeration files = dirConn.list(true);
            while(files.hasMoreElements()) {
                String f = (String) files.nextElement();
                String ret = searchThumbnail(baseDir, name + f, searchName);
                if (ret != null) {
                    // Close the connection and return
                    dirConn.close();
                    return ret;
                }
            }
            dirConn.close();
        }
        return null;
    }


    private Image createThumbnailFromThumbDB() throws IOException {

        // We shall search for the thumbnails, browsing subdirectories
        Image thumbnail = null;
        String fullName = searchThumbnail("file:///" + path, "_PAlbTN/", name);
        if (fullName != null) {
            FileAdapter fa = new FileAdapter(fullName, true);
            InputStream is = fa.openInputStream();
            thumbnail = Image.createImage(is);
            if (rotate) {
                // We must rotate it as it is landscape
                thumbnail = Image.createImage(thumbnail, 0, 0,
                        thumbnail.getWidth(), thumbnail.getHeight(),
                        Sprite.TRANS_ROT90);
            }
            fa.close();
        }
        return thumbnail;
    }
}

 
