/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2010 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 *
 *
 */
package com.funambol.mailclient.ui.controller;

import java.io.IOException;
import java.util.Date;
import java.util.Vector;

import com.funambol.mail.Folder;
import com.funambol.mail.AccountFolder;
import com.funambol.mail.MailException;
import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;

import com.funambol.mailclient.CTPManager;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.mm.MessageManager;
import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.ui.view.InboxMessageList;
import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.mailclient.syncml.FolderData;
import com.funambol.mailclient.syncml.MailSyncItemHelper;

import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.protocol.DevInf;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.syncml.spds.SyncReport;
import com.funambol.syncml.spds.BasicSyncListener;
import com.funambol.util.Log;
import com.funambol.util.MailDateFormatter;
import com.funambol.mailclient.ui.utils.UserNotificationManager;


/* Note:
 * In multisession mode (when we synchronize more repositories in the same
 * session), we assume that the MailSyncListener is the main listener.
 * The main listener is in charge of performing operations related to the
 * beginning/end of the entire session (such as enabling/disabling sync button).
 */
public class MailSyncListener extends BasicSyncListener {

    private static final String TAG_LOG = "MailSyncListener";

    /** Number of items received so far (in the current session) */
    private int received;
    /** Number of items sent so far (in the current session) */
    private int sent;
    /** Total number of items to be received in this session */
    private int toReceive;
    /** Total number of items to be sent in this session */
    private int toSend;
    private boolean alerted;
    /** Indicates if this sync is part of a multisession (sync of more
     * repositories)
     **/
    private boolean multisession;
    /** Session exit status */
    private int status = -1;
    /** Indicates if this is the first block to be sent */
    private boolean firstSend;

    private int syncMode;
    
    private boolean isFirstMessageReceived;
    
    private Vector receivedAccounts = null;

    /** Creates a new instance of MailSyncListener */
    public MailSyncListener() {

        init();
    }

    public void itemReceived(Object item) {

        Log.debug(TAG_LOG, "Item received");
        
        // we need to reset the inbox but we don't want the user
        // to catch us performing in a slow sync required
        // from the server... so we refresh the  inbox 
        // only when we are doing a refresh from server (that
        // has probably been requested by our client)
        // reload of th inbox has been moved to itemReceived() method
        
        if ((syncMode == SyncML.ALERT_CODE_REFRESH_FROM_SERVER 
               || syncMode == SyncML.ALERT_CODE_SLOW)
               && isFirstMessageReceived) {
            Log.debug(TAG_LOG, "Reloading Message List");
            UIController.reloadInboxMessageListMessages();
            isFirstMessageReceived=false;
        }
       
        // Safety check, even though this should never be the case
        if (item == null) {
            return;
        }

        if(item instanceof Message) {
            Message msg = (Message) item;

            msg.setKey(MailSyncItemHelper.filterDefaultAccount(msg.getKey(), false));

            String userInfoMessage;

            // Update the number of received items
            received++;

            if (toReceive != ITEMS_NUMBER_UNKNOWN) {
                userInfoMessage = Localization.getMessages().
                        getMESSAGE_X_OF_Y_RECEIVED(received, toReceive)+"...";
            } else {
                userInfoMessage = Localization.getMessages().
                        getMESSAGE_X_RECEIVED(received)+"...";
            }
            UIController.updateInboxMessageList(userInfoMessage);

            String folderPath = MessageManager.getInstance().folderFromKey(msg.getKey());
            Folder targetFolder = MessageManager.getInstance().getFolder(folderPath);

            if (targetFolder != null) {
                String parentRole = targetFolder.getRole();
                // TODO: play sound only if the message has unread flag
                if ((FolderData.INBOX_ROLE).equals(parentRole)) {
                    if (!alerted && (!msg.getFlags().isSet(MessageFlags.OPENED))) {
                        UserNotificationManager unm = UserNotificationManager.getInstance();
                        unm.playNewEmailTone();
                        unm.showNewEmailIcon();
                        alerted = true;
                    }
                    InboxMessageList messageList = UIController.getInboxMessageList();
                    try {
                        messageList.addMessage(msg);
                    } catch (MailException ex) {
                        Log.error(TAG_LOG, "handleInboxMessage: " + ex.toString());
                        ex.printStackTrace();
                    }
                }
            }
        } else if(item instanceof AccountFolder) {
            // Notify the inbox of the new account
            UIController.getInboxMessageList().newAccountAdded((AccountFolder)item);
            if (receivedAccounts == null) {
                receivedAccounts = new Vector();
            }
            receivedAccounts.addElement(item);
        }
    }

    public void startSession() {

        Log.debug(TAG_LOG, "Start session");

        if (!multisession) {
            initializeSession();
        }
    }

    private void initializeSession() {
        //DISABLE GetMail Command during sync email
        UIController.enableSyncCommands(false);
        // Update msg
        String userInfoMessage = Localization.getMessages().CONNECTING;
        UIController.updateInboxMessageList(userInfoMessage);
    }

    public void endSession(SyncReport report) {

        Log.debug(TAG_LOG, "End session");
        this.status = report.getSyncStatus();

        if (Log.getLogLevel() >= Log.INFO) {
            Log.info(TAG_LOG, report.toString());
        }

        // If this is not a multisession then we update the UI
        if (!multisession) {
            finalizeSession();
        }
        
        MailClientConfig mcc = null;
        try {
            mcc = ConfigManager.getConfig();
        } catch (ConfigException ex) {
            ex.printStackTrace();
        }
        
        if (mcc.getCtpPushStatus()==mcc.CTP_ENABLED) {
            if (!CTPManager.getInstance().isRunning()) {
                Log.debug(TAG_LOG, "CTP is not running, start it");
                CTPManager.getInstance().startCTP();
            } else {
                Log.debug(TAG_LOG, "CTP is already running");
            }
        }

        // Tell the InboxMessageList that the sync ended
        UIController.getInboxMessageList().setSyncEnded();
    }

    public void endConnecting(int status) {

        Log.debug(TAG_LOG, "End connecting");

        // We anticipate the "Checking emails..." message as we want it to
        // last a little bit. If we display it in the startReceiving it gets
        // clobbered too soon by the "Receiving #"
        String msg = Localization.getMessages().LOADING;

        // Resetting inbox: the server answer OK to the 
        // Refresh from server request
        if (status == SyncML.ALERT_CODE_REFRESH_FROM_SERVER) {
            msg = Localization.getMessages().RESETTING_MSG;
            // Messages are deleted only after the connection has been 
            // established
            //UIController.messageManager.emptyFolder(UIController.messageManager.INBOX_ID);
        }

        UIController.updateInboxMessageList(msg);
    }

    public void syncStarted(int syncMode) {
        this.syncMode=syncMode;
    }
    
    public boolean startSyncing(int syncMode, DevInf devInf) {
    
        // we always allow the sync to start. If needed, here we can 
        // ask the user if it's ok to perform a sync (e.g. for slow syncs)

        // Update the server devinf
        if(devInf != null) {
            try {
                ServerCapStore.getInstance().save(devInf);
            } catch (IOException ioe) {
                Log.error(TAG_LOG, "Cannot save server caps " + ioe.toString());
            }
        }

        receivedAccounts = null;

        return true; 
    } 

    public void startReceiving(int numItems) {

        Log.debug(TAG_LOG, "Start receiving (" + numItems + ")");
        // Save the number of items to be received
        toReceive = numItems;
    }

    public void endReceiving() {

        Log.debug(TAG_LOG, "End receiving");
        toReceive = 0;
    }

    public void itemDeleted(Object item) {

        Log.debug(TAG_LOG, "Item deleted");
        String key = MailSyncItemHelper.filterDefaultAccount((String) item, false);

        if(isAccountkey(key)) {
            UIController.getInboxMessageList().accountRemoved(key);
        } else {
            //if message is in inbox, we delete it.
            UIController.getInboxMessageList().deleteMessage(key);
        }
    }

    public void itemUpdated(Object item, Object update) {

        Log.debug(TAG_LOG, "Item updated by msgFlagsUpdated");

        // In some case the source does not perform the update and passes over
        // a null update object. In these cases we do not have to update
        // anything, so we ignore this event
        if (update == null) {
            return;
        }

        String key = MailSyncItemHelper.filterDefaultAccount((String) item, false);
        MessageFlags newFlags = (MessageFlags) update;

        //if message is in inbox, we update the flags
        UIController.getInboxMessageList().updateMessageFlags(key, newFlags);
    }

    public void itemUpdated(Object item) {

        String key = MailSyncItemHelper.filterDefaultAccount((String) item, false);
        String folderPath = MessageManager.getInstance().folderFromKey(key);
        Folder folder = MessageManager.getInstance().getFolder(folderPath);

        // We only update items in the INBOX_ID
        if (folder.getRole().equals(FolderData.INBOX_ROLE)) {
            //MessageFlags newMask = FlagQueue.get(key);

            MessageManager mm = MessageManager.getInstance();
            Message msg = mm.getMessage(folder, key);

            if (msg != null) {
                MessageFlags newFlags = msg.getFlags();

                UIController.getInboxMessageList().updateMessageFlags(key,
                        newFlags);
            } else {
                Log.info(TAG_LOG, "Message not found: it could have been deleted");
            }
        }
    }

    /**
     * Check if the specified key represents an AccountFolder
     * @param key
     * @return
     */
    public boolean isAccountkey(String key) {
        String first = key.substring(1);
        String second = first.substring(first.indexOf("/")+1);
        return second.indexOf("/") < 0;
    }
    
    public void startSending(int numNewItems, int numUpdItems, int numDelItems) {

        Log.debug(TAG_LOG, "Start sending (" + numNewItems + "," + numUpdItems + "," + numDelItems + ")");
        toSend = numNewItems;

        //if we have something NEW to send. Other changes do not trigger the
        //SENDING message. In the future we may display a send updates msg
        if (toSend != 0 && firstSend) {
            // The number of items to be sent does not correspond to the number
            // of emails to be sent, but rather to the number of total changes.
            // Since this number can be misleading, at the moment we just
            // display a generic sending message
            String userInfoMessage = Localization.getMessages().SENDING_DOTTED;

            // TODO Is this really necessary? can we be outside a session???
            UIController.enableSyncCommands(false);
            UIController.updateInboxMessageList(userInfoMessage);
        }
    }
 
    public void itemAddSendingStarted(String key, String parent, int size) {

        Log.debug(TAG_LOG, "New item sent");
        String msg;
        if (toSend > 0) {
            // If the overall number of new items to send is known we show the
            // it in the message
            msg = Localization.getMessages().getSENDINGX_OF_Y(++sent, toSend);
        } else {
            msg = Localization.getMessages().getSENDINGX(++sent);
        }
        UIController.updateInboxMessageList(msg);
    }

    public void endSending() {

        Log.debug(TAG_LOG, "End sending");
        toSend = 0;
        firstSend = false;

    // TODO Why don't we re-enable the sync commands??? I think this must be
    // consistent with the above

    // We do not update the title here and wait for it to be rewritten at
    // the beginning of the receving phase. This way we avoid an annoying
    // transition Loading --> Inbox --> Loading and we keep the Loading.
    }

    public void dataReceived(String date, int size) {

        Log.debug(TAG_LOG, "Data received");

        // date like Thu, 03 May 2007 14:45:38 GMT
        try {
            if (!ConfigManager.getConfig().getTimeZoneUpdated()) {

                Date gmtServerdate = MailDateFormatter.parseRfc2822Date(date);
                Date deviceDate = new Date();

                int millisServerDate = (int) gmtServerdate.getTime();
                int millisDeviceDate = (int) deviceDate.getTime();
                int hourServer = (int) (millisServerDate / 3600000); //60*60*1000
                int hourDevice = (int) (millisDeviceDate / 3600000); //60*60*1000
                int hourDiff = hourDevice - hourServer;
                int minServer = (int) (millisServerDate / 60000); //60*60*1000
                int minDevice = (int) (millisDeviceDate / 60000); //60*60*1000
                int minDiff = minDevice - minServer - (hourDiff * 60);
                StringBuffer offset = new StringBuffer();
                offset.append((hourDiff > 0) ? "+" : "-");
                int mathHourDiff = (Math.abs(hourDiff));
                offset.append((mathHourDiff <= 9) ? "0" + mathHourDiff : "" + mathHourDiff);
                int mathMinDiff = (Math.abs(minDiff));
                if (mathMinDiff <= 20 || mathMinDiff >= 55) {
                    offset.append("00");
                } else if (mathMinDiff > 20 || mathMinDiff < 40) {
                    offset.append("30");
                } else if (mathMinDiff >= 40 || mathMinDiff < 55) {
                    offset.append("45");
                }

                // Save the new config.
                ConfigManager.getConfig().setTimeZone(offset.toString());
                ConfigManager.saveConfig();
            }
        } catch (ConfigException cex) {
            Log.error(TAG_LOG, "dataReceived - " + cex.toString());
        } catch (IOException ioex) {
            Log.error(TAG_LOG, "dataReceived - " + ioex.toString());
        }
    }

    /**
     * This method returns the status as reported to the endSession method.
     * The value returned is only meaningful at the end of a sync, otherwise
     * -1 is returned
     */
    public int getStatus() {
        return status;
    }

    /**
     * @return an error string based on the error status received
     */
    private static String getErrorMessage(int status) {
        Log.debug(TAG_LOG, "Getting Error Message from error status: [" + status + "]");

        String message;
        switch (status) {
            case SyncSource.STATUS_CONNECTION_ERROR:
                message = Localization.getMessages().ACCOUNT_LISTENER_ERROR_CONNECTIONG;
                break;
            case SyncSource.STATUS_RECV_ERROR:
                message = Localization.getMessages().ACCOUNT_LISTENER_ERROR_RECEIVING;
                break;
            case SyncSource.STATUS_SEND_ERROR:
                message = Localization.getMessages().ACCOUNT_LISTENER_ERROR_SENDING;
                break;
            case SyncSource.STATUS_SERVER_ERROR:
                message = Localization.getMessages().ACCOUNT_LISTENER_ERROR_FROM_SERVER;
                break;
            default:
                message = Localization.getMessages().ACCOUNT_LISTENER_ERROR_DEFAULT;
                break;

        }
        return message;
    }

    /**
     * This method does not belong to the SyncListener interface.
     * A multi session is a concept introduced by the SyncClient which is able
     * to perform the sync of multiple repositories.
     *
     * @param repositories repositories to be synchronized
     */
    public void startMultiSession(boolean repositories[]) {

        initializeSession();
        multisession = true;
    }

    /**
     * This method does not belong to the SyncListener interface.
     * A multi session is a concept introduced by the SyncClient which is able
     * to perform the sync of multiple repositories.
     *
     */
    public void endMultiSession() {

        // At the end of the multisession we must reset the Inbox title with the
        // number of new messages received
        finalizeSession();
        multisession = false;
    }

    private void finalizeSession() {

        //String userInfoMessage;
        if (status != SyncSource.STATUS_SUCCESS && status == SyncSource.STATUS_SEND_ERROR) {
            //userInfoMessage = getErrorMessage(status);
            UIController.showErrorAlert(getErrorMessage(status), UIController.getInboxMessageList());
        }

        //else {
        // userInfoMessage = Localization.getMessages().getX_MESSAGES_RECEIVED(received);
        //UIController.updateInboxMessageList(Localization.getMessages().getX_MESSAGES_RECEIVED(received));
        //}
        // Update the title
        //UIController.updateInboxMessageList(userInfoMessage);

        //Display the "no new message" inbox localized title only when a maual 
        //sync is performed.
        if (received<=0&&UIController.getSyncCaller()!=UIController.USER) {
            //in this case (0 messages received after a non manual sync) the 
            //default inbox title is used in order not to worry
            //the user about no new messages found after scheduled and SAN 
            //initiated sync 
            Log.debug(TAG_LOG, "Avoid to update the inbox title");
            UIController.updateInboxMessageList(Localization.getMessages().INBOX_DEFAULT_TITLE);
        } else {
            //All other cases:
            //1) Manual sync and 0 messages received
            //2) Non manual sync and more than 0 messages received
            //Definitely only in case of a manual sync, the user is notified 
            //with the localized message "No new messages".
            Log.debug(TAG_LOG, "Updating the inbox title");
            UIController.updateInboxMessageList(Localization.getMessages().getX_MESSAGES_RECEIVED(received));
        }

        // If we received a new account, we shall check that all the necessary
        // folders were received. If any of them is missing, we create it here

        if (receivedAccounts != null) {
            for(int i=0;i<receivedAccounts.size();++i) {
                AccountFolder newAccount = (AccountFolder)receivedAccounts.elementAt(i);
                // We expect inbox and outbox to be sent because they are
                // mandatory for synchronization.
                // Sent and drafts are optional, so we verify if we get them.
                try {
                    createFolderIfNecessary(newAccount, "drafts", FolderData.DRAFTS_ROLE);
                } catch (MailException me) {
                    Log.error(TAG_LOG, "Cannot create missing drafts folder for account "
                              + newAccount.getFullName());
                }
                try {
                    createFolderIfNecessary(newAccount, "sent", FolderData.SENT_ROLE);
                } catch (MailException me) {
                    Log.error(TAG_LOG, "Cannot create missing sent folder for account "
                              + newAccount.getFullName());
                }
            }
        }
                
        UIController.enableSyncCommands(true);

        // Perform a check for new versions (if required)
        //UpdaterManager.getInstance().check();
    }

    /**
     * Creates a folder if it does not exist already.
     */
    private void createFolderIfNecessary(AccountFolder account, String folderName,
                                         String role) throws MailException
    {
        Folder folder = null;
        MessageManager mm = MessageManager.getInstance();
        try {
            folder = mm.getAccountFolderFromRole(account, role);
        } catch (MailException me) {
            // If we cannot access the folder, then we re-create it. This is not
            // necessarly an error, so we log at info
            Log.info(TAG_LOG, "Cannot access folder with role " + role + " recreating it");
        }

        if (folder == null) {
            Log.info(TAG_LOG, "Creating drafts folder for " + account.getFullName());
            folder = mm.createFolder(account, folderName, role, account.getCreated());
            mm.addFolder(folder);
        }
    }

    private void init() {

        received = 0;
        sent = 0;
        toReceive = ITEMS_NUMBER_UNKNOWN;
        toSend = ITEMS_NUMBER_UNKNOWN;
        alerted = false;
        multisession = false;
        firstSend    = true;
        isFirstMessageReceived  = true; 
    }
}
