/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.ui.controller;

import com.funambol.mailclient.config.ConfigException;
import com.funambol.storage.AbstractRecordStore;
import com.funambol.syncml.protocol.DevInf;
import com.funambol.syncml.protocol.Ext;
import com.funambol.util.CodedException;
import com.funambol.util.Log;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Vector;

import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreNotFoundException;

/**
 * This class is responsible of the storing of the server capabilities.
 */
public class ServerCapStore {

    private static final String TAG_LOG = "ServerCapStore";

    //the store name
    private static final String STORE_NAME = "ServerCapabilities";

    // config versions
    private static final int VERSION_100 = 100; // First version
    private static final int VERSION_101 = 101; // Added Ext field for MEA

    private static final int CURRENT_VERSION = VERSION_101;

    /** Check whether the SyncML Devinf properies are required from server */
    private static boolean requireServerDevInf = false;

    /**
     * This is the current version. It represents the version of the stored
     * capabilities.
     */
    private int version = -1;
    
    // the object with servercaps
    private DevInf serverCapabilities = null;

    private static ServerCapStore instance = null;

    //--------------------------------------------------------- constructors
    //private constructor to prevent this to be istantiated
    private ServerCapStore() {
        try {
            loadCapabilities();
            // The capabilities are valid if the version is the expected one,
            // otherwise we consider them invalid
            if (version < CURRENT_VERSION) {
                serverCapabilities = null;
            }
        } catch (IOException e) {
            serverCapabilities = null;
        }
    }

    public static ServerCapStore getInstance() {
        if (instance == null) {
            instance = new ServerCapStore();
        }
        return instance;
    }

    //--------------------------------------------------------- public methods
    /**
     * Delete all the saved capabilities from memory and persistent storage.
     */
    public void delete() {
        Log.info(TAG_LOG, "Deleting server capabilities saved config");
        serverCapabilities = null;
        deleteStoredConfig();
    }

    /**
     * Get the server capabilities from storage
     * 
     * @return the server capabilities (maybe null if not available for any
     * reason)
     */
    public DevInf get() {
        if(serverCapabilities == null) {
            try {
                loadCapabilities();
            } catch(Exception ex) { 
                Log.error(TAG_LOG, "Error while loading server capabilities");
            }
        }
        return serverCapabilities;
    }

    /**
     * Updates the Server Capabilities. The new DevInf is required in order to
     * add the missing Ext properties.
     * 
     * @param inf
     */
    public void save(DevInf inf) throws IOException {
        if(inf == null) {
            return;
        }
        serverCapabilities = inf;
        saveCapabilities();
    }

    /**
     * Check if the Multiple Email Account features are supported by the server
     * @return
     * @throws com.funambol.mailclient.config.ConfigException
     * @throws com.funambol.util.CodedException
     */
    public boolean isExtSupported(String extName) {
        Log.trace(TAG_LOG, "isMultipleEmailAccountSupported server cap check");
        boolean ret = false;
        try {
            DevInf sc = get();
            if (sc != null) {
                Vector exts = sc.getExts();
                for (int i = 0; i < exts.size(); i++) {
                    if (((Ext) exts.elementAt(i)).getXNam().equals(extName)) {
                        Log.debug(TAG_LOG, extName + " supported");
                        ret = true;
                        break;
                    }
                }
            }
        } catch(Exception ex) {
            Log.info(TAG_LOG, "Exception while retrieving server capabilities: " + ex);
        }
        Log.trace(TAG_LOG, "isExtSupported: " + extName + " returns " + ret);
        return ret;
    }

    //--------------------------------------------------------- private methods
    private void safeCloseRecordStore(AbstractRecordStore s) {
        if (s != null) {
            try {
                s.closeRecordStore();
            } catch (RecordStoreException e) {
                //nothing to do here!
                Log.error(TAG_LOG, "Cannot close server cap record store");
            }
        }
    }

    // this nukes the rms
    private void deleteStoredConfig() {
        try {
            Log.info(TAG_LOG, "nuking stored server capabilities config");
            AbstractRecordStore.deleteRecordStore(STORE_NAME);
        } catch (RecordStoreNotFoundException ex) {
            // nothing to do here, we're happy the record store was not there! :D
        } catch (RecordStoreException ex) {
            Log.error(TAG_LOG, "error deleting recordstore " + ex);
        }
    }

    /**
     * Write config version to the first record of the recordstore
     */
    private void initRecordStore() {
        Log.trace(TAG_LOG, "initRecordStore");
        AbstractRecordStore s = null;
        try {
            // ensure we're initing an empty recordstore
            s = AbstractRecordStore.openRecordStore(STORE_NAME, true);
            if (s.getNumRecords() > 0) {
                s.closeRecordStore();
                deleteStoredConfig();
                s = s.openRecordStore(STORE_NAME, true);
            }
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DataOutputStream os = new DataOutputStream(baos);
            os.writeInt(version);
            byte[] data = baos.toByteArray();
            s.addRecord(data, 0, data.length);
        } catch (IOException ex) {
            Log.error(TAG_LOG, "Cannot init server cap record store " + ex.toString());
        } catch (RecordStoreException ex) {
            Log.error(TAG_LOG, "Cannot init server cap record store " + ex.toString());
        } finally {
            safeCloseRecordStore(s);
        }
    }

    private void saveCapabilities() throws IOException {
        Log.info(TAG_LOG, "saving config.");
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        DataOutputStream os = new DataOutputStream(baos);
        AbstractRecordStore s = null;
        serialize(os);

        try {
            deleteStoredConfig();
            initRecordStore();
            s = AbstractRecordStore.openRecordStore(STORE_NAME, false);
            // now we have a recordstore opened with 1 records
            s.addRecord(baos.toByteArray(), 0, baos.toByteArray().length);
        } catch (RecordStoreException ex) {
            Log.error(TAG_LOG, "Cannot save server cap record store " + ex.toString());
            throw new IOException(ex.toString());
        } finally {
            // close the recordstore
            safeCloseRecordStore(s);
        }
    }

    /**
     * Read server cap from rms.
     * @throws IOException if the store cannot be read for any reason but it
     * does not exist
     */
    private void loadCapabilities() throws IOException {
        AbstractRecordStore rs = null;
        try {
            Log.info(TAG_LOG, "Getting server cap from storage");
            serverCapabilities = null;
            rs = AbstractRecordStore.openRecordStore(STORE_NAME, false);
            // the recordstore contains version number and the record with the 
            // server capabilities. if we have a number of records that is
            // different from 2 we have an issue
            if (rs.getNumRecords() == 2) {
                Log.debug(TAG_LOG, "2 records found");
                version = getStoredVersion(rs);
                //deserialize method fills serverCapabilities if able
                deserialize(getStoredCapabilities(rs));
            } else {
                Log.info(TAG_LOG, rs.getNumRecords() + " records found, not expected");
            }
        } catch (RecordStoreNotFoundException rse) {
            // The capabilities have never been got, return null
            Log.info(TAG_LOG, "Server capabilities store not present");
        } catch (Exception ex) {
            Log.error(TAG_LOG, "Error reading server cap from the store " + ex);
            throw new IOException(ex.toString());
        } finally {
            safeCloseRecordStore(rs);
        }
    }

    private void serialize(DataOutputStream out) throws IOException {

        Log.info(TAG_LOG, "Serializing stuff");
        boolean supportSlowSync = false;
        boolean supportMultipleEmailAccount = false;
        Vector v = serverCapabilities.getExts();

        for (int i = 0; i <v.size(); i++) {
            Ext e = (Ext) v.elementAt(i);
            if (e.getXNam().equals(Ext.X_FUNAMBOL_SMARTSLOW)) {
                supportSlowSync = true;
            } else if (e.getXNam().equals(Ext.X_FUNAMBOL_MULTIPLE_EMAIL_ACCOUNT)) {
                supportMultipleEmailAccount = true;
            }
        }
        out.writeBoolean(supportSlowSync);
        if(version >= VERSION_101) {
            out.writeBoolean(supportMultipleEmailAccount);
        }
        Log.info(TAG_LOG, "writing config: smartslow set to " + supportSlowSync);
        Log.info(TAG_LOG, "writing config: mea set to " + supportMultipleEmailAccount);
    }

    private void deserialize(DataInputStream in) throws IOException {

        boolean supportSlowSync = in.readBoolean();
        boolean supportMultipleEmailAccount = false;

        if(version >= VERSION_101) {
            supportMultipleEmailAccount = in.readBoolean();
        }

        serverCapabilities = new DevInf();
        Vector extVector = new Vector();

        if (supportSlowSync) {
            Ext slow = new Ext(Ext.X_FUNAMBOL_SMARTSLOW, new String[0]);
            extVector.addElement(slow); 
        }
        if (supportMultipleEmailAccount) {
            Ext mea = new Ext(Ext.X_FUNAMBOL_MULTIPLE_EMAIL_ACCOUNT, new String[0]);
            extVector.addElement(mea);
        }
        if(extVector.size() > 0) {
            serverCapabilities.setExts(extVector);
        }
    }

    private int getStoredVersion(AbstractRecordStore rs) {
        if(version != -1) {
            return version;
        }
        try {
            byte[] data = rs.getRecord(1);
            DataInputStream dis = new DataInputStream(new ByteArrayInputStream(data));
            version = dis.readInt();
        } catch (Exception ex) {
            Log.error(TAG_LOG, "Error while reading stored version: " + ex);
            ex.printStackTrace();
        }
        return version;
    }

    private DataInputStream getStoredCapabilities(AbstractRecordStore rs) {
        try {
            byte[] data = rs.getRecord(2);
            return new DataInputStream(new ByteArrayInputStream(data));
        } catch (Exception ex) {
            Log.error(TAG_LOG, "Error while reading stored capabilities: " + ex);
            ex.printStackTrace();
            return null;
        }
    }
}
