/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.ui.view;

import com.funambol.mailclient.sm.SyncClient;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.config.*;
import com.funambol.mailclient.UserAccount;
import com.funambol.mailclient.MailAccount;
import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.ui.controller.ServerCapStore;
import com.funambol.syncml.protocol.DevInf;
import com.funambol.syncml.protocol.Ext;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Item;
import javax.microedition.lcdui.TextField;

/**
 * This form shows all the settings related to the
 * Account functions
 * The user will be able to configure username, password,
 * sync server url.
 * The form shows also the deviceId used internally to
 * synchronize
 */
public class AccountSettingsForm extends TestableForm implements CommandListener {

    private Command backCommand;
    private Command saveCommand;
    private TextField serverUrlTF;
    private TextField userNameTF;
    private TextField passwordTF;

    private TextField visibleNameTF;
    private TextField emailAddressTF;
    private MailClientConfig mailClientConfig;
    private boolean syncIsBusy;

    /**
     * Creates a new instance of AccountSettingsForm
     */
    public AccountSettingsForm() {
        super(Localization.getMessages().CONNECTION_SETTINGS_FORM_TITLE);

        mailClientConfig = UIController.getConfig().clone();

        //#ifdef isBlackberry
        //# saveCommand = new Command(Localization.getMessages().SAVE_COMMAND, Command.OK, 5);
        //# backCommand = new Command(Localization.getMessages().BACK_COMMAND, Command.BACK, 10);
        //# addCommand(UIController.blackberryExitCommand);
        //#else
        backCommand = UIController.backCommand;
        saveCommand = UIController.saveCommand;
        //#endif
        syncIsBusy = SyncClient.getSyncClient().isBusy();

        layout();
        
        addCommand(backCommand);
        if (!UIController.isSyncInProgress()) {
            addCommand(saveCommand);
        }
        setCommandListener(this);
    }

    public void layout() {
        deleteAll();

        append(getServerUrlTF());
        append(getUserNameTF());
        append(getPasswordTF());

        // Check if the server capabilities are already received
        ServerCapStore serverCap = ServerCapStore.getInstance();
        DevInf serverDevinf = serverCap.get();

        if(serverDevinf != null) {
            append(getVisibleNameTF());

            // If the server doesn't support multiple email accounts we shall
            // include the email address in the account settings
            if(!serverCap.isExtSupported(Ext.X_FUNAMBOL_MULTIPLE_EMAIL_ACCOUNT)) {
                append(getEmailAddressTF());
            }
        }
    }

    /**
     * 
     * used to find the item that should be focused.
     * if username or passowrd textfield are empty, that item is returned
     * otherwise server textfield is returned
     * 
     * @return the item that should be focused
     */
    public Item getFocusedItem() {
        if ("".equals(getUserNameTF().getString())) {
            Log.debug("returning username TF");
            return getUserNameTF();
        } else if ("".equals(getPasswordTF().getString())) {
            Log.debug("returning pwd TF");
            return getPasswordTF();
            
        } else {
            Log.debug("returning url TF");
            return getServerUrlTF();
        }
        
    }
    
    /**
     * command handling method
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command == backCommand) {
            UIController.showBackScreen();
            UIController.accountSettingsForm = null;
        } else if (command == saveCommand) {
            // save settings

            // Sanity checks: empty url or default and empty credentials
            //                are not valid.
            // FIXME: why can these be null??

            if (StringUtil.isNullOrEmpty(userNameTF.getString()) ||
                    StringUtil.isNullOrEmpty(serverUrlTF.getString()) ||
                    StringUtil.isNullOrEmpty(passwordTF.getString())) {
                Log.info("Empty Url or credentials field found: fill it!!");
                UIController.showAlert(
                        Localization.getMessages().CSF_EMPTY_CREDENTIALS_ERROR);
                return;
            } else if (emailAddressTF != null && !checkValidEmail(emailAddressTF.getString())) {
                Log.info("Email Address field needs valid email address value.");
                UIController.showAlert(
                       Localization.getMessages().CSF_INVALID_EMAIL_ERROR);
                return;
            }

            //before saving the account trim the fields
            trimFields();

            // Get the current account
            UserAccount userAccount = UIController.getConfig().getUserAccount();

            // If the account was already set
            if (userAccount.isSet()) {
                // If username or url are changed, we must alert the user
                // and reset the account.
                boolean userChanged =
                        !userNameTF.getString().equals(userAccount.getUser());
                boolean urlChanged =
                        !serverUrlTF.getString().equals(userAccount.getUrl());

                if (userChanged || urlChanged) {
                    mailClientConfig = getNewConfigFromUser();
                    UIController.saveNewAccountWithConfirmation(mailClientConfig);
                } else {
                    mailClientConfig = getNewConfigFromUser();
                    UIController.saveNewConfig(mailClientConfig);
                    UIController.showBackScreen();
                }
            } else {
                // if it's the first start, save the config and
                // start the initial sync.
                mailClientConfig = getNewConfigFromUser();
                UIController.saveNewConfig(mailClientConfig);
                UIController.showInboxScreen();
                UIController.setSyncCaller(UIController.USER);
                UIController.sync(true);
            }
        } else if (command == UIController.blackberryExitCommand) {
            UIController.midlet.destroyApp(false);
        }
    }

    private MailClientConfig getNewConfigFromUser() {
        UserAccount userAccount = new UserAccount();
        userAccount.setUser(userNameTF.getString());
        userAccount.setPassword(passwordTF.getString());
        userAccount.setUrl(serverUrlTF.getString());

        //Problem: It lose .jad "remoteURI" and "PIMRemoteURI" set parameters
        //when starting first time. The next two lines fix the problem.
        userAccount.setRemoteURI(UIController.getConfig().getUserAccount().getRemoteURI());
        userAccount.setPimRemoteURI(UIController.getConfig().getUserAccount().getPimRemoteURI());

        mailClientConfig.setUserAccount(userAccount);

        if(visibleNameTF != null) {
            String vn = visibleNameTF.getString();
            if(emailAddressTF != null) {
                MailAccount mAccount = new MailAccount(vn, emailAddressTF.getString());
                mailClientConfig.setMailAccount(mAccount, true);
            } else {
                mailClientConfig.setVisibleName(vn);
            }
        }
        Log.debug("[AccountSettingForm.getNewConfigFromUser] New Account is:\n" + userAccount.toString());
        return mailClientConfig;

    }

    private TextField getServerUrlTF() {
        if (serverUrlTF == null) {
            serverUrlTF = new TextField(Localization.getMessages().CSF_SERVER_URL_TF_LABEL,
                    mailClientConfig.getUserAccount().getUrl(), 120, TextField.ANY | TextField.NON_PREDICTIVE);
            serverUrlTF.setInitialInputMode("MIDP_LOWERCASE_LATIN");
        }

        setUneditableTextField(serverUrlTF);
        return serverUrlTF;
    }

    private TextField getUserNameTF() {
        if (userNameTF == null) {

	        String username = mailClientConfig.getUserAccount().getUser();
	        username = username==null?"":username;
        
            userNameTF = new TextField(Localization.getMessages().CSF_USERNAME_TF_LABEL,
                    username, 120, TextField.ANY | TextField.NON_PREDICTIVE);
            userNameTF.setInitialInputMode("MIDP_LOWERCASE_LATIN");

        }
        userNameTF.setLayout(TextField.LAYOUT_EXPAND|TextField.LAYOUT_2|TextField.LAYOUT_NEWLINE_AFTER);
        setUneditableTextField(userNameTF);
        return userNameTF;
    }

    private TextField getPasswordTF() {
        if (passwordTF == null) {
            passwordTF = new TextField(Localization.getMessages().CSF_PASSWORD_TF_LABEL,
                    mailClientConfig.getUserAccount().getPassword(), 120, TextField.PASSWORD);
            passwordTF.setInitialInputMode("MIDP_LOWERCASE_LATIN");
        }
        passwordTF.setLayout(TextField.LAYOUT_EXPAND|TextField.LAYOUT_2);
        setUneditableTextField(passwordTF);
        return passwordTF;
    }

    private TextField getVisibleNameTF() {
        if (visibleNameTF == null) {
            visibleNameTF =
                    new TextField(Localization.getMessages().CSF_VISIBLE_NAME_TF_LABEL,
                    mailClientConfig.getMailAccount(0).getVisibleName(), 120, TextField.ANY | TextField.NON_PREDICTIVE);
            visibleNameTF.setInitialInputMode("MIDP_LOWERCASE_LATIN");
        }
        setUneditableTextField(visibleNameTF);
        return visibleNameTF;
    }

    private TextField getEmailAddressTF() {
        if (emailAddressTF == null) {
            emailAddressTF =
                   new TextField(Localization.getMessages().CSF_EMAIL_ADDRESS_TF_LABEL,
                   mailClientConfig.getMailAccount(0).getEmailAddress(), 120, TextField.ANY | TextField.NON_PREDICTIVE);
            emailAddressTF.setInitialInputMode("MIDP_LOWERCASE_LATIN");
        }
        setUneditableTextField(emailAddressTF);
        return emailAddressTF;
    }
    
    
    //This method disable editing of TextFields when
    //a sync process was started
    private void setUneditableTextField(TextField textField) {
        //#ifndef EditTextField
        if (syncIsBusy) {
            textField.setConstraints(TextField.UNEDITABLE | textField.getConstraints());
        }
    //#endif
    }

    private boolean checkValidEmail(String emailTF) {
        if ((emailTF == null) || (emailTF.equals(""))) {
            return true;
        } else {
            int position=emailTF.indexOf("@");
            if (position == -1 || position == 0 || position == emailTF.length() - 1) {
                return false;
            }
        }
        return true;
    }

    // trim fields removing useless and wrong spaces at the beginning and at the end
    private void trimFields() {
        TextField tempTF = null;
        for (int i = 0; i < this.size(); i++) {
            tempTF = (TextField) this.get(i);
            tempTF.setString(tempTF.getString().trim());
        }
    }

    /**
     * enable or disable the send command
     * @param enable if true enable the command, if false disable the command
     */
    public void enableSaveCommand(boolean enable) {
        Log.debug(this, "Enable save commands: " + enable);
        if (enable) {
            addCommand(saveCommand);
        } else {
            removeCommand(saveCommand);
        }
    }
}
