/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.ui.view;

import java.util.Vector;
import java.util.TimerTask;
import java.util.Timer;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.Font;

import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.loc.LocalizedMessages;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.controller.PopupAction;
import com.funambol.mailclient.ui.view.components.ScrollBar;
import com.funambol.mailclient.ui.utils.UiUtils;
import com.funambol.util.StringUtil;
import com.funambol.util.Log;


/**
 * This class represents an alert on the screen. An alert can serve different
 * purposes, from informing the user about errors to asking the user permission
 * for certain operations.
 * These alerts are drawn on a low level graphical component so that the
 * application has full control on the look and feel of the component itself.
 *
 * Each alert can be configured in several ways. It can be configured for
 * autodismiss after a certain amount of time has passed without user click.
 *
 * Another option is to show the alert as modal. In this case the call to show
 * is blocking and the control returns back to the caller once the user dismiss
 * the alert or the timeout expires.
 *
 * It is also possible to register a listener for the alert. When the user
 * clicks on one of the available option, then the listener is invoked so that
 * it can perform the necessary actions.
 *
 * It is possible to supply a text to be shown in the alert and a long description.
 * When the long description is provided, there is a menu entry for the details.
 * The details open into a new screen where it is possible to have very long messages.
 *
 * The alert handles the problem of overlapping alerts. By default it is not
 * possible to stack multiple alerts on the screen. Unless the user forces a
 * different behavior. By default when an alert is shown on the screen and a new
 * request comes to show a new one, the second one is discarded and the new
 * alert is not shown.
 *
 */ 
public class AlertScreen extends CustomMenuCanvas implements TestableScreen, CommandListener {

    public static final int FOREVER = 0; 

    public static final int CONFIRMATION_YES_NO = 0;
    public static final int INFORMATION_OK      = 1;
    public static final int BAR_HEIGHT          = 15;

    private static final int TITLE_SEPARATION = 5;
    private static final int TITLE_LINES_SEPARATION = 2;
    private static final int TEXT_BORDER_DISTANCE = 10;

    private static final int BAR_RED_COLOR = 0xe78b00;
    private static final int BAR_BLACK_COLOR = 0x0;
    private static final int ALERT_BACKGROUND_COLOR = 0x00333333;
    private static final int TITLE_COLOR = 0xC9C9C9;
    private static final int MESSAGE_COLOR = 0xC9C9C9;
    private static final int FRAME_COLOR_1 = 0x696969;
    private static final int FRAME_COLOR_2 = 0x999999;
    private static final int EMPTY_BACKGROUND_COLOR = 0xFFFFFF;

    private static final String TAG_LOG = "AlertScreen";

    private static Image bar = null;

    private static boolean isAnAlertShown = false;
    private static boolean showSingleAlert = true;

    private Image backgroundImage;
    private String shortMessage;
    private String longMessage;
    private String title;
    private Drawer drawer;
    private GraphicalTheme theme;
    private Font dialogFont;
    private Font titleFont;
    private String[] shortLines;
    private String[] titleLines;
    private int dialogHeight, dialogWidth;
    private int startingX, startingY;
    private Displayable nextScreen;
    private int type;
    private int textWidth;
    private PopupAction popupAction;
    private int screenWidth, screenHeight;
    private boolean modal;
    private int timeout;
    private boolean isShown;
    private Timer timer;
    private boolean wipeBackground = false;
    private String yesLabel, noLabel, okLabel;
    private boolean commandsAdded;
    private boolean scrollDownEnabled = false;
    private int firstVisible = 0;
    private boolean scrollBarVisible = false;
    private ScrollBar scrollBar;

    private Command yesCommand, noCommand, detailsCommand;
    private Command okCommand;

    private Object lock = new Object();

    /**
     * Constructs an AlertScreen.
     * @param backgroundImage the image to be placed in the screen background
     * @param type alert type (e.g. INFORMATION_OK, CONFIRMATION_YES_NO)
     * @param title the alert title
     * @param shortMessage the message to show in the alert
     * @param longMessage the message to show in the details (if null no details
     * are made available)
     * @param modal true for modal alert (the show method blocks execution)
     * @param timeout the autodismiss timeout (if it is set to FOREVER, the
     * alert does not autodismiss)
     * @param nextScreen the next screen to show once the alert is closed. If
     * this value is null, the alert shows the previous screen on its closure
     * (relying on the UIController back stack)
     * @param action the action to be performed when the user clicks a button
     * (can be null)
     */
    public AlertScreen(Image backgroundImage, int type, String title,
                       String shortMessage, String longMessage,
                       boolean modal, int timeout, Displayable nextScreen,
                       PopupAction action)
    {
        super();

        this.backgroundImage = backgroundImage;
        this.shortMessage = shortMessage;
        this.longMessage = longMessage;
        this.title       = title;
        this.nextScreen  = nextScreen;
        this.type        = type;
        this.popupAction = action;
        this.modal       = modal;

        // If the alert requires user choice, we do not allow the autodismiss
        if (type != CONFIRMATION_YES_NO) {
            this.timeout = timeout;
        } else {
            this.timeout = -1;
        }

        if (backgroundImage != null) {
            this.backgroundImage = blurImage(backgroundImage);
        }

        drawer = UIController.getDrawer();
        theme = drawer.getGraphicalTheme();
        dialogFont = theme.getDefaultFont();
        titleFont  = theme.getBoldFont();

        yesLabel = Localization.getMessages().POPUP_YES;
        noLabel  = Localization.getMessages().POPUP_NO;
        okLabel  = Localization.getMessages().GENERIC_OK;

        commandsAdded = false;

        init();
    }

    private void addCommands() {
        if (!commandsAdded) {
            // Add the proper commands, depending on the alert type
            if (type == CONFIRMATION_YES_NO) {
                int yesPriority;
                int noPriority;
//#ifdef isBlackberry
//#             yesPriority = 0;
//#             noPriority  = 1;
//#else
                yesPriority = 1;
                noPriority  = 0;
//#endif
                yesCommand = new Command(yesLabel, UIController.COMMAND_TYPE, yesPriority);
                noCommand  = new Command(noLabel, UIController.COMMAND_TYPE, noPriority);
                addCommand(yesCommand);
                addCommand(noCommand);
            } else if (type == INFORMATION_OK) {
                okCommand = new Command(okLabel, UIController.COMMAND_TYPE, 0);
                // For the BB we do not add a command for this and allow the
                // user to quit the alert with the back key
                addCommand(okCommand);
            }

            if (longMessage != null) {
                String label = Localization.getMessages().DETAILS_COMMAND_LABEL;
                detailsCommand = new Command(label, UIController.COMMAND_TYPE, 2);
                addCommand(detailsCommand);
            }
            commandsAdded = true;
        }
    }

    public void showSingleAlert(boolean value) {
        showSingleAlert = value;
    }

    public void show() {

        // If an alert is already shown, then we do not show anything else, and
        // just return (this is true if the alert screen is in single alert
        // mode)
        if (isAnAlertShown && showSingleAlert) {
            return;
        }

        // Add the commands if they are not there yet
        addCommands();

        isShown = true;
        isAnAlertShown = true;
        // Start a timer if necessary
        if (timeout > 0) {
            timer = new Timer();
            AlertTimer alertTimer = new AlertTimer();
            timer.schedule(alertTimer, timeout);
        } else {
            timer = null;
        }
        // Show this screen
        if (nextScreen == null) {
            UIController.showScreenWithBackStack(this);
        } else {
            UIController.showScreenNoBackStack(this);
        }
        // Block execution for modal dialogs
        if (modal) {
            // Now wait for the user answer
            synchronized(lock) {
                try {
                    lock.wait();
                } catch (InterruptedException ie) {
                    Log.error(TAG_LOG, "Modal interrupted");
                }
            }
        }
    }

    public void setYesLabel(String label) {
        if (label != null) {
            yesLabel = label;
        }
    }

    public void setNoLabel(String label) {
        if (label != null) {
            noLabel = label;
        }
    }

    public void setOkLabel(String label) {
        if (label != null) {
            okLabel = label;
        }
    }

    private void init() {
        // Compute coordinates one for all
        if (screenWidth != getWidth() || screenHeight != getHeight()) {
            screenWidth = getWidth();
            screenHeight = getHeight();
        } else {
            return;
        }


        // We occupy 75% of the screen width
        dialogWidth = (screenWidth * 75) / 100;
        textWidth = (dialogWidth * 80) / 100;

        if (title != null) {
            titleLines = UiUtils.getStringArray(title, textWidth - 4, titleFont);
        }

        if (shortMessage != null) {
            shortLines = UiUtils.getStringArray(shortMessage, textWidth - 4, dialogFont);
        } else {
            shortLines = new String[0];
        }

        int dialogMaxHeight = (screenHeight * 80 / 100);
        dialogHeight = shortLines.length * dialogFont.getHeight();
        dialogHeight += TEXT_BORDER_DISTANCE * 2;
        if (titleLines != null) {
            dialogHeight += titleLines.length * titleFont.getHeight() + TITLE_SEPARATION;
            if (titleLines.length > 1) {
                dialogHeight += ((titleLines.length - 1) * TITLE_LINES_SEPARATION);
            }
        }
        // Reserve space for the gradient bar
        dialogHeight += dialogFont.getHeight() + 3;

        if (dialogHeight > dialogMaxHeight) {
            // TODO what shall we do here?
            dialogHeight = dialogMaxHeight;
        }

        // Compute the x and y coordinates
        startingX = (screenWidth - dialogWidth) / 2;
        startingY = (screenHeight - dialogHeight) / 2;

        // Prepare the bar if necessary (the bar is static and kept across
        // class instances
        int barWidth = dialogWidth - 6;
        if (bar == null || bar.getWidth() != barWidth) {
            bar = initBar(barWidth, BAR_HEIGHT);
        }
    }

    private static Image initBar(int width, int height) {
        Image res = Image.createImage(width, height);
        Graphics g = res.getGraphics();

        for(int r=0;r<height;++r) {
            int red = BAR_RED_COLOR;
            int black = BAR_BLACK_COLOR;
            for(int c=0;c<width;c+=5) {
                if (c % 10 == 0) {
                    g.setColor(red);
                } else {
                    g.setColor(black);
                }

                g.drawLine(c,r,c+5,r);
            }
        }

        return rotateAndApplyGrad(res);
    }

    protected void paintContent(Graphics graphics) {

        init();

        // If we have a background image, we first display it
        // The snapshot was taken on a different screen size,
        // we cannot reuse it here, unless we decide to resize it, but it is
        // not worth doing it
        if (backgroundImage != null) {
            if (backgroundImage.getWidth() == screenWidth) {
                graphics.drawImage(backgroundImage, 0, 0, Graphics.TOP | Graphics.LEFT);
            } else {
                // Clear the background to avoid using it dirty
                graphics.setColor(EMPTY_BACKGROUND_COLOR);
                graphics.fillRect(0, 0, getWidth(), getHeight());
            }
        } else {
            if (wipeBackground) {
                graphics.setColor(EMPTY_BACKGROUND_COLOR);
                graphics.fillRect(0, 0, getWidth(), getHeight());
            }
        }

        // Now paint the dialog
        // Start from the background
        graphics.setColor(ALERT_BACKGROUND_COLOR);
        graphics.fillRoundRect(startingX, startingY, dialogWidth, dialogHeight, 10, 10);
        // Now the surrounding frame
        graphics.setColor(FRAME_COLOR_1);
        graphics.drawRoundRect(startingX + 1, startingY + 1, dialogWidth -3 , dialogHeight -3, 10, 10);
        graphics.setColor(FRAME_COLOR_2);
        graphics.drawRoundRect(startingX + 2, startingY + 2, dialogWidth -5 , dialogHeight -5, 10, 10);
        // Draw the top bar
        graphics.drawImage(bar, startingX + 3, startingY + 3, Graphics.TOP | Graphics.LEFT);
        int currentY = startingY + bar.getHeight() + 3;

        // Draw the title
        graphics.setFont(titleFont);
        graphics.setColor(TITLE_COLOR);
        currentY += TEXT_BORDER_DISTANCE;
        if (titleLines != null) {
            for(int j=0;j<titleLines.length;++j) {
                currentY += titleFont.getHeight();
                String titleLine = titleLines[j];
                int titleXOffset = (dialogWidth - titleFont.stringWidth(titleLine)) / 2;
                titleXOffset += startingX;
                graphics.drawString(titleLine, titleXOffset, currentY,
                        Graphics.BOTTOM | Graphics.LEFT);
                if (j != titleLines.length - 1) {
                    currentY += TITLE_LINES_SEPARATION;
                }
            }
            currentY += TITLE_SEPARATION;
        }
        graphics.setFont(dialogFont);
        // Now draw the message
        graphics.setColor(MESSAGE_COLOR);
        scrollDownEnabled = false;
        int startingYForMessage = currentY;
        int numVisibles = 0;
        int lastMessageY = 0;
        for(int i=firstVisible;i<shortLines.length;++i,++numVisibles) {
            // Each message is centered
            int x = startingX + ((dialogWidth - dialogFont.stringWidth(shortLines[i])) / 2);
            currentY += dialogFont.getHeight();
            // Check if we are out of vertical space (we subtract 3 to take into
            // account the space reserved for the frame)
            if (currentY >= startingY + dialogHeight - 3) {
                scrollDownEnabled = true;
                // If we need a scroll bar, then we get ready to draw it
                scrollBarVisible  = true;
                break;
            }
            graphics.drawString(shortLines[i], x, currentY,
                                Graphics.BOTTOM | Graphics.LEFT);
            lastMessageY = currentY;
        }

        if (scrollBarVisible) {
            if (scrollBar == null) {
                int x0 = startingX + dialogWidth - 12;
                int y0 = startingYForMessage;
                scrollBar = new ScrollBar(x0,y0,6, lastMessageY - y0 - 2,
                                          shortLines.length,
                                          numVisibles);
                scrollBar.setVisible(true);
            } else {
                scrollBar.updateBarPosition(firstVisible);
            }
            scrollBar.paint(graphics);
        }
    }

    protected void handleKeyPressed(int keyCode) {
        int dir = getGameAction(keyCode);
        // todo: check if we can filter out keypad numbers
        // by checking if keycode is > 0.
        if (isDownKey(dir, keyCode)) {
            scrollDown(1);
        } else if (isUpKey(dir, keyCode)) {
            scrollUp(1);
        } else if (keyCode == -7) {
//#ifdef isBlackberry
//#         handleBackKeyBB();
//#endif
        } else if (isFireKey(dir, keyCode)) {
//#ifdef isBlackberry
//#         openMainMenu();
//#endif
        }
    }

    private void handleBackKeyBB() {
        if (type == INFORMATION_OK) {
            // On BB the back command is not associated to any menu entry, but it
            // behaves like the "OK" if the alert is just an info screen
            commandAction(okCommand, this);
        } else if (type == CONFIRMATION_YES_NO) {
            commandAction(noCommand, this);
        }
    }

    private void scrollDown(int amount) {
        if (scrollDownEnabled) {
            firstVisible++;
            repaint();
        }
    }

    private void scrollUp(int amount) {
        if (firstVisible > 0) {
            firstVisible--;
            repaint();
        }
    }

    protected void handleKeyRepeated(int keyCode) {
        handleKeyPressed(keyCode);
    }

    protected void handleTouchEvent(FTouchEvent event) {
        int type     = event.getType();
        int tapCount = event.getTapCount();

        switch (type) {
            case FTouchEvent.DRAGGING_DOWN:
                scrollDown(1);
                break;
            case FTouchEvent.DRAGGING_UP:
                scrollUp(1);
                break;
        }
    }

    protected void openMainMenu() {
//#ifdef isBlackberry
//#         wipeBackground = true;
//#endif
        super.openMainMenu();
    }

    public void commandAction(Command command, Displayable displayable) {

        if (command == yesCommand || command == okCommand) {
            if (popupAction != null) {
                popupAction.confirm();
            }
            closeAlert();
        } else if (command == noCommand) {
            if (popupAction != null) {
                popupAction.cancel();
            }
            closeAlert();
        } else if (command == detailsCommand) {
            // When we show the details we need to clean the underlying
            // graphics, because otherwise on the back action we may leave the
            // details in the background (if the original screen was not
            // alertable)
            UIController.showAlertDetails(longMessage);
            wipeBackground = true;
        } else {
            super.commandAction(command, displayable);
        }
    }

    private void closeAlert() {
        isShown = false;
        // If we have a timer, then we cancel it
        if (timer != null) {
            timer.cancel();
        }
        // If the alert is modal, then we unlock it and let the execution
        // continue
        if (modal) {
            synchronized(lock) {
                lock.notify();
            }
        }
        if (popupAction == null) {
            if (nextScreen == null) {
                UIController.showBackScreen();
            } else {
                UIController.display.setCurrent(nextScreen);
            }
        }
        isAnAlertShown = false;
    }

    private Image blurImage(Image image) {
        Log.debug(TAG_LOG, "Blurring image");
        try {
            int rgbData[] = new int[image.getWidth() * image.getHeight()];
            image.getRGB(rgbData, 0, image.getWidth(), 0, 0, image.getWidth(), image.getHeight());
            for(int i=0;i<rgbData.length;++i) {
                int value = rgbData[i];
                int r = (value & 0x00FF0000) >> 16;
                int g = (value & 0x0000FF00) >> 8;
                int b = value & 0x000000FF;

                r -= 64;
                g -= 64;
                b -= 64;
                if (r < 0) {
                    r = 0;
                }
                if (g < 0) {
                    g = 0;
                }
                if (b < 0) {
                    b = 0;
                }
                value = 0XFF000000 | (r << 16) | (g << 8) |b;
                //value = value & 0xFFC8C8C8;
                rgbData[i] = value;
            }
            image = Image.createRGBImage(rgbData, image.getWidth(), image.getHeight(), true);
        } catch (Exception e) {
            Log.error(TAG_LOG, "Cannot blur image " + e.toString());
        }
        return image;
    }

    public static Image rotateAndApplyGrad(Image src) {
        int sw = src.getWidth();
        int sh = src.getHeight();
        int[] srcData = new int[sw * sh];
        src.getRGB(srcData, 0, sw, 0, 0, sw, sh);
        int[] dstData = new int[sw * sh];
        //double rads = angle * Math.PI / 180.f;
        //float sa = (float) Math.sin(rads);
        //float ca = (float) Math.cos(rads);
        float sa = (float) 0.71f;
        float ca = (float) 0.71f;
        int isa = (int) (256 * sa);
        int ica = (int) (256 * ca);

        int my = - (sh >> 1);
        for(int i = 0; i < sh; i++) {
            int wpos = i * sw;
            int xacc = my * isa - (sw >> 1) * ica + ((sw >> 1) << 8);
            int yacc = my * ica + (sw >> 1) * isa + ((sh >> 1) << 8);
            for(int j = 0; j < sw; j++) {
                int srcx = (xacc >> 8);
                int srcy = (yacc >> 8);
                if(srcx < 0) srcx = 0;
                if(srcy < 0) srcy = 0;
                if(srcx > sw - 1) srcx = sw - 1;
                if(srcy > sh - 1) srcy = sh - 1;
                dstData[wpos++] = srcData[srcx + srcy * sw];
                xacc += ica;
                yacc -= isa;
            }
            my++;
        }

        // Apply the gradient on the output image
        /*
        int[] reddishColors = UiUtils.getGradient(0xffb84d, 0xe78b00, sh);
        int[] blackishColors = UiUtils.getGradient(0xc9c9c9, 0x000000, sh);
        for(int i=0;i<sh;i++) {
            int newBlack = blackishColors[i];
            int newRed   = reddishColors[i];
            for(int j=0;j<sw;j++) {
                int pixel = dstData[i*sw+j];
                if ((pixel & 0xFFFFFF) == 0x0 ) {
                    // This is a black one
                    dstData[i*sw+j] = newBlack;
                } else {
                    dstData[i*sw+j] = newRed;
                }
            }
        }
        */

        for(int i=0;i<sh / 2;i++) {
            int newBlack = 0x999999;
            int newRed   = 0xffb84d;
            for(int j=0;j<sw;j++) {
                int pixel = dstData[i*sw+j];
                if ((pixel & 0xFFFFFF) == 0x0 ) {
                    // This is a black one
                    dstData[i*sw+j] = newBlack;
                } else {
                    dstData[i*sw+j] = newRed;
                }
            }
        }


        Image img = Image.createRGBImage(dstData, sw, sh, false);
        return img;
    }

    public String getTitle() {
        return title;
    }

    public String getMessage() {
        return shortMessage;
    }

    protected class AlertTimer extends TimerTask {

        public AlertTimer() {
        }

        public void run() {
            if (isShown) {
                closeAlert();
            }
        }
    }
}
 
