/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


package com.funambol.mailclient.ui.view;

import com.funambol.mail.Address;
import com.funambol.mail.MailException;
import com.funambol.mailclient.cm.Contact;
import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.utils.UiUtils;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;

/**
 * A class that describes a Contact item object to be painted as List element.
 * This class provides methods for handling the state of the contact
 * (unselected, selected as TO, CC, BCC), for managing the selected email address
 * (default, secondary or tertiary address), for creating an Address from the
 * ContactItem and for painting the contactitem on a given graphics object
 */
public class ContactItem {

    private static final String TAG_LOG = "ContactItem";

    /**
     * ADDRESS_STATE_NONE is composed by Address.TO|Address.CC|Address.BCC and shifted.
     * this way we're sure that ADDRESS_STATE_NONE is different from each of Address.TO,
     * Address.CC, Address.BCC
     */
    public static final int ADDRESS_STATE_NONE =( Address.TO | Address.CC | Address.BCC ) <<1;

    /** Items margins */
    private final int MARGIN = 3;
    private final int LABEL_MARGIN_LEFT  = 5;
    private final int LABEL_MARGIN_RIGHT = 5;
    private final int CONTACT_INFO_MARGIN_LEFT = 4;

    /**
     * Current state of the contact. one of ContactItem.ADDRESS_STATE_NONE, Address.TO,
     * Address.CC, Address.BCC
     */
    private int state;

    /** The contact */
    private Contact contact;

    /**
     * The selected email index.
     * one of Contact.DEFAULT_EMAIL, Contact.SECONDARY_EMAIL,
     * Contact.TERTIARY_EMAIL
     */
    private int selectedEmailIndex;

    /** The height of this contact items */
    private int height;

    /** The drawer used for painting */
    private Drawer drawer;

    /** Standard font */
    private Font font;

    /** Bold font */
    private Font boldFont;

    /** Max label width in pixel */
    private int maxLabelWidth;

    private boolean lastActive;

    /**
     * Creates a new instance of ContactItem with the given contact.
     */
    public ContactItem(Contact c) {
        this(c, ADDRESS_STATE_NONE);
    }

    /**
     * Creates a new instance of ContactItem with the given contact and state.
     */
    public ContactItem(Contact c, int state) {

        this(c, state, Contact.DEFAULT_EMAIL);
        toFirstEmail();
    }

    /**
     * Create a new ContactItem with given data
     * @param contact the contact
     * @param state one of Address.TO Address.CC Address.BCC or
     * FunCanvasContactList.ADDRESS_STATE_NONE
     * @param selectedEmail one of Contact.DEFAULT, Contact.SECONDARY_EMAIL
     * or Contact.TERTIARY_EMAIL
     */
    public ContactItem(Contact c, int state, int selectedEmail) {
        this.contact = c;
        this.state = state;
        this.selectedEmailIndex = selectedEmail;
        this.drawer = UIController.getDrawer();
        this.font = drawer.getGraphicalTheme().getDefaultFont();
        this.boldFont = drawer.getGraphicalTheme().getBoldFont();
    }

    /**
     * Change the state of the item the the next one. <br>
     * sequence is ADDRESS_STATE_NONE -> TO -> CC -> BCC -> ADDRESS_STATE_NONE
     *
     * @return the state at the end of the method call, as one of
     * <ul>
     *  <li>ADDRESS_STATE_NONE</li>
     *  <li>Address.TO</li>
     *  <li>Address.CC</li>
     *  <li>Address.BCC</li>
     * </ul>
     */
    public int toNextState() {

        switch (state) {
            case ADDRESS_STATE_NONE:
                state=Address.TO;
                break;
            case Address.TO:
                state=Address.CC;
                break;
            case Address.CC:
                state=Address.BCC;
                break;
            case Address.BCC:
                state=ADDRESS_STATE_NONE;
                break;
            default:
                state=ADDRESS_STATE_NONE;
        }
        return state;
    }

    /**
     * Shift to the first available email
     */
    public int toFirstEmail() {
        if (! StringUtil.isNullOrEmpty( contact.getDefaultEmail()) ) {
            selectedEmailIndex = Contact.DEFAULT_EMAIL;
        } else if (! StringUtil.isNullOrEmpty( contact.getEmail_2()) ) {
            selectedEmailIndex = Contact.SECONDARY_EMAIL;
        } else if (! StringUtil.isNullOrEmpty( contact.getEmail_3()) ) {
            selectedEmailIndex = Contact.TERTIARY_EMAIL;
        }
        return selectedEmailIndex;
    }

    /**
     * Cycle to the next email changing the value of selectedImailIndex.
     */
    public void toNextEmail() {
        switch (selectedEmailIndex) {
            case Contact.DEFAULT_EMAIL:
                selectedEmailIndex = Contact.SECONDARY_EMAIL;
                break;
            case Contact.SECONDARY_EMAIL:
                selectedEmailIndex = Contact.TERTIARY_EMAIL;
                break;
            case Contact.TERTIARY_EMAIL:
            default: 
                selectedEmailIndex = Contact.DEFAULT_EMAIL;
                break;
        }
        if (StringUtil.isNullOrEmpty( getSelectedEmail())) {
            toNextEmail();
        }
    }

    /**
     * Cycle to the previous email changing the value of selectedImailIndex.
     */
    public void toPreviousEmail() {
        switch (selectedEmailIndex) {
            case Contact.DEFAULT_EMAIL:
                selectedEmailIndex = Contact.TERTIARY_EMAIL;
                break;
            case Contact.TERTIARY_EMAIL:
                selectedEmailIndex = Contact.SECONDARY_EMAIL;
                break;
            case Contact.SECONDARY_EMAIL:
            default:
                selectedEmailIndex = Contact.DEFAULT_EMAIL;
                break;
        }
        if (StringUtil.isNullOrEmpty( getSelectedEmail())) {
            toPreviousEmail();
        }
    }

    /**
     * @return the contact in use.
     */
    public Contact getContact() {
        return contact;
    }

    /**
     * @return the address or null if no email is selected
     */
    public Address getAddress() {
        if (state==ADDRESS_STATE_NONE)
            return null;
        else {
            try {
                return contact.getAddress(state, selectedEmailIndex);
            } catch (MailException ex) {
                //TODO: better exception handling and localization
                Log.error(TAG_LOG, "Error getting Address from contact item");
                ex.printStackTrace();
                return null;
            }
        }
    }

    /**
     * @return the selected email as a string
     */
    private String getSelectedEmail() {
        switch (selectedEmailIndex) {
            case Contact.SECONDARY_EMAIL:
                return contact.getEmail_2();
            case Contact.TERTIARY_EMAIL:
                return contact.getEmail_3();
            case Contact.DEFAULT_EMAIL:
            default:
                return contact.getDefaultEmail();
        }
    }

    /**
     * Paint the ContactItem on the current graphics starting at 0,0
     * so a translate operation is often needed to correctly align the
     * graphics object
     * <br>
     * this method draws the contactItem as an item with width equals to
     * graphics.getClipWidth().
     *
     * @return the height of the painted area, this should be the same as
     * calculateheight()
     *
     * @param graphics the Graphics object to paint on.
     * @param elementId The current element id.
     * @param isActive the current active / non active state of the element
     *
     */
    public int paint(Graphics graphics, int elementId, boolean isActive) {

        // Calculate the height of the item
        int h = calculateHeight(isActive);
        int w = graphics.getClipWidth();

        int original_h = graphics.getClipHeight();

        // Computes the contact information position.
        int textNameH = 0;
        int textAddressH = 0;
        if (isActive) {
            textNameH    =   h/4 - boldFont.getHeight()/2 + MARGIN;
            textAddressH = 3*h/4 - boldFont.getHeight()/2;
        } else {
            textNameH = h/2 - font.getHeight()/2;
        }

        //
        // Draw the background
        //
        int grayedRegionWidth = LABEL_MARGIN_LEFT + getMaxLabelWidth() + LABEL_MARGIN_RIGHT;
        if (isActive) {
            UIController.getDrawer().drawHighlightedCustomItemSquaredBackground(graphics, 0, 0, w, h);
        } else {
            UIController.getDrawer().drawBackground(graphics, 0, 0, w, h, 
                    (elementId+1)%2, false, grayedRegionWidth, -9);
        }

        //
        // Draw the arrows icon
        //
        int arrowsWidth = 0;
        if ( hasMultipleEmails() ) {

            int arrowHeight  = calculateHeight(true)/4;
            int arrowWidth   = arrowHeight/2;
            int arrowSpace   = arrowHeight/2-1;
            int arrowsMargin = arrowWidth+MARGIN;
            arrowsWidth = arrowWidth + arrowSpace + arrowWidth + arrowsMargin;

            graphics.translate(w-arrowsWidth, h/2);
            drawSwitchEmailIcon(graphics, arrowWidth, arrowSpace);
            graphics.translate(-w+arrowsWidth, -h/2);
        }

        //
        // Draw the item label (TO: CC: BCC:)
        //
        drawLabel(graphics, h);

        //
        // Draw display name and mail address if active
        //
        graphics.setClip(grayedRegionWidth+CONTACT_INFO_MARGIN_LEFT, 0,
                         w-(grayedRegionWidth+CONTACT_INFO_MARGIN_LEFT), h);
        graphics.translate(grayedRegionWidth+CONTACT_INFO_MARGIN_LEFT, 0);
        if (isActive) {
            drawer.drawBoldString(UiUtils.cutString(contact.getVisibleName(),
                                  graphics.getClipWidth()-arrowsWidth, boldFont),
                                  graphics, 0, textNameH, isActive);
            drawer.drawString(UiUtils.cutString(getSelectedEmail(),
                              graphics.getClipWidth()-arrowsWidth, font),
                              graphics, 0, textAddressH, isActive);
        } else {
            drawer.drawString(UiUtils.cutString(contact.getVisibleName(),
                              graphics.getClipWidth()-arrowsWidth, font),
                              graphics, 0, textNameH, isActive);
        }
        graphics.translate(-(grayedRegionWidth+CONTACT_INFO_MARGIN_LEFT), 0);
        graphics.setClip(0, 0, w, original_h);

        return h;
    }

    /**
     * Compute the height of the item
     */
    private int calculateHeight(boolean isActive) {
        if (height==0 || lastActive != isActive) {
            lastActive = isActive;
            height = MARGIN*2;
            height += Math.max(font.getHeight(), boldFont.getHeight());
            if ( isActive  ) {
                height += boldFont.getHeight();
                height += 2*MARGIN;
            }
        }
        return height;
    }

    /**
     * Compute max label width for elements
     */
    private int getMaxLabelWidth() {
        if ( maxLabelWidth == 0) {
            Font labelFont = UIController.getDrawer().getGraphicalTheme().getLabelFont();
            int w1 = labelFont.stringWidth(Localization.getMessages().TO_LABEL);
            int w2 = labelFont.stringWidth(Localization.getMessages().CC_LABEL);
            int w3 = labelFont.stringWidth(Localization.getMessages().BCC_LABEL);
            maxLabelWidth = Math.max( w1, Math.max(w2, w3));

            // Adding padding width
            maxLabelWidth += labelFont.stringWidth(" ")*2;
        }
        return maxLabelWidth;
    }

    /**
     * Draw the switch email icons (arrows)
     */
    private void drawSwitchEmailIcon(Graphics graphics, int width, int space) {

        final int ARROW_SIZE   = width;
        final int ARROW_MARGIN = space;

        if ( hasMultipleEmails() ) {
            // Set color black for the arrows
            graphics.setColor(0, 0, 0);
            
            graphics.fillTriangle( 0, 0, ARROW_SIZE, ARROW_SIZE,
                                         ARROW_SIZE, -ARROW_SIZE);
            
            graphics.fillTriangle( 2*ARROW_SIZE+ARROW_MARGIN, 0,
                                   ARROW_SIZE+ARROW_MARGIN, ARROW_SIZE,
                                   ARROW_SIZE+ARROW_MARGIN, -ARROW_SIZE);

            // Set color white for the shadow
            graphics.setColor(255, 255, 255);
            graphics.drawLine( 0, 1, ARROW_SIZE-1, ARROW_SIZE);
            graphics.drawLine( ARROW_SIZE, ARROW_SIZE-1,
                               ARROW_SIZE, -ARROW_SIZE+1);
            graphics.drawLine( ARROW_SIZE+ARROW_MARGIN, ARROW_SIZE,
                               2*ARROW_SIZE+ARROW_MARGIN-1, 1);
        }
    }

    /**
     * Draws the label ("to:", "cc:","bcc:")
     */
    private void drawLabel(Graphics g, int h) {

        String label = getAddressTypeLabel();
        if(StringUtil.isNullOrEmpty(label)) {
            return;
        }
        
        Drawer d          = UIController.getDrawer();
        GraphicalTheme gt = d.getGraphicalTheme();
        Font labelFont    = gt.getLabelFont();

        // Define the label sizes
        int labelTextHeight = labelFont.getHeight();
        int labelPaddingX   = labelFont.stringWidth(" ");
        int labelWidth      = getMaxLabelWidth();
        int labelHeight     = labelTextHeight;
        int labelMarginLeft = LABEL_MARGIN_LEFT;
        int labelMarginTop  = (h-labelHeight)/2;
        int labelArcWidth   = 10;
        int labelArcHeight  = labelArcWidth;

        d.drawCustomItemLabel(label, labelFont, g,
                labelMarginLeft, labelMarginTop,
                labelWidth,      labelHeight,
                labelArcWidth,   labelArcHeight,
                labelMarginLeft + labelPaddingX,
                labelMarginTop  + labelTextHeight,
                Graphics.BOTTOM | Graphics.LEFT);
    }

    /**
     *  @return the appropriate label given the address state
     */
    private String getAddressTypeLabel() {
        switch (state) {
            case Address.TO:
                return Localization.getMessages().TO_LABEL;
            case Address.CC:
                return Localization.getMessages().CC_LABEL;
            case Address.BCC:
                return Localization.getMessages().BCC_LABEL;
            default: // ADDRESS_STATE_NONE
                return "";
        }
    }

    /**
     * Return the current state as one of:
     * Address.TO, Address.CC, Address.BCC or ContactItem.ADDRESS_STATE_NONE
     */
    public int getState() {
        return state;
    }

    /**
     * Set the state.
     *
     * @param state the state to be set. this has to be one of
     * Address.TO,  Address.CC, Address.BCC or ContactItem.ADDRESS_STATE_NONE
     * or the function will do nothing
     */
    public void setState(int state) {
        if( (state == Address.TO) ||
            (state == Address.CC) ||
            (state == Address.BCC) ||
            (state == ADDRESS_STATE_NONE) ) {
            this.state = state;
        }
    }

    /**
     * Set the current contact
     * @param c the new contac
     */
    public void setContact(Contact c) {
        this.contact = c;
    }

    /**
     * Check if the contact item has more than one email
     * @return true if this contact has a not null or empty
     * secondary or tertiary email
     */
    public boolean hasMultipleEmails() {
        int count = 0;
        if(!StringUtil.isNullOrEmpty( contact.getDefaultEmail())) {
            count++;
        }
        if(!StringUtil.isNullOrEmpty( contact.getEmail_2())) {
            count++;
        }
        if(!StringUtil.isNullOrEmpty( contact.getEmail_3())) {
            count++;
        }
        if (count < 2)  {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Reset the contact item height. since the height is cached, this method
     * should be called when the height should be recalculated
     * (e.g. if a secondary email has been added)
     */
    public void resetHeight() {
        this.height = 0;
    }

    /**
     * Set the selected email.
     * value can be one of Contact.DEFAULT_EMAIL, Contac.SECONDARY_EMAIL,
     * Contac.TERTIARY_EMAIL
     */
    public void setSelectedEmailIndex(int emailIndex) {
        this.selectedEmailIndex = emailIndex;
    }
}
