/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.ui.view;

import java.util.Vector;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.Font;

import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.loc.LocalizedMessages;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.utils.UiUtils;
import com.funambol.mailclient.ui.utils.DevInfo;
import com.funambol.util.StringUtil;
import com.funambol.util.Log;


/**
 * This class represent a canvas which can handle its own menus. The add/remove
 * commands are reimplemented to draw low level menus. The preferred way to
 * describe a menu is by using MenuItem objects. But for backward compatibility
 * purposes, the <i>Command</i> interface is kept so that no changes are
 * required to users of the Canvas.
 * Users shall extend this class to implement the abstract methods. This class
 * must be able to handle "paint" and "keyPressed" events. If such events are
 * not for the menu, then they are propagated to the subclasses, using the
 * abstract methods.
 */
public abstract class CustomMenuCanvas extends TouchCanvas implements TestableScreen, AlertableScreen {

    private static final String TAG_LOG = "CustomMenuCanvas";

    private static final int ARROW_WIDTH         = 8;
    private static final int MENU_TEXT_X_OFFSET  = 6;
    private static final int TEXT_ARROW_DISTANCE = 2;
    private static final int MENU_TOP_MIN_DISTANCE = 40;
    private static final int MENU_ITEMS_DISTANCE   = 5;
    private static final int MENU_SEPARATOR_HEIGHT = 4;
    private static final int MENU_SCROLL_BAR_WIDTH = 3;
    private static final int MENU_SCROLL_BAR_X_DISTANCE = 3;
    private static final int MENU_SCROLL_BAR_Y_DISTANCE = 3;
    private static final int MENU_FRAME_WIDTH = 3;
    private static final int MENU_BOTTOM_TRIANGLE_HEIGHT = 6;
    private static final int MENU_BOTTOM_TRIANGLE_WIDTH = 6;
    private static final int MENU_ICON_X_DISTANCE = 10;
    private static final int SUBMENUS_X_OFFSET = 20;

    //#ifdef traslateTitle
        protected static final int TITLE_Y_OFFSET = 30;
    //#else
    //# protected static final int TITLE_Y_OFFSET = 0;
    //#endif


    private GraphicalTheme theme = GraphicalThemeFactory.getGraphicalTheme();

    /**
     * Specifies if the menu is currently on display
     */
    private boolean menuOnDisplay = false; 

    /**
     * This is the stack of opened menus
     */
    private Vector   openedMenus = new Vector();

    /**
     * Represents the number of soft keys
     */
    private int numberSoftKeys = 2;

    /**
     * This is the main menu
     */
    private Menu mainMenu = new Menu(Localization.getMessages().MENU_LABEL);

    /**
     * This is a list of soft keys actions associated to available
     * keys
     */
    private Vector softKeys = new Vector();

    // TODO: generalize and make it a list of previous sk
    private SoftKey previousSk1 = null;
    private SoftKey previousSk2 = null;

    private Command mainMenuSelectCommand = new Command(Localization.getMessages().SELECT_MENU_OPTION,
                                                        Command.SCREEN, 0);
    private Command mainMenuCancelCommand = new Command(Localization.getMessages().CANCEL_MENU_OPTION,
                                                        Command.SCREEN, 0);

    /**
     * The font to be used in the menu
     */
    private Font menuFont;

    /**
     * Menu bar height. A value of zero indicates that no menu bar shall be
     * displayed
     */
    private int menuBarHeight;

    /**
     * This variable indicates if the layout is dirty, i.e. commands were
     * added/removed after the last paint.
     */
    private boolean layoutDirty = true;

    /**
     * This is the list of menu items (including items in the menu and
     * associated to soft keys.
     */
    private Vector items = new Vector();

    /**
     * This is the canvas total width. The code cannot rely on super.getWidth()
     * as on some devices it is not reported correctly (e.g. V3xx). This
     * variable is instead modified by the sizeChanged call back method.
     */
    private int totalWidth;

    /**
     * This is the canvas total height. The code cannot rely on super.getHeight()
     * as on some devices it is not reported correctly (e.g. V3xx). This
     * variable is instead modified by the sizeChanged call back method.
     */
    private int totalHeight;

    /**
     * Specifies if soft keys shall be placed left to right or right to left.
     * By default the placement is left to right.
     */
    private boolean placeLeftToRight = true;
    //private boolean placeLeftToRight = false;

    /**
     * True if the repaint request was issued to refresh the menu. This is used
     * to improve performance, avoiding a complete screen repaint when just the
     * menu needs repainting
     */
    private boolean paintForMenu = false;
    
    public CustomMenuCanvas() {
        super();

        // Set in full screen mode as we paint our own menus
        // Disable full screen mode. It will be enabled later
        setFullScreenMode(true);

        Drawer drawer = UIController.getDrawer();
        menuFont = drawer.getGraphicalTheme().getMenuFont();

        // The menu bar is for a single line of text
        menuBarHeight = menuFont.getHeight() + 3;

        totalWidth = super.getWidth();
        totalHeight = super.getHeight();

        //#ifdef isBlackberry
        //# menuBarHeight = 0;
        //# numberSoftKeys = 0;
        //#endif
    }

    public int getHeight() {
        if (totalHeight > 0) {
            return totalHeight - menuBarHeight;
        } else {
            return super.getHeight() - menuBarHeight;
        }
    }

    public void commandAction(Command command, Displayable displayable) {
        // These options correspond to the main menu soft keys
        if (command == mainMenuSelectCommand) {
            fireMenu();
        } else if (command == mainMenuCancelCommand) {
            cancelMenu();
        }
    }

    public void addCommand(Command command) {
        MenuItem menuItem = new MenuItem(command);
        addMenuItem(menuItem);
    }

    public void addMenuItem(MenuItem item) {
        // Preliminary check: do not add twice the same command
        // (while we scan the list, we also compute the insert position based on
        // the menu entry priority)
        int pos = -1;
        for(int i=0;i<items.size();++i) {
            MenuItem otherItem = (MenuItem) items.elementAt(i);

            if (item.getCommand() == otherItem.getCommand() && item.getCommand() != null) {
                return;
            }
            if (otherItem == item) {
                return;
            }
            if (pos == -1 && item.getPriority() < otherItem.getPriority()) {
                pos = i;
            }
        }
        // Since the command is not in the current list, we add it
        if (pos != -1) {
            items.insertElementAt(item, pos);
        } else {
            items.addElement(item);
        }
        layoutDirty = true;
    }

    public void updateMenuItem(MenuItem item) {
        for(int i=0;i<items.size();++i) {
            MenuItem otherItem = (MenuItem) items.elementAt(i);
            if (item.getText().equals(otherItem.getText())) {
                items.setElementAt(item, i);
            }
        }
        layoutDirty = true;
    }

    public void removeMenuItem(MenuItem item) {
        for(int i=0;i<items.size();++i) {
            MenuItem otherItem = (MenuItem) items.elementAt(i);
            if (item.getText().equals(otherItem.getText())) {
                items.removeElementAt(i);
            }
        }
        layoutDirty = true;
    }

    public void addSeparator(int priority) {
        MenuItem separatorItem = new MenuItem(priority);
        addMenuItem(separatorItem);
    }

    public void removeCommand(Command command) {
        // Search in the list of items, and remove this command
        for(int i=0;i<items.size();++i) {
            MenuItem item = (MenuItem) items.elementAt(i);
            if (item.getCommand() == command) {
                items.removeElementAt(i);
                break;
            }
        }
        layoutDirty = true;
    }

    public void enableCommand(Command command) {
        enableDisableCommand(command, true);
        layoutDirty = true;
    }

    public void disableCommand(Command command) {
        enableDisableCommand(command, false);
        layoutDirty = true;
    }

    /**
     * This is equivalent to keyPressed after filtering out events for the menu.
     */
    protected abstract void handleKeyPressed(int keyCode);

    /**
     * This is equivalent to keyRepeated after filtering out events for the menu.
     */
    protected abstract void handleKeyRepeated(int keyCode);

    /**
     * This is equivalent to paint
     */
    protected abstract void paintContent(Graphics graphics);

    /**
     * This is equivalent to handleFTouchEvent after filtering out events for the menu.
     */
    protected abstract void handleTouchEvent(FTouchEvent event);

    private void enableDisableCommand(Command command, boolean enabled) {
        for(int i=0;i<items.size();++i) {
            MenuItem item = (MenuItem) items.elementAt(i);
            if (item.getCommand() == command) {
                if (enabled) {
                    item.enable();
                } else {
                    item.disable();
                }
                return;
            } else if(item.isSubMenu()) {
                Menu submenu = item.getSubMenu();
                for(int j=0; j<submenu.totalNumberOfItems(); ++j) {
                    MenuItem submenuitem = (MenuItem) submenu.getMenuItem(j);
                    if (submenuitem.getCommand() == command) {
                        if (enabled) {
                            submenuitem.enable();
                        } else {
                            submenuitem.disable();
                        }
                        return;
                    } 
                }
            }
        }
    }

    /**
     * handle keyrepeated events
     * @param keycode the keycode of the key
     */
    protected void keyRepeated(int keyCode) {
        if (isMenuOnDisplay() || isSoftKey(keyCode) != -1) {
            int dir = getGameAction(keyCode);
            // this is an event for the menu
            if (isDownKey(dir, keyCode)) {
                scrollDownMenu();
            } else if (isUpKey(dir, keyCode)) {
                scrollUpMenu();
            }
        } else {
            handleKeyRepeated(keyCode);
        }
    }

    /**
     * handle the up / down navigation
     */
    protected void keyPressed(int keyCode) {

        if (isMenuOnDisplay() || isSoftKey(keyCode) != -1) {
            int dir = getGameAction(keyCode);

//#ifdef isBlackberry
//#         // On BB the left key (menu) is not handled like a soft key. It can
//#         // be used to popup the menu.
//#         if (isSoft1(dir, keyCode)) {
//#             if (isMenuOnDisplay()) {
//#                 fireMenu();
//#             } else {
//#                 openMainMenu();
//#             }
//#             return;
//#         }
//#         // On BB the right key (back) is not handled like a soft key,
//#         // but has special handling as its purpose is not dynamic. It can
//#         // serve only as back
//#         if (isSoft2(dir, keyCode)) {
//#             if (isMenuOnDisplay()) {
//#                 cancelMenu();
//#             } else {
//#                 handleKeyPressed(keyCode);
//#             }
//#             return;
//#         }
//#endif

            // Soft keys are always enabled
            if (placeLeftToRight) {
                if (isSoft2(dir, keyCode)) {
                    if (softKeys.size() > 0) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(0);
                        softKeyPressed(sk);
                    }
                } else if (isSoft1(dir, keyCode)) {
                    if (softKeys.size() > 1) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(1);
                        softKeyPressed(sk);
                    } else if (numberSoftKeys == 1) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(0);
                        softKeyPressed(sk);
                    }
                }
            } else {
                if (isSoft2(dir, keyCode)) {
                    if (softKeys.size() > 1) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(1);
                        softKeyPressed(sk);
                    } else if (numberSoftKeys == 1) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(0);
                        softKeyPressed(sk);
                    }
                } else if (isSoft1(dir, keyCode)) {
                    if (softKeys.size() > 0) {
                        SoftKey sk = (SoftKey)softKeys.elementAt(0);
                        softKeyPressed(sk);
                    }
                }
            }

            // Now handle action keys
            if (menuOnDisplay) {
                // this is an event for the menu
                if (isDownKey(dir, keyCode)) {
                    scrollDownMenu();
                } else if (isUpKey(dir, keyCode)) {
                    scrollUpMenu();
                } else if (isFireKey(dir, keyCode)) {
                    fireMenu();
                } else if (isRightKey(dir, keyCode)) {
                    OpenedMenuStatus status = getCurrentMenuStatus();
                    Menu menu = status.menu;
                    MenuItem item = menu.getMenuItem(status.index);
                    if (item.isSubMenu()) {
                        openSubMenu(item, status);
                    }else if (openedMenus.size()<=1){
                        handleKeyPressed(keyCode);
                    }
                } else if (isLeftKey(dir, keyCode)) {
                    // The left key can be used to close any submenus (not the main
                    // menu)
                    OpenedMenuStatus status = getCurrentMenuStatus();
                    Menu menu = status.menu;
                    MenuItem item = menu.getMenuItem(status.index);

                    if (openedMenus.size()>1) {
                        cancelMenu();
                    }else if (!item.isSubMenu()){
                        handleKeyPressed(keyCode);
                    }
                }else{
                    handleKeyPressed(keyCode);
                }
            }
        } else {
            handleKeyPressed(keyCode);
        }
    }

    protected boolean isMenuOnDisplay() {
        return menuOnDisplay;
    }

    private void closeMenu() {
        // Close the main menu
        menuOnDisplay = false;
        paintForMenu = false;
        // Restore the previous soft keys
        softKeys.removeAllElements();
        if (previousSk1 != null) {
            softKeys.addElement(previousSk1);
        }
        if (previousSk2 != null) {
            softKeys.addElement(previousSk2);
        }
    }

    private void cancelMenu() {
        // Close one level only
        if (openedMenus.size() > 0) {
            openedMenus.removeElementAt(openedMenus.size() - 1);

            if (openedMenus.size() == 0) {
                closeMenu();
            }
            // Since the menu disappeared, we need to redraw
            repaint();
        } else {
            Log.error(TAG_LOG, "No menu to cancel");
        }
    }


    private OpenedMenuStatus prepareMenu(Menu menu, OpenedMenuStatus parent) {
        OpenedMenuStatus menuStatus = new OpenedMenuStatus();
        menuStatus.menu = menu;

        // Compute menu height, width and first selectable item
        int totItems = menu.totalNumberOfItems();
        int width  = 0;
        int height = 0;
        int firstEnabled = -1;
        boolean scroll = false;

        for(int i=0;i<totItems;++i) {
            int deltaHeight;
            MenuItem item = menu.getMenuItem(i);
            if (!item.isSeparator()) {
                String label = item.getText();
                int itemWidth = menuFont.stringWidth(label);
                if(!item.getCommandInfo().equals("")){
                    itemWidth += menuFont.stringWidth(item.getCommandInfo())+ MENU_TEXT_X_OFFSET;
                }
                if (item.isSubMenu()) {
                    // We need to take into account the space needed for the
                    // arrow
                    itemWidth += ARROW_WIDTH + TEXT_ARROW_DISTANCE;
                }
                if (item.getIcon() != null) {
                    // We need to take into account the space needed for the
                    // SelectableMenuItem icon
                    itemWidth += item.getIcon().getWidth() +
                            MENU_ICON_X_DISTANCE;
                }
                if (itemWidth > width) {
                    width = itemWidth;
                }

                if (width + SUBMENUS_X_OFFSET > getWidth()) {
                    width = getWidth() - SUBMENUS_X_OFFSET -  ( 4 * MENU_FRAME_WIDTH) - MENU_SCROLL_BAR_WIDTH - MENU_SCROLL_BAR_Y_DISTANCE;
                }
                //}
                deltaHeight = menuFont.getHeight() + MENU_ITEMS_DISTANCE;
                if (item.isEnabled() && firstEnabled == -1) {
                    firstEnabled = i;
                }
            } else {
                // This is a separator
                deltaHeight = MENU_SEPARATOR_HEIGHT;
            }
            // Increment the menu height
            if (!scroll && (height + deltaHeight < getHeight() - MENU_TOP_MIN_DISTANCE)) {
                height += deltaHeight;
                menuStatus.lastVisible = i;
            } else {
                scroll = true;
            }
        }
        menuStatus.index = firstEnabled;
        // The last visible is exclusive, it is the first one not shown, while
        // the loop above sets the last one shown, so we need to increment by
        // one
        menuStatus.lastVisible++;

        menuStatus.scroll = scroll;

        // Increase the width to separate the last char from the border
        menuStatus.width = width + MENU_TEXT_X_OFFSET * 2;
        menuStatus.height = height;

        return menuStatus;
    }

    private void softKeyPressed(SoftKey sk) {

        if (sk.menuItem != null) {
            // This is a shortcut menu
            MenuItem item = sk.menuItem;
            Command command = item.getCommand();

            if (command != null && item.isEnabled()) {
                commandAction(command, this);
            }
        } else {
            // This is the main menu
            if (menuOnDisplay) {
                cancelMenu();
            } else {
                openMainMenu();
            }
        }
    }

    protected void openMainMenu() {
        openedMenus.removeAllElements();
        menuOnDisplay = true;
        // Stack the global menu
        OpenedMenuStatus status = prepareMenu(mainMenu, null);
        openedMenus.addElement(status);

        // Replace soft keys with menu shortcuts
        replaceSoftKeysOnMenuOpen();
        repaintForMenu();
    }

    private void replaceSoftKeysOnMenuOpen() {
        // Replace the soft keys
        if (softKeys.size() > 0) {
            previousSk1 = (SoftKey) softKeys.elementAt(0);
        }
        if (softKeys.size() > 1) {
            previousSk2 = (SoftKey) softKeys.elementAt(1);
        }

        softKeys.removeAllElements();

        if (numberSoftKeys > 0) {
            SoftKey cancelKey = new SoftKey();
            // On the BB the back function is always associated to the back
            // button, so the menu key serves to select
            MenuItem newMenuItem;
//#ifdef isBlackberry
//#         newMenuItem = new MenuItem(mainMenuSelectCommand);
//#else
            newMenuItem = new MenuItem(mainMenuCancelCommand);
//#endif
            cancelKey.menuItem = newMenuItem;
            softKeys.addElement(cancelKey);
        }

        if (numberSoftKeys > 1) {
            SoftKey selectKey = new SoftKey();
            MenuItem selectMenuItem = new MenuItem(mainMenuSelectCommand);
            selectKey.menuItem = selectMenuItem;
            softKeys.addElement(selectKey);
        }
    }


    protected int isSoftKey(int keyCode) {
        if (isSoft1(getGameAction(keyCode), keyCode)) {
            return 0;
        } else if (isSoft2(getGameAction(keyCode), keyCode)) {
            return 1;
        }
        return -1;
    }

    private boolean isSoft1(int dir, int keyCode) {
        String name = getKeyName(keyCode);
        if (StringUtil.equalsIgnoreCase("SOFT1", name)) {
            return true;
        }
        if (dir == 0 && keyCode == -6) {
            return true;
        }
        return false;
    }

    private boolean isSoft2(int dir, int keyCode) {
        String name = getKeyName(keyCode);
        if (StringUtil.equalsIgnoreCase("SOFT2", name)) {
            return true;
        }
        if (dir == 0 && keyCode == -7) {
            return true;
        }
        return false;
    }

    protected boolean isDownKey(int dir, int keyCode) {
        boolean isDownKey =
                ((dir == Canvas.DOWN) &&
                ((keyCode < 0) || (keyCode == Canvas.DOWN)));

        return isDownKey;
    }

    protected boolean isUpKey(int dir, int keyCode) {
        boolean isUpKey =
                ((dir == Canvas.UP) &&
                ((keyCode < 0) || (keyCode == Canvas.UP)));

        return isUpKey;
    }

    protected boolean isLeftKey(int dir, int keyCode) {
        boolean isLeftKey =
                ((dir == Canvas.LEFT) &&
                ((keyCode < 0) || (keyCode == Canvas.LEFT)));

        return isLeftKey;
    }

    protected boolean isRightKey(int dir, int keyCode) {
        boolean isRightKey =
                ((dir == Canvas.RIGHT) &&
                ((keyCode < 0) || (keyCode == Canvas.RIGHT)));

        return isRightKey;
    }


    protected boolean isFireKey(int dir, int keyCode) {

        if (((dir == Canvas.FIRE) &&
            ((keyCode < 0) || (keyCode == Canvas.FIRE) || (keyCode == 20))))
        {
            return true;
        } else if (dir == 0 && keyCode == Canvas.FIRE) {
            // On most devices there is no game key associated to the fire, but this is a fire
            if (DevInfo.getFamily() != null && (DevInfo.getFamily().equals("E71") ||
                                                DevInfo.getFamily().equals("E71X"))) {
                return false;
            }
            //#ifdef isBlackberry
            //# // On BB Pearl, this is a clear
            //# if (DevInfo.getFamily() != null && DevInfo.getFamily().equals("8100")) {
            //#     return false;
            //# }
            //#endif
            return true;
            // The following code is for the BB Storm where the fire is
            // identified by this pair of codes
        //#ifdef isBlackberry
        //# } else if (DevInfo.getFamily() != null && DevInfo.getFamily().equals("9500") &&
        //#            dir == 5 && keyCode == Canvas.FIRE) {
        //#     return true;
        //#endif
        } else if (dir == Canvas.FIRE && keyCode == 53) {
            // This is the fire on Nokia 2nd edition
            return true;
        } else {
            return false;
        }
    }

    private OpenedMenuStatus getCurrentMenuStatus() {
        int statusSize = openedMenus.size();
        if (statusSize > 0) {
            return (OpenedMenuStatus) openedMenus.elementAt(statusSize - 1);
        } else {
            return null;
        }
    }

    private void updateOpenedMenuStatusFromLast(Menu menu, OpenedMenuStatus status) {
        int totItems = menu.totalNumberOfItems();
        int deltaHeight = 0;
        int height = 0;
        int i = 0;
        for(i=totItems-1;i>=0;--i) {
            MenuItem item = menu.getMenuItem(i);
            if (!item.isSeparator()) {
                deltaHeight = menuFont.getHeight() + MENU_ITEMS_DISTANCE;
            } else {
                // This is a separator
                deltaHeight = MENU_SEPARATOR_HEIGHT;
            }

            if (height + deltaHeight < getHeight() - MENU_TOP_MIN_DISTANCE) {
                height += deltaHeight;
            } else {
                break;
            }
        }
        status.lastVisible = totItems;
        status.firstVisible = i + 1;
        status.height = height;
    }

    private void updateOpenedMenuStatus(Menu menu, OpenedMenuStatus status,
                                        int lastVisible) {
        int height = 0;
        int i;
        int totItems = menu.totalNumberOfItems();
        boolean scroll = false;
        
        if(status.firstVisible == -1){
            status.firstVisible =  0;
        }

        

        for(i=status.firstVisible;i<totItems;++i) {
            int deltaHeight;
            MenuItem item = menu.getMenuItem(i);
            if (!item.isSeparator()) {
                deltaHeight = menuFont.getHeight() + MENU_ITEMS_DISTANCE;
            } else {
                // This is a separator
                deltaHeight = MENU_SEPARATOR_HEIGHT;
            }
            // Increment the menu height
            if (height + deltaHeight < getHeight() - MENU_TOP_MIN_DISTANCE) {
                height += deltaHeight;
            } else {
                scroll = true;
         
                break;
            }
        }
        if (!scroll) {
            i = totItems;
        }
        status.lastVisible = i;
    }


    private void scrollDownMenu() {
        OpenedMenuStatus status = getCurrentMenuStatus();
        Menu menu = status.menu;
        boolean update = false;
        int delta = 0;
        int menuIndex = status.index;
        // Search the next enabled item
        while (menu.totalNumberOfItems() > 0 && (menuIndex + delta) < (menu.totalNumberOfItems() -1)) {
            delta++;
            MenuItem item = menu.getMenuItem(menuIndex + delta);
            if (item.isEnabled()) {
                update = true;
                break;
            }
        }
        if (update) {
            status.index = menuIndex + delta;
            int fontHeight  = menuFont.getHeight();
            // Check if we need to scroll down
            int lastVisible = status.lastVisible;

            if (status.index >= status.lastVisible) {
                // Scroll down
                status.firstVisible+=delta;
                // We need to update the lastVisible in the menu
                updateOpenedMenuStatus(menu, status, lastVisible);
            }
            repaintForMenu();
        } else {
            // Scroll down from the top because we must wrap around
            status.index = 0;
            status.firstVisible = 0;
            if (menu.getMenuItem(0).isEnabled() || status.scroll == false) {
                int firstEnabled = 0;
                while (!menu.getMenuItem(firstEnabled).isEnabled()) {
                    firstEnabled++;
                }
                status.index = firstEnabled;
                updateOpenedMenuStatus(menu, status, -1);
                repaintForMenu();
            } else {
                status.lastVisible = -1;
                scrollDownMenu();
            }
        }
    }

    private void scrollUpMenu() {
        OpenedMenuStatus status = getCurrentMenuStatus();
        Menu menu = status.menu;
        boolean update = false;
        int menuIndex = status.index;
        int delta = 0;
        while ((menuIndex - delta) > 0) {
            delta++;
            MenuItem item = menu.getMenuItem(menuIndex - delta);
            if (item.isEnabled()) {
                update = true;
                break;
            }
        }
        if (update) {
            status.index -= delta;
            if (status.index < status.firstVisible) {
                // Scroll up
                status.firstVisible-=delta;
                updateOpenedMenuStatus(menu, status, -1);
            }
            repaintForMenu();
        } else {
            // Scroll up from the bottom because we must wrap around
            status.index = status.menu.totalNumberOfItems() - 1;
            if (menu.getMenuItem(status.index).isEnabled()) {
                updateOpenedMenuStatusFromLast(menu, status);
                repaintForMenu();
            } else {
                status.lastVisible = status.menu.totalNumberOfItems();
                scrollUpMenu();
            }
        }
    }

    protected class OpenedMenuStatus {
        public Menu menu = null;
        public int index = 0;
        public int firstVisible = 0;
        public int lastVisible  = 0;
        public int width = 0;
        public int height = 0;
        public boolean scroll = false;
        public int x;
        public int y;
        public int selectedItemY = 0;
    }

    protected void openSubMenu(MenuItem item, OpenedMenuStatus parent) {
        // Stack the new submenu
        OpenedMenuStatus menuStatus = prepareMenu(item.getSubMenu(), parent);
        openedMenus.addElement(menuStatus);
        repaintForMenu();
    }

    private void fireMenu() {
        OpenedMenuStatus status = getCurrentMenuStatus();
        Menu menu = status.menu;
        MenuItem item = menu.getMenuItem(status.index);

        // If this is a command, we just execute the commandAction, but
        // if it a submenu we shall handle it here
        if (item.isSubMenu()) {
            openSubMenu(item, status);
        } else {
            Command command = item.getCommand();
            // Close completely the menu
            closeMenu();
            //menuOnDisplay = false;
            repaint();
            // Now execute the user command
            commandAction(command, this);
        }
    }

    /**
     * Paint the screen
     */
    protected void paint(Graphics graphics) {
        paint(graphics, true, false);
    }

    /**
     * Paint the screen
     *
     * @param graphics the graphics
     * @param optimized if true the method checks the paintForMenu flag and if
     *        this is set, it only paints the menu, otherwise everything gets painted.
     *        This is done to speed up the repaint when there are changes that
     *        are related to the menu only.
     */
    protected void paint(Graphics graphics, boolean optimized, boolean ignoreTitleOffset) {
        // We need to reset this flag every time we paint because after display
        // a high level screen, we need to get rid of the high level menu bar
        setFullScreenMode(true);
        graphics.setColor(theme.getMenuBarBgColor());
        if (!ignoreTitleOffset) {
            graphics.fillRect(0, 0, totalWidth, TITLE_Y_OFFSET);
            graphics.setColor(theme.getTitleColor());
            //#ifdef traslateTitle
            graphics.setFont(theme.getBoldFont());
            graphics.drawString(Localization.getMessages().TOP_BAR_NAME, totalWidth/2,TITLE_Y_OFFSET/2+(theme.getBoldFont().getHeight())/2, Graphics.BOTTOM|Graphics.HCENTER);
            graphics.translate(0, TITLE_Y_OFFSET);
            //#endif
        }
        
        // This is a workaround for a bug on Motorola V3xx. On these devices
        // asking for the super.getHeight() may return the size in non full size
        // mode. The only reliable way to get this size is to use the
        // sizeChanged callback. But this callback does not work on other
        // systems, like the WTK emulator. For this reason we take the biggest
        // screen size between the sizeChanged and super.getHeight().
        int newHeight = super.getHeight();
        if (newHeight > totalHeight) {
            totalHeight = newHeight;
        }

        // paint the actual content if we are not just refreshing the menu (or
        // if we were explicitely asked for this)
        if (optimized) {
            if (!paintForMenu) {
                paintContent(graphics);
            } else {
                paintForMenu = false;
            }
        } else {
            paintContent(graphics);
        }

        // If the set of commands changed, we need to recompute the layout
        if (layoutDirty) {
            computeCommandsLayout();
        }

        if (menuOnDisplay) {
            repaintMenu(graphics);
        }

        if (menuBarHeight > 0) {
            paintMenuBar(graphics);
        }
    }

    private void repaintMenu(Graphics graphics) {
        int yOffset = 0;
        int x0 = 0;

        
        if (!placeLeftToRight) {
            if (openedMenus.size() > 0) {
                OpenedMenuStatus menuStatus = (OpenedMenuStatus)openedMenus.elementAt(0);
                x0 = totalWidth - menuStatus.width - MENU_TEXT_X_OFFSET * 2;
            }
        }

        int currentX = x0;
        for(int i=0;i<openedMenus.size();++i) {
            OpenedMenuStatus menuStatus = (OpenedMenuStatus)openedMenus.elementAt(i);
            int menuHeight = menuStatus.height;
            int y0;

            if (getHeight() - menuHeight - yOffset >= 0) {
                y0 = getHeight() - menuHeight - yOffset;
            } else {
                y0 = 0;
            }

            menuStatus.y = y0;

            //The submenu must be drawn near the submenu arrow of its parent
            //Pay attention that this procedure is valid until the drawing logic
            //is to draw the submenu from bottom to top. Reverting this logic can
            //produce inconsistent drawing over the screen border.
            if (i>0) {
                OpenedMenuStatus parent = (OpenedMenuStatus)openedMenus.elementAt(i-1);
                int firstY = parent.selectedItemY - MENU_BOTTOM_TRIANGLE_HEIGHT+TITLE_Y_OFFSET;
                
                if (menuStatus.y > firstY || menuStatus.y == 0) {
                    menuStatus.y = firstY;
                }

                //calculating  X submenu
                int parentWidth = parent.width;
                int subMenuX = 0;

                int totalMenuWidth = parentWidth+menuStatus.width+6*MENU_FRAME_WIDTH+
                        MENU_SCROLL_BAR_WIDTH+2*MENU_SCROLL_BAR_X_DISTANCE;

                if(totalMenuWidth>= totalWidth){
                    subMenuX = parentWidth - (totalMenuWidth - totalWidth);
                    if(subMenuX<0){
                        subMenuX = 0;
                        menuStatus.width = getWidth() -  ( 6 * MENU_FRAME_WIDTH) - MENU_SCROLL_BAR_WIDTH - MENU_SCROLL_BAR_Y_DISTANCE ;
                    }
                }else{
                    subMenuX = parentWidth + ARROW_WIDTH/2;
                }
           
                menuStatus.x = subMenuX;
            }

            paintMenu(graphics, menuStatus);
            if (placeLeftToRight) {
                currentX += menuStatus.x;
            } else {
                currentX -= menuStatus.x;
            }
            yOffset += 10;
        }
    }

    private class SoftKey {
        public MenuItem menuItem;
        public Menu menu;
        public int index;
    }

    private void paintMenuBar(Graphics g) {
        int height = totalHeight;
        // Clip the menu bar
        g.setClip(0, height - menuBarHeight-TITLE_Y_OFFSET, getWidth(), menuBarHeight);
        // Draw the menu bar background
        g.setColor(theme.getMenuBarBgColor());
        g.fillRect(0, height - menuBarHeight-TITLE_Y_OFFSET, getWidth(), menuBarHeight);
        // Now write the options, this depends on the number of soft keys
        // available and if there is at least one menu item
        g.setColor(theme.getMenuBarFgColor());

        for(int i=0;i<softKeys.size();++i) {
            SoftKey sk = (SoftKey)softKeys.elementAt(i);
            // The first key is displayed on the right (if we have more soft
            // keys)
            MenuItem item = sk.menuItem;
            String text = null;
            if (item != null) {
                text = item.getText();
            } else {
                Menu menu = sk.menu;
                if (menu != null) {
                    text = menu.getTitle();
                }
            }

            if (text != null) {
                int x;
                boolean right;
                if (placeLeftToRight  && i == 0 ||
                    !placeLeftToRight && i == 1) {
                    right = true;
                } else {
                    right = false;
                }

                if (right) {
                    x = getWidth() - menuFont.stringWidth(text) - MENU_TEXT_X_OFFSET; 
                } else {
                    x = MENU_TEXT_X_OFFSET;
                }

                if (item == null || item.isEnabled()) {
                    g.setFont(menuFont);
                    g.setColor(theme.getMenuItemFgColor());
                } else {
                    g.setColor(0x888888);
                }
                int fontHeight = menuFont.getHeight();
                int half = (menuBarHeight - fontHeight) / 2;
                int y = height - menuBarHeight-TITLE_Y_OFFSET + half + fontHeight;
                g.drawString(text, x, y, Graphics.BOTTOM | Graphics.LEFT);
            }
        }
    }


    private void paintMenu(Graphics graphics, OpenedMenuStatus openedMenu) {

        Menu menu = openedMenu.menu;
        int menuIndex = openedMenu.index;
        int width     = openedMenu.width;

        int totItems = menu.totalNumberOfItems();
        int menuHeight = openedMenu.height;

        int y0 = openedMenu.y - MENU_BOTTOM_TRIANGLE_HEIGHT - TITLE_Y_OFFSET;
        int x0 = openedMenu.x + MENU_TEXT_X_OFFSET/2;
        int scrollBarSpace = MENU_SCROLL_BAR_WIDTH + MENU_SCROLL_BAR_X_DISTANCE;
        int extraXSpace = MENU_TEXT_X_OFFSET;
        
        // Clip to the menu region
        graphics.setClip(x0, y0 - MENU_FRAME_WIDTH * 2 - 2, width + extraXSpace + scrollBarSpace + 1 + MENU_FRAME_WIDTH * 2, totalHeight);
        // Clear the background
        graphics.setColor(theme.getMenuItemBgColor());
        graphics.fillRect(x0, y0 - MENU_FRAME_WIDTH - 2, width + extraXSpace + scrollBarSpace + MENU_FRAME_WIDTH * 2, menuHeight + 2 + MENU_FRAME_WIDTH * 2 - 3);
        // Draw the frame
        graphics.setColor(0, 0, 0);
        graphics.drawRect(x0, y0 - MENU_FRAME_WIDTH - 2, width + extraXSpace + scrollBarSpace + MENU_FRAME_WIDTH * 2, menuHeight + 2 + MENU_FRAME_WIDTH * 2 - 3);
        graphics.setColor(theme.getMenuScrollBarBgColor());
        graphics.drawRect(x0+2, y0 - MENU_FRAME_WIDTH - 2 + 2, width + extraXSpace + scrollBarSpace - 4 + MENU_FRAME_WIDTH * 2, menuHeight + 2 + MENU_FRAME_WIDTH * 2 - 7);
        graphics.setColor(theme.getMenuItemBgColor());
        graphics.drawRect(x0+3, y0 - MENU_FRAME_WIDTH - 2 + 3, width + extraXSpace + scrollBarSpace - 6 + MENU_FRAME_WIDTH * 2, menuHeight + 2 + MENU_FRAME_WIDTH * 2 - 9);

        if (openedMenu == (OpenedMenuStatus)openedMenus.elementAt(0)) {
            //Draw the triangle anchored to the soft button if this is the first
            //menu tab
            int triangleX1 = x0 + 10;
            int triangleX2 = x0 + 20;
            int triangleX3 = x0 + 30;
            int tY0 = menuHeight + MENU_FRAME_WIDTH - 3;
            graphics.setColor(0, 0, 0);
            graphics.fillTriangle(triangleX1, y0 + tY0 , triangleX2, getHeight()-TITLE_Y_OFFSET, triangleX3, y0 + tY0);
            graphics.setColor(theme.getMenuScrollBarBgColor());
            graphics.fillTriangle(triangleX1, y0 + tY0 -2, triangleX2, getHeight() -2-TITLE_Y_OFFSET, triangleX3, y0 + tY0 -2);
            graphics.setColor(theme.getMenuItemBgColor());
            graphics.fillTriangle(triangleX1, y0 + tY0 -3, triangleX2, getHeight() -3-TITLE_Y_OFFSET, triangleX3, y0 + tY0 -3);
        }

        extraXSpace += MENU_FRAME_WIDTH;
        // This is the bar background
        //the scrollBarBackgruound is always needed but...
        int barX = x0 + width + extraXSpace + MENU_SCROLL_BAR_X_DISTANCE / 2;
        int barY = y0 + MENU_SCROLL_BAR_Y_DISTANCE;
        int scrollBarHeight = menuHeight - 2 * MENU_SCROLL_BAR_Y_DISTANCE;
        graphics.setColor(theme.getMenuScrollBarBgColor());
        graphics.fillRoundRect(barX, barY, MENU_SCROLL_BAR_WIDTH,
                               scrollBarHeight,
                                6, 6);

        //... we show the active scroll bar only if necessary
        if (openedMenu.scroll) {
            // This is the bar foreground. The size depends on how many items
            // are visible at the same time. The position depends on the
            // firstVisible item
            int thumbRelSize = ((openedMenu.lastVisible - openedMenu.firstVisible) * 100) / totItems;
            int thumbSize = (scrollBarHeight * thumbRelSize) / 100;
            graphics.setColor(theme.getMenuScrollBarFgColor());
            int fgBarY = barY;
            if (openedMenu.firstVisible > 0) {
                fgBarY += (scrollBarHeight / totItems) * openedMenu.firstVisible;
            }
            // If this is the last item, we make sure we get to the end of the
            // scroll bar (rounding math may prevent this)
            if (openedMenu.lastVisible >=  totItems) {
                thumbSize = scrollBarHeight - (fgBarY - barY);
            }
            graphics.fillRoundRect(barX, fgBarY, MENU_SCROLL_BAR_WIDTH,
                                   thumbSize,
                                   6, 6);
        }

        int fontHeight  = menuFont.getHeight();

        // Now print the labels
        int currentY = y0;
        int firstVisible = openedMenu.firstVisible;
        for(int i=firstVisible;i<openedMenu.lastVisible;++i) {
            MenuItem item = menu.getMenuItem(i);
            if (item.isSeparator()) {
                graphics.setColor(0, 0, 0);
                //graphics.drawLine(x0, currentY, x0 + width + extraXSpace, currentY);
                graphics.drawLine(x0 + MENU_TEXT_X_OFFSET, currentY, barX - MENU_TEXT_X_OFFSET, currentY);
                graphics.setColor(theme.getMenuScrollBarBgColor());
                graphics.drawLine(x0 + MENU_TEXT_X_OFFSET, currentY + 1, barX - MENU_TEXT_X_OFFSET, currentY + 1);
                graphics.drawLine(x0 + MENU_TEXT_X_OFFSET, currentY + 2, barX - MENU_TEXT_X_OFFSET, currentY + 2);
                currentY += MENU_SEPARATOR_HEIGHT;
            } else {
                String label = item.getText();
                int availableLabelWidth = width - MENU_TEXT_X_OFFSET;
                if (menuIndex == i) {
                    // this is the selected item
                    graphics.setFont(menuFont);
                    // Draw the background
                    graphics.setColor(theme.getMenuItemSelectedBgColor());
                    graphics.fillRoundRect(x0 + MENU_TEXT_X_OFFSET,
                                           currentY,
                                           width - MENU_TEXT_X_OFFSET,
                                           fontHeight + 2, 6, 6);
                    
                    //Draws the gradiented label around the selected item
                    int linesNumber = fontHeight + 3;
                    int[] colors = UiUtils.getGradient(0x02dcff, 0x0a7edf, linesNumber);
                    int roundedAmount = 3;
                    for (int j = 0; j < linesNumber; j++) {
                        if (j==0||j==linesNumber-1) {
                            roundedAmount = 1;
                        } else if (j==1||j==linesNumber-2) {
                            roundedAmount = 0;
                        }

                        graphics.setColor(colors[j]);
                        graphics.drawLine(x0 + MENU_TEXT_X_OFFSET + roundedAmount, currentY + j, barX - MENU_TEXT_X_OFFSET - roundedAmount, currentY + j);
                    }

                    //The item is selected. We need to understand where to draw
                    //the related submenus, thus we take note of the y
                    //coordinate of the current menu selected item.
                    openedMenu.selectedItemY = currentY;

                    // Prepare to draw the item
                    graphics.setColor(theme.getMenuItemSelectedFgColor());
                } else {
                    graphics.setColor(theme.getMenuItemFgColor());
                    graphics.setFont(menuFont);
                    if (!item.isEnabled()) {
                        graphics.setColor(theme.getMenuItemDisabledFgColor());
                    }
                }
                // If this is a submenu we shall draw the little arrow
                if (item.isSubMenu()) {
                    // The arrow shall be centered in the text line
                    int reduction = 3;
                    int arrowHeight = fontHeight - reduction * 2;
                    int arrowX = x0 + width - ARROW_WIDTH;
                    int arrowY = currentY + (fontHeight + 2)/2
                            - arrowHeight/2;
                    graphics.setColor(theme.getSubMenuArrowColor());
                    graphics.fillTriangle(arrowX + ARROW_WIDTH/2, arrowY,
                            x0 + width -2 + ARROW_WIDTH/2, arrowY + (arrowHeight/2),
                            arrowX + ARROW_WIDTH/2, arrowY + arrowHeight);
                    availableLabelWidth -= ARROW_WIDTH;
                    graphics.setColor(theme.getMenuItemSelectedFgColor());
                }
          
               
               
                Image icon = item.getIcon();
                if(!item.getCommandInfo().equals("")){
                   availableLabelWidth -=  menuFont.stringWidth(item.getCommandInfo())+ 2*MENU_TEXT_X_OFFSET;
                }
                if(icon!=null){
                    availableLabelWidth -= icon.getWidth() + MENU_ICON_X_DISTANCE;
                }
                label = UiUtils.cutString(label, availableLabelWidth, menuFont);
                int labelWidth = menuFont.stringWidth(label);
                
                int stringX = x0 + labelWidth + MENU_TEXT_X_OFFSET;
                if(!item.getCommandInfo().equals("")){
                    stringX += 2*MENU_TEXT_X_OFFSET;
                    int color = graphics.getColor();
                    graphics.setColor(item.getComandInfoFgColor());
                    graphics.drawString(item.getCommandInfo(), stringX,
                                currentY + (MENU_ITEMS_DISTANCE / 2) + fontHeight,
                                Graphics.BOTTOM|Graphics.LEFT);
                    graphics.setColor(color);
                }

                
                if(icon != null) {
                    int iconX = stringX + menuFont.stringWidth(item.getCommandInfo())+ MENU_ICON_X_DISTANCE;
                    int iconY = currentY + (MENU_ITEMS_DISTANCE / 2) +
                            fontHeight/2 - icon.getHeight()/2;
                    graphics.drawImage(icon, iconX, iconY, Graphics.TOP | Graphics.LEFT);
                }
                
                graphics.drawString(label, x0 + 2 * MENU_TEXT_X_OFFSET,
                                    currentY + (MENU_ITEMS_DISTANCE / 2) + fontHeight,
                                    Graphics.BOTTOM|Graphics.LEFT);
                currentY += fontHeight + MENU_ITEMS_DISTANCE;
            }
        }
    }

    private void computeCommandsLayout() {
        softKeys.removeAllElements();
        mainMenu = new Menu(Localization.getMessages().MENU_LABEL);
        boolean lastIsSep = false;
        // compute the layout for all commands, given the current set of entries
        for(int i=0;i<items.size();++i) {
            MenuItem item = (MenuItem) items.elementAt(i);
            // As long as we have space in the soft keys we place everything
            // there (skipping separators)
            if (softKeys.size() < numberSoftKeys - 1) {
                // There is room in the soft keys
                if (item.isSeparator()) {
                    Log.trace(TAG_LOG, "Skipping separator");
                } else {
                    SoftKey sk = new SoftKey();
                    sk.menuItem = item;
                    sk.index = softKeys.size();
                    softKeys.addElement(sk);
                }
            } else {
                // If this is the last item, then we can assign the remaining
                // soft key
                if (items.size() == (i+1) && softKeys.size() < numberSoftKeys) {
                    SoftKey sk = new SoftKey();
                    sk.menuItem = item;
                    sk.index = softKeys.size();
                    softKeys.addElement(sk);
                } else {
                    // If we haven't created a soft key for the main menu, then
                    // we do it here
                    if (softKeys.size() < numberSoftKeys) {
                        SoftKey sk = new SoftKey();
                        sk.menu = mainMenu;
                        sk.menuItem = null;
                        softKeys.addElement(sk);
                    }
                    // Create the new menu item
                    if (lastIsSep && item.isSeparator()) {
                        Log.trace(TAG_LOG, "Skipping adjacent separators");
                    } else {
                        mainMenu.addMenuItem(item);
                    }
                    lastIsSep = item.isSeparator();
                }
            }
        }

        // if the menu is currently on display, we shall replace the right soft
        // key with the "cancel" key
        if (isMenuOnDisplay()) {
            replaceSoftKeysOnMenuOpen();
        }

        layoutDirty = false;
    }

    protected void sizeChanged(int w, int h) {
        totalWidth = w;
        totalHeight = h;
    }

    private void repaintForMenu() {
        paintForMenu = true;
        repaint();
    }

    protected void handleFTouchEvent(FTouchEvent event) {
        Log.trace(TAG_LOG, "handleTouchEvent");
        int type     = event.getType();
        int tapCount = event.getTapCount();
        int y = event.getPressY(0);

        Log.debug(TAG_LOG, "handleFTouchEvent " + type + " -- " + tapCount);

        // We handle events in the menu bar or in the menu when it is opened
        if (y >= (totalHeight - menuBarHeight)) {
            // TODO FIXME (This shall be fixed once we support touchscreen
            // devices not BlackBerry)
        } else if (isMenuOnDisplay()) {
            OpenedMenuStatus menuStatus = getCurrentMenuStatus();
            int x = event.getPressX(0);
            //The sensible touch zone is the current menu drawin region
            int yMenuRegionStart = menuStatus.y;
            int yMenuRegionEnd = menuStatus.y + menuStatus.height;
            if (y >= yMenuRegionStart && y <= yMenuRegionEnd &&
                x > menuStatus.x && x < menuStatus.x + menuStatus.width) {
                // Click happened in the menu. We can select, fire or scroll
                switch (type) {
                    case FTouchEvent.PRESSED:
                    {
                        // An item shall be selected
                        int index = selectMenuItem(y);
                        if (index >= 0) {
                            //#ifndef isBlackberry
                            int activeElement =  menuStatus.index;
                            if(activeElement == index){
                                keyPressed(Canvas.FIRE);
                            }else{
                                menuStatus.index = index;
                                repaintForMenu();
                            }
                            //#else
                            //# menuStatus.index = index;
                            //# repaintForMenu();
                            //#endif  
                        }
                        
                        break;
                    }
                    case FTouchEvent.CLICK:
                    {
                        // Now we may need to open the msg
                        if (tapCount > 1) {
                            keyPressed(Canvas.FIRE);
                        }
                        break;
                    }
                    case FTouchEvent.DRAGGING_DOWN:
                        scrollUpMenu();
                        break;
                    case FTouchEvent.DRAGGING_UP:
                        scrollDownMenu();
                        break;
                    default:
                        Log.error(TAG_LOG, "Unhandled touch event in emnu");
                        break;
                }
            }
        } else {
            // We do not handle this event, so we pass it through
            handleTouchEvent(event);
        }
    }

    private int selectMenuItem(int y) {
        // The item selected depends on where the user clicked and the portion
        // of the menu which is currently visible
        OpenedMenuStatus menuStatus = getCurrentMenuStatus();
        Menu menu = menuStatus.menu;
        int currentY = menuStatus.y;
        int nextCurrentY = currentY;
        int fontHeight  = menuFont.getHeight();
        for(int i=menuStatus.firstVisible;i < menuStatus.lastVisible;++i) {
            MenuItem item = menu.getMenuItem(i);

            if (item.isSeparator()) {
                nextCurrentY += MENU_SEPARATOR_HEIGHT;
            } else {
                nextCurrentY += fontHeight + MENU_ITEMS_DISTANCE;
            }

            if (y >= currentY && y < nextCurrentY) {
                // Found it!
                if (item.isEnabled()) {
                    return i;
                } else {
                    return -1;
                }
            }
        }
        return -1;
    }

    
    public Image takeSnapshot() {
        // This is a workaround for a bug on Motorola V3xx. On these devices
        // asking for the super.getHeight() may return the size in non full size
        // mode. The only reliable way to get this size is to use the
        // sizeChanged callback. But this callback does not work on other
        // systems, like the WTK emulator. For this reason we take the biggest
        // screen size between the sizeChanged and super.getHeight().
        int height = totalHeight;
        int width  = super.getWidth();

        // Compute the bitmap for the two canvas
        Image canvasImage = Image.createImage(width, height - TITLE_Y_OFFSET);
        Graphics graphicsCanvas = canvasImage.getGraphics();
        paint(graphicsCanvas, false, true);
        return canvasImage;
    }

    /////////////////////////////////////////////////////////////
    /* Implementation of the TestableScreen interface */
    /////////////////////////////////////////////////////////////

//#ifdef TEST_MODE
//# public Vector getCommands() {
//#     return getCommandsInternal();
//# }
//#endif

//#ifdef TEST_MODE
//# public Vector getTextFields() {
//#     return new Vector();
//# }
//#endif

//#ifdef TEST_MODE
//# public void simulateKeySinglePress(int keyCode) {
//#     keyPressed(keyCode);
//# }
//#endif

    private Vector getCommandsInternal() {
        Vector res = new Vector();
        for(int i=0;i<items.size();++i) {
            MenuItem item = (MenuItem) items.elementAt(i);
            if (item.isSubMenu()) {
                // TODO: We should recurse here
                Menu subMenu = item.getSubMenu();
                for (int j=0;j<subMenu.totalNumberOfItems();++j) {
                    MenuItem subItem = subMenu.getMenuItem(j);
                    if (!subItem.isSeparator() && !subItem.isSubMenu()) {
                        Command command = subItem.getCommand();
                        res.addElement(command);
                    }
                }
            } else if (!item.isSeparator()) {
                Command command = item.getCommand();
                res.addElement(command);
            }
        }
        return res;
    }

}
