/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.ui.view.components;

import com.funambol.mailclient.ui.view.*;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.utils.UiUtils;
import com.funambol.util.Log;
import javax.microedition.lcdui.Graphics;

/**
 * This class represents an EditField in the Funambol low level UI. Such an item
 * can be added to Canvas to show a field that the user can edit. The class is
 * just responsible for painting the content and it does not handle events,
 * which are processed by the container.
 */
public class BodyItem extends EditField {

    private static final int FRAME_WIDTH = 4;
    private static final int FRAME_HEIGHT = 4;
    private int totalRowNumber = 0;
    private int visibleRowNumber = 0;
    private int firstVisibleRowIndex = 0;
    private int lastVisibleRowIndex = 0;
    private String[] bodyContent = null;
        
    
    private ScrollBar scrollBar = null;
    private boolean isScrollBarEnabled = false;
    private int sbX = 0;
    private int sbY = 0;
    private int sbW = 0;
    private int sbH = 0; 
    
    public BodyItem(Drawer d, GraphicalTheme t) {
        super(d, t);
    }
    
    public BodyItem(String label, String text, Drawer d, GraphicalTheme t) {
        super(label, text, d, t);
    }
    
    public BodyItem(String[] label, String[] text, Drawer d, GraphicalTheme t) {
        super(label, text, d, t);
    }

    public boolean isScrollable() {
        
        if (bodyContent!=null) {
            return totalItemH / (bodyContent.length * textHeight) < 1;
        }
        return false;
    }

    public int getFirstVisibleRowIndex() {
        return firstVisibleRowIndex;
    }

    public void setFirstVisibleRowIndex(int index) {
        this.firstVisibleRowIndex = index;
        this.isChanged = true;
    }


    public void scrollDown(int amount) {
        while (amount>0) {
            if (!isLastLineReached()) {
                firstVisibleRowIndex ++;
                lastVisibleRowIndex ++;
            }
            amount--;
        }
        //firstVisibleRowIndex += amount;
        //lastVisibleRowIndex += amount;
        if (isScrollBarNeeded()) {
            scrollBar.updateBarPosition(firstVisibleRowIndex);
        }
    }

    public void scrollUp(int amount) {
        while (amount>0) {
            if (!isFirstLineReached()) {
                firstVisibleRowIndex --;
                lastVisibleRowIndex --;
            }
            amount--;
        }
        //firstVisibleRowIndex -= amount;
        //lastVisibleRowIndex -= amount;
        if (isScrollBarNeeded()) {
            scrollBar.updateBarPosition(firstVisibleRowIndex);
        }
    }

    public boolean isFirstLineReached() {
            
        return firstVisibleRowIndex == 0;
    }

    public boolean isLastLineReached() {
            
        return lastVisibleRowIndex == bodyContent.length;
    }

    public void enableScrollBar(boolean isScrollBarEnabled) {
        this.isScrollBarEnabled = isScrollBarEnabled;
    }
    
    
    private boolean isScrollBarNeeded() {
        return (isScrollBarEnabled && (totalRowNumber > visibleRowNumber) && isHighlighted());
    }
    
    protected void recalculate(int y, Graphics g) {

        //if the body content is not empty, draws the empty recall the
        //Content of the message body
        if (getItemContent().length()==0) {
            super.recalculate(y, g);
            totalItemH = screenHeight - y;
        } else {
            recalculateBodyTextY(y);

            totalItemH = screenHeight - y;

            //recalculates the indentation based on the text font
            indentX = textFont.stringWidth(" ");
            
            //The drawing of the bar must take into account
            //the frame width and height
            sbW = indentX;
            sbH = totalItemH - FRAME_HEIGHT;
            sbX = screenWidth - sbW - FRAME_WIDTH/2;
            sbY = y + FRAME_WIDTH/2;

            //The content of the body item is left and right indented
            //and is dependent by the presence of the scrollbar and the frame
            int bodyContentWidth = screenWidth - 2 * indentX - sbW - FRAME_WIDTH;
 
            //Recalculate the Strings array using the bodyContentWidth dimension
            String content = getItemContent();
            bodyContent = UiUtils.getStringArray(content, bodyContentWidth, textFont);
            
            //total rows number
            totalRowNumber = bodyContent.length;

            //visible rows number
            visibleRowNumber = (totalItemH - FRAME_WIDTH - FRAME_WIDTH/4) / textHeight;

            if (visibleRowNumber >= totalRowNumber) {
                visibleRowNumber = totalRowNumber;
            }

            lastVisibleRowIndex = visibleRowNumber + firstVisibleRowIndex;

            if (scrollBar == null) {
                scrollBar = new ScrollBar(sbX, sbY, sbW, sbH, totalRowNumber, visibleRowNumber);
            } else {
                scrollBar.updateScrollBar(sbX, sbY, sbW, sbH, totalRowNumber, visibleRowNumber);
            }


            if (isScrollBarNeeded()) {
                scrollBar.setVisible(isScrollBarEnabled);
            }
            
            setChanged(false);
        }
    }

    protected void drawItemBackGround(Graphics g, int x, int y, int w, int totalItemH) {
        //draw body item background if it is highlighted, 
        //leave it white otherwise
        if (this.isHighlighted() && isBackGroundNeeded) {
            d.drawHighlightedBodyItemBackground(g, x, y, w, totalItemH);
            //Draw the body highlighted background
            /*int[] color = gt.getBodyHighLightedTextColor();
            g.setColor(color[0], color[1], color[2]);
            g.fillRect(x, y, w, totalItemH);
            
            //Draw the body highlighted frame
            color = gt.getBodyHighLightedFrameColor();
            g.setColor(color[6], color[7], color[8]);
            g.drawRect(x+2, y+2, w - 5, totalItemH - 5);
            g.setColor(color[3], color[4], color[5]);
            g.drawRect(x+1, y+1, w - 3, totalItemH - 3);
            g.setColor(color[0], color[1], color[2]);
            g.drawRect(x, y, w - 1, totalItemH - 1);*/
        } else {
            g.setColor(255, 255, 255);
            g.fillRect(x, y, w, totalItemH);
        }
    }
    
    protected void drawItemContent(int i, Graphics g) {
        if (getItemContent().length() == 0) {
            super.drawItemContent(i, g);
        } else {
            //if doesn't need scroll reset the indexes
            if (!isScrollable()) {
                firstVisibleRowIndex = 0;
                lastVisibleRowIndex = bodyContent.length;
            }
            
            int row = 0;
            for (int j = firstVisibleRowIndex; j < lastVisibleRowIndex; j++) {
                d.drawCustomItemTextString(bodyContent[j], g, indentX + FRAME_WIDTH/2,
                                           textAfterLabelY[0] + row * textHeight, layout,
                                           textFont, gt.getForegroundColor());
                row++;
            }
            
            if (isScrollBarNeeded()) {
                scrollBar.updateBarPosition(firstVisibleRowIndex);
                scrollBar.setVisible(true);
                scrollBar.paint(g);
            }
        }
    }

    private void recalculateBodyTextY(int y) {
        textHeight = textFont.getHeight();
        deltaFrame = textHeight / labelRatio;
        if (labelY == null) {
            labelY = new int[1];
        }
        labelY[0] = y + deltaFrame;

        if (textAfterLabelY == null) {
            textAfterLabelY = new int[1];
        }
        textAfterLabelY[0] = labelY[0] + textHeight;
    }
}
