/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2009 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 *
 */

package com.funambol.mailclient.ui.view.components;

import com.funambol.mailclient.ui.view.CustomMenuCanvas;
import com.funambol.mailclient.ui.view.GraphicalTheme;
import com.funambol.mailclient.ui.view.GraphicalThemeFactory;
import com.funambol.mailclient.ui.view.FTouchEvent;
import com.funambol.mailclient.ui.view.ItemStateListener;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.util.Log;

import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.Displayable;

import java.util.Vector;

/**
 * <code>Form</code> represents a container of <code>Item</code> objects. It
 * allows to add and remove items through the <code>append</code> and
 * <code>delete</code> methods.
 *
 * The title is showed through the <code>Title</code> component.
 */
public abstract class Form extends CustomMenuCanvas 
        implements ItemStateListener, CommandListener {

    // The Vector of Item objects
    private Vector items = null;

    // Keep track of the current selected item
    private Item currentItem = null;

    // The Title
    private Title title = null;

    // The ScrollBar
    private ScrollBar scrollBar = null;

    // The GraphicalTheme
    GraphicalTheme theme = GraphicalThemeFactory.getGraphicalTheme();

    /**
     * Create a new Form given the title.
     * @param title
     */
    public Form(String title) {
        super();
        this.items = new Vector();
        this.title = new Title(theme, title, null, getWidth());
    }

    /**
     * Appends a new <code>Item</code> to the <code>Form</code>.
     * 
     * @param item The <code>Item</code> object.
     * @throws NullPointerException If the given item is null.
     */
    public void append(Item item) throws NullPointerException {
        items.addElement(item);
        if(currentItem == null) {
            activateItem(item);
        }
    }

    /**
     * Inserts a new <code>Item</code> to the <code>Form</code> at
     * the given position.
     * 
     * @param item The <code>Item</code> object.
     * @param position The position within the form.
     * @throws NullPointerException If the given item is null.
     * @throws IndexOutOfBoundsException If position is invalid.
     */
    public void insert(Item item, int position)
            throws NullPointerException, IndexOutOfBoundsException {
        items.insertElementAt(item, position);
    }

    /**
     * Deletes an item at the given position.
     *
     * @param position The position within the form.
     * @throws IndexOutOfBoundsException If position is invalid.
     */
    public void delete(int position) throws IndexOutOfBoundsException {
        items.removeElementAt(position);
    }

    /**
     * Get the <code>Item</code> from the given poition.
     * 
     * @param position The position within the form.
     * @return The <code>Item</code> object.
     * @throws IndexOutOfBoundsException If position is invalid.
     */
    public Item get(int position) throws IndexOutOfBoundsException {
        return (Item)items.elementAt(position);
    }

    /**
     * Get the <code>Item</code> from the given poition.
     *
     * @param position The position within the form.
     * @param item The <code>Item</code> object.
     * @throws IndexOutOfBoundsException If position is invalid.
     * @throws NullPointerException If the given item is null.
     */
    public void set(int position, Item item)
            throws NullPointerException, IndexOutOfBoundsException {
        items.setElementAt(item, position);
    }

    /**
     * Get the number of the items contained in the form.
     * @return The item count.
     */
    public int size() {
        return items.size();
    }

    /**
     * Activates the given item.
     * 
     * @param item An existing Item
     */
    public void activateItem(Item item) {
        if(currentItem != null) {
            currentItem.setHighlighted(false);
            removeItemCommands(currentItem.getCommands());
        }
        // Set the new current item
        currentItem = item;

        currentItem.setHighlighted(true);
        addItemCommands(currentItem.getCommands());

        // Update the subtitle
        title.setSubTitle(currentItem.getToolTipText());
    }

    /**
     * Activates the item the given y coordinate.
     *
     * @param y The y coordinate.
     */
    private void activateItem(int y) {
        Item item = null;
        for (int i = 0; i<items.size(); i++) {
            item = (Item) items.elementAt(i);
            int startCoordinate = item.getY();
            if (y >= startCoordinate && y <= startCoordinate + item.getHeight()) {
                activateItem(item);
            }
        }
    }

    /**
     * Handles the given key code.
     * @param keyCode The key code.
     */
    protected void handleKeyPressed(int keyCode) {
        int dir = getGameAction(keyCode);
        if (isFireKey(dir, keyCode)) {
            fireCommand();
        } else if (isDownKey(dir, keyCode)) {
            handleScrollUpEvent();
        } else if (isUpKey(dir, keyCode)) {
            handleScrollDownEvent();
        }
    }

    /**
     * Gets the next item.
     */
    private void getNextItem() {
        int currentIndex = items.indexOf(currentItem);
        if (currentIndex < items.size()-1) {
            activateItem((Item)items.elementAt(currentIndex+1));
        }
    }

    /**
     * Gets the previous item.
     */
    private void getPreviousItem() {
        int currentIndex = items.indexOf(currentItem);
        if (currentIndex > 0) {
            activateItem((Item)items.elementAt(currentIndex-1));
        }
    }

    /**
     * Handles a <code>FTouchEvent</code> event.
     * @param event The <code>FTouchEvent</code> event.
     */
    protected void handleTouchEvent(FTouchEvent event) {
        int type     = event.getType();
        int tapCount = event.getTapCount();

        switch (type) {
            case FTouchEvent.PRESSED:
                int y = event.getPressY(0);
                if (y >= 0) {
                    //activate the item drawn on that y
                    activateItem(y);
                }
                repaint();
                break;
            case FTouchEvent.CLICK:
                // Now we may need to open the msg
                if (tapCount > 1) {
                    // Multi taps are considered as double tap
                    //fireCommand();
                }
                break;
            case FTouchEvent.DRAGGING_DOWN:
                handleScrollDownEvent();
                break;
            case FTouchEvent.DRAGGING_UP:
                handleScrollUpEvent();
                break;
            default:
                break;
        }
    }

    /**
     * Handles the scrool down event.
     */
    private void handleScrollDownEvent() {
        //Implement here the scroll logic for the body item
        getPreviousItem();
        repaint();
    }

    /**
     * Handles the scrool up event.
     */
    private void handleScrollUpEvent() {
        //Implement the scroll logic for the body item
        getNextItem();
        repaint();
    }

    /**
     * Paint the form content.
     * @param graphics
     */
    protected void paintContent(Graphics graphics) {

        int width = getWidth();
        int height = getHeight();

        UIController.getDrawer().drawCanvasBackground(graphics, width, height);

        //Paint Title
        int y = title.paint(graphics, width, height);

        for (int i = 0; i < items.size(); i++) {
            Item item = (Item) items.elementAt(i);

            //Needed by tiltable screens devices to manage the
            //width and height
            item.setScreenWidth(width);
            item.setScreenHeight(height);

            //Needs for the touch screen devices to understand
            //Where the item is initially painted
            item.setY(y+1);
            item.invalidate();
            item.setVisible(true);
            if (item.isVisible()) {
                y = item.paint(graphics, 0, y, width);
            }
        }
    }
    
    /**
     * Handle the key repeated event.
     * @param keyCode
     */
    protected void handleKeyRepeated(int keyCode) {
        keyPressed(keyCode);
    }

    /**
     * Removes the given commands
     * @param commands
     */
    private void removeItemCommands(Vector commands) {
        for(int i=0; i<commands.size(); i++) {
            removeCommand((Command)commands.elementAt(i));
        }
    }

    /**
     * Adds the given commands
     * @param commands
     */
    private void addItemCommands(Vector commands) {
        for(int i=0; i<commands.size(); i++) {
            addCommand((Command)commands.elementAt(i));
        }
    }

    /**
     * @see CommandListener#commandAction(javax.microedition.lcdui.Command,
     * javax.microedition.lcdui.Displayable)
     */
    public void commandAction(Command command, Displayable d) {
        // Forward the command to the currentItem
        currentItem.commandAction(command, this);
    }

    /**
     * @see ItemStateListener#notifyChanged() a
     */
    public void notifyChanged() {
        // TBD
    }

    /**
     * Handles fire command
     */
    private void fireCommand() {
        // TBD
    }
}
