/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.ui.view.components;

import com.funambol.mailclient.ui.view.ItemStateListener;

import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;

import java.util.Vector;

/**
 * Represents an abstact type of item that can be used within a
 * <code>Form</code> component.
 */
public abstract class Item implements CommandListener {

    protected boolean isVisible     = false;
    protected boolean isHighlighted = false;
    protected boolean isLighter = false;
    protected boolean isChanged     = false;

    private String  toolTipText = null;
    private Vector  commands    = null;

    protected int screenHeight = 0;
    protected int screenWidth  = 0;
    private int itemy          = 0;

    public ItemStateListener stateListener = null;

    /**
     * Create a new <code>Item</code>
     */
    public Item() {
        commands = new Vector();
    }

    /**
     * Create a new <code>Item</code> given the tool tip text.
     *
     * @param toolTipText the tool tip text.
     */
    public Item(String toolTipText) {
        this();
        setToolTipText(toolTipText);
    }

    /**
     * Set the tool tip text associated to this item.
     * @param text
     */
    public void setToolTipText(String text) {
        this.toolTipText = text;
    }

    /**
     * Get the tool tip text associated to this item.
     * @return
     */
    public String getToolTipText() {
        return this.toolTipText;
    }

    /**
     * The the commands <code>Vector</code>
     * @return The <code>Vector</code> containing <code>Command</code> items.
     */
    public Vector getCommands() {
        return commands;
    }

    /**
     * Add a new command
     * @param command The new command
     */
    public void addCommand(Command command) {
        commands.addElement(command);
    }

    /**
     * Set this item visibility.
     * @param isVisible True if the item shall be visible.
     */
    public void setVisible(boolean isVisible) {
        this.isVisible = isVisible;
    }

    /**
     * Check whether this item is visible or not.
     * @return 
     */
    public boolean isVisible() {
        return this.isVisible;
    }

    /**
     * Set the highlight property. This may cause the item to be drawed
     * differently.
     *
     * @param isHighLighted True if this item shall be highlighted
     */
    public void setHighlighted(boolean isHighlighted) {
        this.isHighlighted = isHighlighted;
    }

    /**
     * Check whether this item is highlighted or not.
     * @return
     */
    public boolean isHighlighted() {
        return this.isHighlighted;
    }

    /**
     * Set the lighter property. This causes the item to be drawn with lighter
     * colors
     *
     * @param isLighter True if this item shall be lighter
     */
    public void setLighter(boolean isLighter) {
        this.isLighter = isLighter;
    }

    /**
     * Check whether this item is lighter or not.
     * @return boolean true if the item is selected to be lighter gradient drawn
     */
    public boolean isLighter() {
        return this.isLighter;
    }

    /**
     * Invalidates the item view
     */
    protected void invalidate() {
        isChanged = true;
    }

    /**
     * Set this item y coordinate.
     * @param y
     */
    public void setY(int y) {
        this.itemy = y;
    }

    /**
     * Get the item y coordinate.
     * 
     * @return
     */
    public int getY() {
        return this.itemy;
    }

    /**
     * Set the container screen height.
     * @param screenHeight
     */
    public void setScreenHeight(int screenHeight) {
        if (this.screenHeight != screenHeight) {
            invalidate();
        }
        this.screenHeight = screenHeight;
    }

    /**
     * Set the container screen width.
     * @param screenWidth
     */
    public void setScreenWidth(int screenWidth) {
        if (this.screenWidth != screenWidth) {
            invalidate();
        }
        this.screenWidth = screenWidth;
    }

    /**
     * Sets the current <code>ItemStateListener</code>.
     * @param stateListener The <code>ItemStateListener</code> reference.
     */
    public void setItemStateListener(ItemStateListener stateListener) {
        this.stateListener = stateListener;
    }
    
    // ------------------------ Abstract methods -----------------------------//
    
    /**
     * Paint the item content.
     * 
     * @param g The item <code>Graphics</code>
     * @param x The x coordinate.
     * @param y The y coordinate.
     * @param w The item width.
     * 
     * @return The item height.
     */
    protected abstract int paint(Graphics g, int x, int y, int w);

    /**
     * Gets the total item height.
     * @return The item height.
     */
    public abstract int getHeight();

    /**
     * Handle the given <code>Command<code>
     * @param command
     * @param displayable
     */
    public abstract void commandAction(Command command, Displayable displayable);
}
