/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient;

import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mail.Store;
import com.funambol.mail.Message;
import com.funambol.mailclient.sm.SyncClient;
import com.funambol.storage.ObjectStore;
import com.funambol.storage.AbstractRecordStore;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.util.Log;

import java.io.IOException;

import javax.microedition.rms.RecordStoreException;

//#ifdef isBlackberry
//# import com.funambol.util.BlackberryUtils;
//#endif

/**
 * The UpgradeManager is responsible of handling the upgrade operations.
 */
public class UpgradeManager {

    /**
     * This method handles the upgrade the client using the information
     * given by the given <code>MailClientConfig</code>.
     * 
     * @param newClientConfig
     * @return True if the upgrade has already started a sync process.
     */
    public static boolean handleUpgrade(MailClientConfig clientConfig) {
        int currentVersion = clientConfig.getStoreVersion();
        Log.debug("[UpgradeManager.handleUpgrade] Current store version: " +
                currentVersion);
        if (currentVersion < Store.VERSION_102) {
            return handleUpgradeForStore102(clientConfig);
        } else if (currentVersion < Store.VERSION_105) {
            return handleUpgradeForStore105(clientConfig);
        }
        return false;
    }

    // Handle upgrade when the Store version is older than 102
    private static boolean handleUpgradeForStore102(MailClientConfig clientConfig) {
        Log.debug("[UpgradeManager.handleUpgradeForStore102] Upgrading for " +
                "store version 102");
        String msg = Localization.getMessages().UPGRADE_REQUIRED;
        UIController.showModalAlert(msg);
        // Reset inbox
        try {
            // This is a non blocking operation, therefore the method exits
            // immediately. If the update is not performed properly (for
            // example because of no network connectivity) then the user is
            // left without any message.
            UIController.resetAccount();
            clientConfig.setStoreVersion(Store.LATEST_VERSION);
            ConfigManager.saveConfig();
            // A sync process has been started
            return true;
        } catch (Exception e) {
            Log.error("[UpgradeManager.handleUpgradeForStore102] Error while " +
                    "resetting account: " + e.toString());
            return false;
        }
    }

    // Handle upgrade when the Store version is older than 105
    private static boolean handleUpgradeForStore105(MailClientConfig clientConfig) {

        Log.debug("[UpgradeManager.handleUpgradeForStore105] Upgrading for " +
                "store version 105");

        UIController.midlet.setSplashPhase("Upgrading client...", 0);

        // The store still doesn't support Multiple Email Account

        // Create default account structure
        MailAccount.createDefaultAccount();

        // Upgrade Message items (add prefix character) and migrate them to the
        // new account folder structure. The old folders are removed
        // definitively.
        try {
            AbstractRecordStore.deleteRecordStore("/Inbox");
        } catch(RecordStoreException ex) {
            Log.error("[UpgradeManager.handleUpgradeForStore105] Unable to " +
                        "remove Inbox folder: " + ex.getMessage());
        }
        migrateFolder("/Outbox", MailAccount.DEFAULT_OUTBOX_PATH);
        migrateFolder("/Drafts", MailAccount.DEFAULT_DRAFTS_PATH);
        migrateFolder("/Sent", MailAccount.DEFAULT_SENT_PATH);

        // Make sure the log level is set to INFO while upgrading client
        clientConfig.setLogLevel(Log.DISABLED);
        clientConfig.setStoreVersion(Store.LATEST_VERSION);

        try {
            ConfigManager.saveConfig();
        } catch (Exception e) {
            Log.error("[UpgradeManager.handleUpgradeForStore105] Error while " +
                    "saving configuration: " + e.toString());
            return false;
        }

        // Reset source anchors to avoid a fast sync
        try {
            UserAccount a = ConfigManager.getConfig().getUserAccount();
            SourceConfig srcCfg = SyncClient.getSyncClient().loadSourceConfig(
                    SourceConfig.MAIL, SourceConfig.EMAIL_OBJECT_TYPE, a.getRemoteURI());

            srcCfg.setLastAnchor(0);
            srcCfg.setNextAnchor(0);

            SyncClient.getSyncClient().saveSourceConfig();

        } catch(Exception ex) {
            Log.error("[UpgradeManager.handleUpgradeForStore105] Unable to " +
                    "reset anchors: " + ex.getMessage());
            ex.printStackTrace();
        }
        
        // Start the Inbox reset
        try {
            //#ifdef isBlackberry
            //#
            //# // On BlackBerry devices the upgrade is done after reboot. It is
            //# // possible that the upgrade task is executed before the data
            //# // connection is turned on. It means that the reset inbox could 
            //# // not start as expected. The only way to avoid this is to check
            //# // the network connection as follows.
            //#
            //# int retries = 15;
            //# while(!BlackberryUtils.isDataConnectionAvailable() && retries > 0) {
            //#     Thread.sleep(2000);
            //#     retries--;
            //# }
            //#endif
            UIController.resetInboxFromServer();
        } catch (Exception e) {
            Log.error("[UpgradeManager.handleUpgradeForStore105] Unable to " +
                    "reset inbox: " + e.getMessage());
            e.printStackTrace();
        }
        return true;
    }

    /**
     * Migrate Message items from an old folder to the new one, the Messages
     * from the old folder are read without the prefix char, that will be
     * written in the new folder.
     */
    private static void migrateFolder(String oldFolder, String newFolder) {
        ObjectStore objs = new ObjectStore();
        Message message = new Message();
        message.setLaziness(Message.NO_LAZY);
        try {
            objs.open(oldFolder);
            int count = objs.size();
            for (int i=0; i<count; i++) {
                int firstIndex = objs.getFirstIndex();
                objs.retrieve(firstIndex, message);
                objs.close();

                objs.open(newFolder);
                objs.store(message);
                objs.close();
                
                objs.open(oldFolder);
                objs.remove(firstIndex);
            }
            objs.close();
            AbstractRecordStore.deleteRecordStore(oldFolder);
        } catch(RecordStoreException ex) {
            Log.debug("Error while migrating " + oldFolder + " folder. Exception: "
                    + ex.toString());
        } catch(IOException ex) {
            Log.debug("Error while reading message in " + oldFolder + " folder." +
                    " Exception: " + ex.toString());
        }
    }
}

