/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.mm;

import java.util.Vector;
import java.util.Hashtable;

import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;

import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreNotFoundException;
import javax.microedition.rms.InvalidRecordIDException;
import javax.microedition.rms.RecordStore;

import com.funambol.platform.FileAdapter;
import com.funambol.mailclient.syncml.EmailData;
import com.funambol.mail.Message;
import com.funambol.mail.BodyPart;
import com.funambol.mail.Multipart;
import com.funambol.storage.ComplexSerializer;
import com.funambol.storage.Serializable;

import com.funambol.util.Base64;
import com.funambol.util.XmlUtil;
import com.funambol.util.Log;

/**
 * This class represents a store for pre-formatted emails. Preformatted emails
 * are items ready to be synchronized over SyncML. The mail sync source before
 * sending an email checks if its preformatted version is available in the store
 * and in this case such a version is used. The application at the moment preformats
 * only emails with attachments.
 * This class does not require that an email is actually completely
 * preformatted. For example the application handles attachments as detached and
 * when an email is preformatted, its attachments are only "linked". The
 * handling of this feature is delegated to the detached mime formatter and the
 * FormattedEmailInputStream.
 * Since a preformatted email does not store any attachment, the size of stored
 * items is rather small and the implementation uses a RecordStore. Even on
 * platforms where RecordStore is limited to 64Kb the space is eough for a large
 * number of items.
 */
public class FormattedEmailStore {
    
    private static FormattedEmailStore instance = null;
    private static final String storeName = "FormattedEmailStore";
    private RecordStore rmsStore  = null;
    private Vector messageRecords = new Vector();
    private Hashtable metaInfo    = null;
    private Hashtable streamsStore = new Hashtable();
    private int       emailSize   = 0;

    /**
     * This class represents the input stream that can be used to read a
     * preformatted email. This component is responsible for getting the
     * bytes of the email from where they are stored (the text body part and
     * the headers are in the formatted store, while the attachments are
     * left in their original files)
     */
    private class FormattedEmailInputStream extends InputStream {

        // Do not change this chunk size, as its value depends on the MIME
        // requirement of having lines no longer than 76 characters
        private static final int CHUNK_SIZE = 74;
        private static final int ENCODED_CHUNK_SIZE = 54;

        private static final char LF = '\n';

        private FormattedEmail email;

        private int totalSize;
        private int idx;
        private int formatIdx;
        private int attIdx;
        private int nextAttachmentIdx = -1;
        private boolean lastChunk;
        private boolean readingAttStream = false;
        private byte buf[];
        private int bufIdx = 0;
        private InputStream attIs;
        private byte format[];
        private BodyPart currentBp;
        private long currentFileSize = 0;
        private long currentFileIdx  = 0;

        public FormattedEmailInputStream(FormattedEmail email) {

            this.email = email;
            format = email.getPreformattedValue();

            int numParts = email.getNumberOfAttachmentParts();
            for(int i=0;i<numParts;++i) {
                BodyPart bp = email.getBodyPart(i);
                // We need the size of all the attachments but the first one
                // which is already included in the format length
                String sizeStr = bp.getSize();
                int attSize = -1;
                if (sizeStr != null) {
                    try {
                        attSize = Integer.parseInt(sizeStr);
                    } catch (Exception e) {
                        Log.error("Invalid attachment size: " + sizeStr);
                    }
                }
                if (attSize == -1) {
                    // We need to compute this attachment size
                    attSize = computeSize(bp);
                }
                int encSize = Base64.computeEncodedSize(attSize);
                totalSize += encSize;

                // every 72 bytes we insert a line break 
                int numBreaks = encSize / 72;
                totalSize += numBreaks;
            }
            if (format != null) {
                totalSize += format.length;
            }

            // Compute the next attachment index
            if (attIdx < email.getNumberOfAttachmentParts()) {
                nextAttachmentIdx = email.getAttachmentPosition(attIdx);
            }
        }

        public int read(byte[] data) throws IOException {
            int len = data.length;

            for(int i=0;i<len;++i) {
                int ch = read();
                if (ch != -1) {
                    data[i] = (byte)ch;
                } else {
                    return i;
                }
            }
            return len;
        }

        /**
         * Read one byte from the stream.
         */
        public int read() throws IOException {
            // We start reading the format stream and wait for the next
            // attachment to begin
            if (idx < totalSize) {
                if (readingAttStream) {
                    if (bufIdx == buf.length) {
                        if (lastChunk) {
                            readingAttStream = false;
                            nextAttachmentIdx = -1;
                            try {
                                attIs.close();
                            } catch (IOException ioe) {
                                Log.error("Cannot close output stream " + ioe.toString());
                            }
                            attIs = null;
                            // Prepare for the next attachment
                            if (attIdx < email.getNumberOfAttachmentParts()) {
                                nextAttachmentIdx = email.getAttachmentPosition(attIdx);
                            } else {
                                nextAttachmentIdx = -1;
                            }
                        } else {
                            // We must read the next chunk
                            buf = readNextChunk(attIs, true);
                            bufIdx = 0;
                            // Return the first byte
                            idx++;
                            return buf[bufIdx++];
                        }
                    } else {
                        idx++;
                        return buf[bufIdx++];
                    }
                } 

                if (formatIdx == nextAttachmentIdx) {
                    // We shall start reading the next attachment here
                    currentBp = email.getBodyPart(attIdx);

                    if (currentBp != null) {
                        if (currentBp instanceof InputStreamBodyPart) {
                            InputStreamBodyPart isbp = (InputStreamBodyPart)currentBp;
                            attIs = isbp.getInputStream();
                            // Compute the size
                            String sizeStr = currentBp.getSize();
                            int attSize = -1;
                            if (sizeStr != null) {
                                try {
                                    attSize = Integer.parseInt(sizeStr);
                                } catch (Exception e) {
                                    Log.error("Invalid attachment size: " + sizeStr);
                                }
                            }
                            if (attSize == -1) {
                                // We need to compute this attachment size
                                attSize = computeSize(currentBp);
                            }
                            currentFileSize = attSize;
                            // We invalidate the input stream here
                            // because if the sending fails after some
                            // bytes were read, the stream is not usable any
                            // longer
                            isbp.setInputStream(null);
                        }

                        if (attIs == null) {
                            // We must reopen the stream
                            String fileName = "file:///" + currentBp.getAttachUrl() + currentBp.getFileName();
                            FileAdapter fa = new FileAdapter(fileName, true);
                            try {
                                attIs = fa.openInputStream();
                                currentFileSize = fa.getSize();
                            } catch (IOException ioe) {
                                Log.error("Cannot open " + fileName);
                                Log.error("This message will not be sent");
                                throw ioe;
                            }
                        }

                        currentFileIdx  = 0;
                    } else {
                        Log.error("Cannot find body part in attachments store ");
                        throw new IOException("Cannot read attachment part");
                    }
                    readingAttStream = true;
                    attIdx++;
                    bufIdx = 0;
                    // Read the next chunk
                    buf = readNextChunk(attIs, true);
                    idx++;
                    return buf[bufIdx++];
                } else {
                    // We are not in an attachment, nor ready to start one
                    idx++;
                    return (int)format[formatIdx++];
                }
            } else {
                return -1;
            }
        }

        public int available() {
            return totalSize - idx;
        }

        public void close() throws IOException {
            if (attIs != null) {
                attIs.close();
            }
            StringBuffer msg = new StringBuffer("Closing FormattedEmailInputStream (");
            msg.append(idx).append("/").append(totalSize).append(")");
            Log.trace(msg.toString());
        }

        private int computeSize(BodyPart bp)  {
            String path = bp.getAttachUrl();
            String file = bp.getFileName();

            String fileName = "file:///" + path + file;
            int attSize = 0;
            try {
                FileAdapter fa = new FileAdapter(fileName, true);
                attSize = (int)fa.getSize();
                fa.close();
            } catch(IOException ioe) {
                Log.error("Cannot compute size for: " + fileName);
            }
            return attSize;
        }

        private byte[] readNextChunk(InputStream is, boolean encode) throws IOException {
            byte dataBuf[];
            long available = currentFileSize - currentFileIdx;
            int chunkSize;
            if (encode) {
                chunkSize = ENCODED_CHUNK_SIZE;
            } else {
                chunkSize = CHUNK_SIZE;
            }
            if (available > chunkSize) {
                dataBuf = new byte[chunkSize];
                lastChunk = false;
            } else {
                dataBuf = new byte[(int)available];
                lastChunk = true;
            }
            is.read(dataBuf);
            currentFileIdx += dataBuf.length;

            if (encode) {
                byte[] encodedBuf = Base64.encode(dataBuf);
                dataBuf = encodedBuf;
            }
            // Add a LINE_BREAK at the end of the chunk if this is not the last one
            if (!lastChunk) {
                byte[] tempBuf = new byte[dataBuf.length + 1];
                System.arraycopy(dataBuf, 0, tempBuf, 0, dataBuf.length);
                // We cannot encode line breaks as CRLF , as this will trigger
                // Funambol's server normalization which transform single CR
                // into CR LF and in case of chunks ending on the CR will change
                // the item content
                tempBuf[dataBuf.length] = LF;
                dataBuf = tempBuf;
            }

            return dataBuf;
        }
    }

    /**
     * Default constructor.
     */
    private FormattedEmailStore() {
        try {
            rmsStore = RecordStore.openRecordStore(storeName, true);
            if (rmsStore.getNumRecords() == 0) {
                metaInfo = new Hashtable();
            } else {
                byte data[] = rmsStore.getRecord(1);
                ByteArrayInputStream bais = new ByteArrayInputStream(data);
                DataInputStream din = new DataInputStream(bais);
                metaInfo = (Hashtable) ComplexSerializer.deserializeObject(din);
            }
            if (rmsStore.getNumRecords() == 0) {
                updateMetaInfo();
            }
        } catch (RecordStoreException rse) {
            Log.error("Cannot open record store " + rse.toString());
            metaInfo = new Hashtable();
        } catch (IOException ioe) {
            Log.error("Cannot read from record store " + ioe.toString());
            metaInfo = new Hashtable();
        }
    }

    /**
     * Get the singleton.
     */
    public synchronized static FormattedEmailStore getInstance() {
        if (instance == null) {
            instance = new FormattedEmailStore();
        }
        return instance;
    }

    /**
     * Add a message to the store. The message is preformatted and links to its
     * attachments are created.
     */
    public synchronized void addMessage(Message message) throws IOException {
        if (rmsStore == null) {
            throw new IOException("Cannot access record store " + storeName);
        }
        try {
            FormattedEmail email = new FormattedEmail();
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            MIMEFormatterDetachedAttachments formatter = new MIMEFormatterDetachedAttachments(email);
            EmailData data = new EmailData(message, formatter);
            data.format(os, true);
            os.close();
            // Now update the formatted email
            email.setPreformattedValue(os.toByteArray());
            // Update the meta info
            metaInfo.put(message.getKey(), email);
            updateMetaInfo();
        } catch (RecordStoreException rmse) {
            throw new IOException(rmse.toString());
        }
    }

    public synchronized void removeMessage(String msgKey) throws IOException {
        Object oldKey = metaInfo.remove(msgKey);
        if (oldKey != null) {
            try {
                updateMetaInfo();
            } catch (RecordStoreException rse) {
                throw new IOException(rse.toString());
            }
        }
    }

    /**
     * Gets an input stream of a preformatted message.
     * @param msgKey is the message key
     */
    public InputStream getMessageInputStream(String msgKey) throws IOException {

        if (rmsStore == null) {
            throw new IOException("Cannot access record store " + storeName);
        }

        try {
            if (rmsStore.getNumRecords() == 0) {
                throw new IOException("Message not found");
            }
            FormattedEmail email = (FormattedEmail)metaInfo.get(msgKey);
            if (email == null) {
                throw new IOException("Message not found");
            }

            FormattedEmailInputStream is = new FormattedEmailInputStream(email);
            return is;
        } catch (RecordStoreException rmse) {
            throw new IOException(rmse.toString());
        }
    }

    /**
     * Checks if the given message is present in the store.
     */
    public boolean contains(Message msg) {
        if (metaInfo != null && metaInfo.get(msg.getKey()) != null) {
            return true;
        } else {
            return false;
        }
    }

    private void updateMetaInfo() throws RecordStoreException, IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        DataOutputStream dout = new DataOutputStream(baos);
        ComplexSerializer.serializeObject(dout, metaInfo);
        byte data[] = baos.toByteArray();
        if (rmsStore.getNumRecords() == 0) {
            rmsStore.addRecord(data, 0, data.length);
        } else {
            rmsStore.setRecord(1, data, 0, data.length);
        }
    }
}
    
