/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.mm;

import com.funambol.mail.MessageFlags;
import com.funambol.mailclient.syncml.FlagQueue;

import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;

import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.syncml.FolderData;
import com.funambol.mail.Message;
import com.funambol.mail.Folder;
import com.funambol.mail.AccountFolder;
import com.funambol.mail.Store;
import com.funambol.mail.StoreFactory;
import com.funambol.mail.ChildrenEnumeration;
import com.funambol.mail.MailException;
import com.funambol.mailclient.MailAccount;

import com.funambol.mailclient.syncml.EmailData;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.util.Log;
import com.funambol.util.XmlException;
import com.funambol.util.ChunkedString;
import javax.microedition.rms.RecordStoreException;

/**
 * This class is used to manage local messages.The MessageManager is responsible
 * for adding/updating/deleting messages. All these operations can be performed
 * on the client and then propagated to the server, or originated on the server
 * and propagated to the client.
 * Message (@see Message)is the basic entity to represent an email. Message is
 * handled almost completely by the mail API, but there is one information in
 * Message which is left to the client to manage. The key is a concept for the
 * client and Messages created by the API leave it empty.
 * The MessageManager is responsible for creating all the Messages in the
 * client. Message should never be created without using the MessageManager. In
 * this context the MessageManager acts as a Message factory for the client.
 * Any time the MessageManager creates a Message it sets the key to a proper
 * value, guaranteeing that no Message is without key. The key is used to
 * save/load messages, therefore it is crucial that each Message has this
 * information properly set.
 * From a message key it is always possible to derive both the message folder
 * and the message record id in the underlying store.
 * Other parts of the client should not make any assumption on how the LUID is
 * formed.
 * The MessageManager has three different set of methods to handle messages:
 * 
 * 1) generic methods to operate on the client messages
 * 2) methods invoked to create/update/delete messages upon a user action. In
 *    these cases the UI notifies that something changed and this information
 *    must be sent to the server on the next synchronization
 * 3) methods invoked to create/update/delete messages upon a server request. In
 *    these cases the information in the system (e.g. RMS) is updated. The
 *    MessageManager does not call back any method, because notifications to the
 *    UI are sent via SyncListener by the SyncManager.
 *
 */
public class MessageManager {

    private static Store msgStore = null;
    private static MessageManager instance = null;

    private static final String LOG_TAG = "MessageManager";

    /**
     * Represents a cache of the folders vector related to a specific account.
     *
     * Use the account fullname as key.
     * The values type is a Vector of Folder objects.
     */
    private Hashtable foldersByAccount = new Hashtable();

    /**
     * Represents a cache of the number of unread messages of a specific folder.
     *
     * Use the folder fullname as key.
     * The values type is Integer.
     */
    private Hashtable unreadMessagesCountByFolder = new Hashtable();

    /**
     * The listener to be notified when the number of unread messages has changed
     */
    private UnreadMessagesListener unreadMessageListener = null;

    /**
     * This is a cache of the accounts available. It is computed by getAccounts
     * and is invalidated by any method which add/delete an account
     */
    private AccountFolder accountFolders[] = null;

    /**
     * Default constructor.
     */
    private MessageManager() {
        try {
            msgStore = StoreFactory.getStore();
        } catch (Exception e) {
            Log.error(LOG_TAG, "Cannot get store: " + e.toString());
            msgStore = null;
        }
    }

    /**
     * Get the <code>MessageManager</code> instance.
     * @return The <code>MessageManager</code> instance.
     */
    public synchronized static MessageManager getInstance() {
        if (instance == null) {
            instance = new MessageManager();
        }
        return instance;
    }
    
    //---------------------------------------------------------- Public Methods

    /**
     * Initialize the mail the store.
     *
     * @param reset If true, erase and re-create the <code>Store</code>.
     *
     * @throws MailException in case of errors.
     */
    public void initStore(boolean reset) throws MailException {
        Log.trace(LOG_TAG, "initStore(" + reset + ")" + msgStore);
        synchronized (msgStore) {
            msgStore.init(reset);
            // clear queue of flags if re-initialization is needed
            if (reset) {
                foldersByAccount = new Hashtable();
                accountFolders = null;
                FlagQueue.clear();
                resetUnreadMessagesCache();
            }
        }
    }

    /**
     * Append a message to a folder. The message key is set after this
     * operation. The method throws an exception if the underlying folder
     * throws an exception.
     *
     * @param msg the message
     * @param folder the folder
     *
     * @throws MailException if the underlying folder throws an exception
     */
    public void appendMessageToFolder(Message msg, Folder folder) throws MailException {
        Log.trace(LOG_TAG, "appendMessageToFolder(Message, Folder)");
        synchronized (msgStore) {
            folder.appendMessage(msg);
            msg.setKey(makeLuid(msg));
            if (!msg.getFlags().isSet(MessageFlags.OPENED)) {
                incrementUnreadMessageCount(folder, 1);
            }
        }
    }

    /**
     * Creates a new Message from a SyncML byte stream. The message is also
     * saved in its folder.
     *
     * @param parentName parent folder name
     * @param emailData  byte stream of the SyncML email
     *
     * @return a new Message object with the proper key defined or
     *         null if either the folder is invalid or the byte stream cannot be
     *         parsed properly.
     *
     * @throws XmlException  if the byte stream cannot be parsed
     * @throws MailException if the message cannot be saved in the folder
     */
    public Message createNewMessage(String parentName, byte[] emailData)
    throws XmlException, MailException {

        Log.trace(LOG_TAG, "createNewMessage");

        String folderName = folderFromParent(parentName);

        if(folderName == null) {
            Log.error(LOG_TAG, "Invalid parent from server: " + parentName);
            throw new MailException(MailException.MESSAGE_ERROR,
                    "Invalid parent from server: " + parentName);
        }

        EmailData ed = new EmailData();
        Message msg = ed.parseMessage(emailData);

        if (msg != null) {
            String msgid = msg.getMessageId();
            Folder parent = getFolder(folderName, true);

            Log.debug(LOG_TAG, "Adding message " + msgid + " to folder " + parent.getName());
            appendMessageToFolder(msg, parent);

            // We set the luid to be more information than just the recordId
            // Our LUID is "folder"/"recordid"
            msg.setKey(makeLuid(msg));
        }
        return msg;
    }

    /**
     * Updates a message which is already stored in a folder (the folder
     * information must be valid)
     *
     * @param msg the message
     * @throws MailException if the underlying folder throws an exception
     */
    public void updateMessage(Message msg) throws MailException {
        Log.trace(LOG_TAG, "updateMessage(Message)");
        synchronized (msgStore) {
            Folder parent = msg.getParent();
            Message current = getMessage(msg.getKey());
            if (msg.getFlags().isSet(MessageFlags.OPENED) &&
                !current.getFlags().isSet(MessageFlags.OPENED)) {
                incrementUnreadMessageCount(parent, -1);
            } else if (!msg.getFlags().isSet(MessageFlags.OPENED) &&
                current.getFlags().isSet(MessageFlags.OPENED)) {
                incrementUnreadMessageCount(parent, 1);
            }
            parent.updateMessage(msg);

            if(parent.getRole().equals(FolderData.INBOX_ROLE)) {
                // If the message is in the inbox, then we update the UI
                UIController.getInboxMessageList().updateMessageFlags(msg.getKey(), msg.getFlags());
            }
        }
    }
    
    /**
     * Updates an existing message by redefining its flags. The message is also
     * written into the folder that gets updated immediately. The method accepts
     * the SyncML complete description of a message, but the content is ignored
     * and only the flags are decoded to update the flags.
     *
     * @param parentName parent folder name
     * @param luid message client id
     * @param emailData SyncML byte stream describing the message
     *
     * @throws MailException if the folder update fails
     * @throws XmlException if the byte stream is not SyncML complaint
     */
    public MessageFlags updateMessageFlags(String parentName, String luid,
                                           byte[] emailData)
    throws XmlException, MailException {

        EmailData ed = new EmailData();
        MessageFlags flags = ed.parseFlags(emailData);

        // Update the message in the store
        try {
            Log.debug(LOG_TAG, "Updating flags for message: " + luid);
            Message saved = getMessage(luid);
            if (saved != null) {
                //if the message has been deleted this would generate NPE
                saved.setKey(luid);
                saved.setFlags(flags);
                updateMessage(saved);
            }
            return flags;
        } catch (MailException me) {
            Log.error(LOG_TAG, "Error saving message: " + luid + " " + me.toString() );
            me.printStackTrace();
            throw me;
        }
    }

    /**
     * Sets Message Flags after a messages one or more of its property flag
     * @param msg the message to be flagged
     * @param flag is the flag to be added to the message <code>msg</code>
     */
    public void updateMessageFlag(Message msg, int flag, boolean value)
    throws MailException {
        Log.trace(LOG_TAG, "updateMessageFlag()");
        synchronized (msgStore) {
            
            boolean current = msg.getFlags().isSet(flag);
            if(current == value) {
                // The flag is already set, nothing to change.
                Log.debug(LOG_TAG, "Flag: " + flag + " already set");
                return;
            }

            //Creates MessageFlags to put into updated items HashTable
            MessageFlags flags = new MessageFlags(msg.getFlags());
            flags.setFlag(flag, value);

            // OPENED or FLAGGED flags must be set into the current
            // msg and saved immediately, the others will be set after the sync.
            switch (flag) {
                case MessageFlags.DELETED:
                    updateMessageFlags(msg, flags);
                    break;
                case MessageFlags.ANSWERED:
                case MessageFlags.DRAFT:
                case MessageFlags.FORWARDED:
                case MessageFlags.PARTIAL:
                    //Call update for SyncML
                    updateMessageFlags(msg, flags);
                    break;
                case MessageFlags.OPENED:
                case MessageFlags.FLAGGED:
                    //Call update for SyncML
                    updateMessageFlags(msg, flags);
                    //Save the updated flags message into the folder
                    msg.getFlags().setFlag(flag, value);
                    updateMessage(msg);
                    break;
                default:
                    break;
            }
        }
    }

    /**
     * Updates an existing message by redefining the sent date and/or the
     * received date. The message is also written into the folder that gets
     * updated immediately.
     *
     * @param sentDate sent date. If null the message sent date is left
     *        unchanged
     *
     * @param receivedDate received date. If null the message received date is
     *        left unchanged
     *
     * @throws MailException if the folder update fails
     */
    public void updateMessageDates(Message msg, Date sentDate, Date receivedDate)
    throws MailException {
        Log.trace(LOG_TAG, "updateMessageDates()");
        synchronized (msgStore) {
            if (sentDate != null) {
                msg.setSentDate(sentDate);
            }
            if (receivedDate != null) {
                msg.setReceivedDate(receivedDate);
            }
            updateMessage(msg);
        }
    }
    
    /**
     * Update message flags Lists to be read by MailSyncSource
     * with this implementation the flags are managed only for the Inbox folder
     *
     *@param msg the the message to be put in the queue
     *@param mf the flags to be set
     */
    public void updateMessageFlags(Message msg, MessageFlags mf) {
        Log.trace(LOG_TAG, "updateMessageFlags()");
        if(msg.getParent().getRole().equals(FolderData.INBOX_ROLE)) {
            FlagQueue.put(msg.getKey(), mf);
        }
    }
    
    /**
     * Remove message flags from the queue
     * with this implementation the flags are managed only for the Inbox folder
     *
     *@param msg the the message to be put in the queue
     *@param mf the flags to be set
     */
    public void removeFromFlagQueue(Message msg) throws MailException {
        Log.trace(LOG_TAG, "removeFromFlagQueue(Message)");
        Folder parent = msg.getParent();
        if (parent == null) {
            throw new MailException(MailException.FOLDER_NOT_FOUND_ERROR, "Can't find the parent of msg " + msg.getKey());
        }
        if(parent.getRole().equals(FolderData.INBOX_ROLE)) {
            FlagQueue.remove(msg.getKey());
        }
    }

    /**
     * Delete the specified message
     * @param msg the message to be deleted
     */
    public void deleteMessage(Message msg) throws MailException {
        Log.trace(LOG_TAG, "deleteMessage(Message)");
        synchronized (msgStore) {
            removeFromFlagQueue(msg);
            try {
                if ((ConfigManager.getConfig().isEnableDeletePropagation())) {
                    updateMessageFlag(msg, MessageFlags.DELETED, true);
                }   
            } catch(ConfigException ex) {
                String errorMsg = "Can't remove message, error while reading " +
                        "configuration.";
                Log.error(LOG_TAG, errorMsg);
                throw new MailException(MailException.MESSAGE_ERROR, errorMsg);
            }
            Folder parent = msg.getParent();
            parent.deleteMessage(msg);
            if (!msg.getFlags().isSet(MessageFlags.OPENED)) {
                incrementUnreadMessageCount(parent, -1);
            }
        }
    }
    
    /**
     * Delete a message from the specified message id.
     * 
     * @param luid message local id
     * @throws MailException if the message deletion fails
     */
    public void deleteMessage(String luid) throws MailException {
        Log.trace(LOG_TAG, "deleteMessage(" + luid + ")");
        synchronized (msgStore) {
            String parent = folderFromKey(luid);
            Folder parentf = msgStore.getFolder(parent);
            String recordId = msgIdFromKey(luid);
            Message msg = parentf.getMessage(recordId);
            FlagQueue.remove(luid);
            try {
                parentf.deleteMessage(msg);
                if (!msg.getFlags().isSet(MessageFlags.OPENED)) {
                    incrementUnreadMessageCount(parentf, -1);
                }
            } catch (MailException exc) {
                // If the message to delete was in the outbox
                // it just means it has been sent. We move these
                // messages from outbox to sent as soon as the server inform us that
                // it accepted the item to send, therefore we do not have this email
                // in the outbox any longer. We can ignore the exception in such a
                // case.
                // TODO: would it be even safer to check that the message is now in
                // the sent folder?
                if (!(exc.getCode() == MailException.ITEM_DELETE_ERROR &&
                      parentf.getRole().equals(FolderData.OUTBOX_ROLE))) {
                    throw exc;
                }
            }
        }
    }

    /**
     * This method should be called on a message that has been accepted by the
     * server as to be sent. The method moves the message from local Outbox to
     * local Sent and updates the folders.
     *
     * @param luid local message id to be moved from outbox to sent
     *
     * @throws MailException if one of the operations on the folder fails
     *         (delete from Outbox or add to the Sent)
     */
    public void messageEffectivelySent(String luid) throws MailException {
        Log.trace(LOG_TAG, "messageEffectivelySent(" + luid + ")");
        synchronized (msgStore) {
            String accountFolderPath = accountFromKey(luid);
            AccountFolder account = (AccountFolder)msgStore.getFolder(accountFolderPath);

            Folder sent   = getAccountFolderFromRole(account, FolderData.SENT_ROLE);
            Folder outbox = getAccountFolderFromRole(account, FolderData.OUTBOX_ROLE);

            Message msg = getMessage(outbox, luid);
            if (msg != null) {
                msg.reloadMessage();
                msg.getFlags().setFlag(MessageFlags.TX_SENDING, false);
                msg.getFlags().setFlag(MessageFlags.TX_SENT, true);
                
                outbox.deleteMessage(msg);
                // Update the luid and the message parent folder
                msg.setParent(sent);
                msg.setKey(makeLuid(msg));
                // And finally append the message into the sent folder
                sent.appendMessage(msg);
            } else {
                Log.error(LOG_TAG, "Message not found from the outbox");
                throw new MailException(MailException.GENERIC_MAIL_ERR,
                        "Message not found from the outbox");
            }
        }
    }

    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * In the meantime it updates the number of unread messages.
     *
     * @param folder Is the folder containing the messages to be loaded.
     * @param listener The listener to be notified of the loading events.
     *
     * @return The Vector of Message held by this folder
     * @throws MailException if errors occur retrieving the folder content
     */
    public Vector getMessagesInFolder(Folder folder,
            MessagesLoaderListener listener) throws MailException {

        Log.trace(LOG_TAG, "getMessagesInFolder(Folder, MessagesLoaderListener)");

        Vector result = new Vector();

        synchronized (msgStore) {

            // Dump memory stats
            Log.stats(LOG_TAG, "Beginning of getMessagesInFolder");

            int count  = 0;
            int unread = 0;
            int total = folder.getMessageCount();

            ChildrenEnumeration children = folder.getChildren();
            while(children.hasMoreElements()) {

                Object element = children.nextElement();
                // consider only Message items
                if(!(element instanceof Message)) {
                    continue;
                }

                Message msg = (Message)element;
                msg.setParent(folder);
                msg.setKey(makeLuid(msg));

                result.addElement(msg);
                count++;

                if (!msg.getFlags().isSet(MessageFlags.OPENED)) {
                    ++unread;
                }
                if(listener != null) {
                    listener.messageLoaded(count, total);
                }
            }
            try {
                // Close the iterator when finishes
                children.close();
            } catch (RecordStoreException ex) {
                Log.error(LOG_TAG, "Can't close folder enumeration.");
            }
            setUnreadMessageCount(folder, unread);

            // Dump memory stats
            Log.stats(LOG_TAG, "End of getMessagesInFolder");
            return result;
        }
    }

    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * In the meantime it updates the number of unread messages.
     *
     * @param folder Is the folder containing the messages to be loaded.
     *
     * @return The Vector of Message held by this folder
     * @throws MailException if errors occur retrieving the folder content
     */
    public Vector getMessagesInFolder(Folder folder) throws MailException {
        Log.trace(LOG_TAG, "getMessagesInFolder(Folder)");
        return getMessagesInFolder(folder, null);
    }

    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * @param folderName Is the folder name.
     * @param listener The listener to be notified of the loading events.
     *
     * @return An array of Message held by this folder
     *
     * @throws MailException if errors occur retrieving the folder content
     */
    public Vector getMessagesInFolder(String folderPath, MessagesLoaderListener listener)
    throws MailException {
        Log.trace(LOG_TAG, "getMessagesInFolder(" + folderPath + ", MessagesLoaderListener)");
        synchronized (msgStore) {
            return getMessagesInFolder(msgStore.getFolder(folderPath), listener);
        }
    }

    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * @param folderName Is the folder name.
     *
     * @return An array of Message held by this folder
     *
     * @throws MailException if errors occur retrieving the folder content
     */
    public Vector getMessagesInFolder(String folderPath) throws MailException {
        Log.trace(LOG_TAG, "getMessagesInFolder(" + folderPath + ")");
        synchronized (msgStore) {
            return getMessagesInFolder(msgStore.getFolder(folderPath));
        }
    }

    /**
     * Return the number of messages in the given folder.
     *
     * @param folder
     * @return The number of messages
     */
    public int getMessagesCount(Folder folder) throws MailException {
        Log.trace(LOG_TAG, "getMessageCount(Folder)");
        synchronized (msgStore) {
            return folder.getMessageCount();
        }
    }

    /**
     * Return the number of unread messages in the given folder.
     *
     * @param folder
     * @return The number of unread messages
     */
    public int getUnreadMessagesCount(Folder folder) throws MailException {
        Log.trace(LOG_TAG, "getUnreadMessagesCount(Folder)");
        synchronized (unreadMessagesCountByFolder) {
            Integer count = (Integer)unreadMessagesCountByFolder.get(folder.getFullName());
            if(count == null) {
                // Updates the unread messages count
                getMessagesInFolder(folder, null);
                count = (Integer)unreadMessagesCountByFolder.get(folder.getFullName());
            }
            // Return 0 if we cannot find the proper count
            return count != null ? count.intValue() : 0;
        }
    }

    /**
     * Return the number of unread messages in the inbox of the given account.
     *
     * @param account
     * @return The number of unread messages
     */
    public int getUnreadMessagesCount(AccountFolder account) throws MailException {
        Log.trace(LOG_TAG, "getUnreadMessagesCount(AccountFolder)");
        Folder inbox = getAccountFolderFromRole(account, FolderData.INBOX_ROLE);
        if(inbox != null) {
            return getUnreadMessagesCount(inbox);
        } else {
            return 0;
        }
    }

    /**
     * Set the unread messages count for the specified folder.
     *
     * NOTE: the folder role must be inbox, otherwise it will be ignored.
     * 
     * @param folder
     * @param count The new unread count
     */
    private void setUnreadMessageCount(Folder folder, int count) {
        // Cache information only for inbox folders
        synchronized(unreadMessagesCountByFolder) {
            if(folder.getRole() != null && folder.getRole().equals(FolderData.INBOX_ROLE)) {
                Integer unread = (Integer)unreadMessagesCountByFolder.get(folder.getFullName());
                if((unread != null) && (unread.intValue() == count)) {
                    return;
                }
                unreadMessagesCountByFolder.put(folder.getFullName(), new Integer(count));
                // Notify the UnreadMessagesListener if setted
                if((unreadMessageListener != null) && !((unread == null) && (count == 0))) {
                    unreadMessageListener.updateUnreadMessages(folder, count);
                }
            }
        }
    }

    /**
     * Increment the unread messages count of the specified folder.
     * 
     * @param folder
     * @param increment
     */
    private void incrementUnreadMessageCount(Folder folder, int increment) {
        synchronized(unreadMessagesCountByFolder) {
            Integer current = (Integer)unreadMessagesCountByFolder.get(folder.getFullName());
            int count = 0;
            if(current != null) {
                count = current.intValue();
            } else {
                count = getUnreadMessagesCount(folder);
            }
            count += increment;
            setUnreadMessageCount(folder, count);
        }
    }

    /**
     * Set the current <code>UnreadMessagesListener</code> to be notified when
     * the number of unread messages has changed.
     * 
     * @param listener
     */
    public void setUnreadMessagesListener(UnreadMessagesListener listener) {
        unreadMessageListener = listener;
    }

    /**
     * Reset the unread messages count cache
     */
    private void resetUnreadMessagesCache() {
        unreadMessagesCountByFolder.clear();
    }

    /**
     * Return a single message fetched from its folder.
     *
     * @param key the message key
     * @return the Message
     */
    public Message getMessage(String key) throws MailException {
        Log.trace(LOG_TAG, "getMessage(" + key + ")");
        synchronized (msgStore) {
            String folderName = folderFromKey(key);
            Folder folder     = msgStore.getFolder(folderName, true);
            return getMessage(folder, key);
        }
    }

    /**
     * Return a single message fetched from its folder.
     *
     * @param folder The folder containing the message
     * @param key The message key
     * @return the Message
     */
    public Message getMessage(Folder folder, String key) throws MailException {
        Log.trace(LOG_TAG, "getMessage(Folder, " + key + ")");
        synchronized (msgStore) {
            String recordId = msgIdFromKey(key);
            // Get the message
            Message msg = folder.getMessage(recordId);
            // Set Message key
            if (msg!=null) {
                //if the message has been previously deleted this generates NPE
                msg.setKey(makeLuid(msg));
            }
            return msg;
        }
    }

    /**
     * Remove a folder and recreate it empty.
     * 
     * @param  folder the <code>Folder</code> to be emptied.
     * @throws MailException In case of errors.
     */
    public void emptyFolder(Folder folder) throws MailException {
        Log.trace(LOG_TAG, "emptyFolder(Folder)");
        synchronized (msgStore) {
            msgStore.removeFolder(folder, true);
            msgStore.addFolder(new Folder(folder.getFullName(), folder.getRole(),
                    new Date(), msgStore));
            // Reset the number of unread messages
            setUnreadMessageCount(folder, 0);
        }
    }
    
    /**
     * Create a new <code>Folder</code> from a SyncML byte stream. The
     * <code>Folder</code> is also added to the related parent. 
     *
     * @param parent The parent name.
     * @param content The SyncML byte stream representing the <code>Folder</code>.
     *
     * @return The new created <code>Folder</code>.
     *
     * @throws XmlException  If the byte stream cannot be parsed.
     * @throws MailException If the <code>Folder</code> cannot be saved.
     */
    public Folder createFolder(String parent, byte[] content)
    throws XmlException, MailException {
        Log.info(LOG_TAG, "createFolder(" + parent + ", content)");

        synchronized (msgStore) {
            FolderData fd = new FolderData();
            fd.parse(content);
            Folder folder = null;

            // Check item type: AccountFolder or a normal Folder
            if(fd.getRole().equals(AccountFolder.ACCOUNT_ROLE)) {
                folder = new AccountFolder(parent + fd.getName(), fd.getCreated(),
                        msgStore,
                        ((ChunkedString)fd.getExtValues(AccountFolder.VISIBLE_NAME).elementAt(0)).toString(),
                        ((ChunkedString)fd.getExtValues(AccountFolder.EMAIL_ADDRESS).elementAt(0)).toString());
            } else {
                folder = new Folder(parent + Folder.FOLDER_SEPARATOR + fd.getName(), fd.getRole(),
                    fd.getCreated(), msgStore);
            }

            // Add the new folder
            addFolder(folder);

            return folder;
        }
    }

    /**
     * Create a new <code>Folder</code> (not an account).
     *
     * @param account The parent account
     * @param name The folder name.
     * @param role The folder role
     * @param created The folder creation date
     *
     * @return The new created <code>Folder</code>.
     *
     * @throws MailException If the <code>Folder</code> cannot be saved.
     */
    public Folder createFolder(AccountFolder account, String name,
                               String role, Date created)
    throws MailException {
        
        Log.trace(LOG_TAG, "createFolder()");
        synchronized (msgStore) {
            StringBuffer fullPath = new StringBuffer();
            fullPath.append(account.getFullName()).append(Folder.FOLDER_SEPARATOR).append(name);

            Folder folder = new Folder(fullPath.toString(), role, created, msgStore);
            addFolder(folder);

            return folder;
        }
    }

    /**
     * Create a new <code>AccountFolder</code>.
     *
     * @param accountName The name of the new account
     * @param created The account creation date
     *
     * @return The new created <code>AccountFolder</code>.
     *
     * @throws MailException If the <code>AccountFolder</code> cannot be saved.
     */
    public AccountFolder createAccountFolder(String accountName, 
                                             Date created, String visibleName,
                                             String emailAddress)
    throws MailException {

        Log.trace(LOG_TAG, "createAccountFolder()");
        synchronized (msgStore) {
            StringBuffer fullPath = new StringBuffer();
            fullPath.append(Folder.ROOT_FOLDER_PATH).append(accountName);
            AccountFolder account = new AccountFolder(fullPath.toString(),
                    created, msgStore, visibleName, emailAddress);
            msgStore.addFolder(account);
            // Invalidate the accounts cache
            accountFolders = null;
            return account;
        }
    }


    /**
     * Add a new <code>Folder</code> to the <code>Store</code>.
     *
     * @param folder The new <code>Folder</code> to be added.
     *
     */
    public void addFolder(Folder folder) {
        Log.trace(LOG_TAG, "addFolder(Folder)");
        synchronized (msgStore) {
            msgStore.addFolder(folder);
            // Initialize the number of unread message count
            setUnreadMessageCount(folder, 0);
            // Invalidate the accounts cache
            accountFolders = null;
        }
    }

    /**
     * Updates an existing <code>Folder</code>.
     *
     * @param parent The parent name.
     * @param content The SyncML byte stream representing the <code>Folder</code>.
     * @param luid The local id of the <code>Folder</code>.
     *
     * @return The updated <code>Folder</code>.
     *
     * @throws XmlException  If the byte stream cannot be parsed.
     * @throws MailException If the <code>Folder</code> cannot be updated.
     */
    public Folder updateFolder(String parent, byte[] content, String luid)
    throws XmlException, MailException {
        Log.trace(LOG_TAG, "updateFolder(" + parent + ", content, " + luid + ")");
        synchronized (msgStore) {
            FolderData fd = new FolderData();
            fd.parse(content);
            String fullname = null;
            if(fd.getRole().equals(AccountFolder.ACCOUNT_ROLE)) {
                fullname = parent + fd.getName();
            } else {
                fullname = parent + Folder.FOLDER_SEPARATOR + fd.getName();
            }
            if(!fullname.equals(luid)) {
                // Cannot update a folder name
                throw new MailException(MailException.FOLDER_ERROR,
                        "Cannot update Folder " + fullname + ": the Folder name must be the same " +
                        "as the existing one.");
            }
            Folder folder = null;
            if(fd.getRole().equals(AccountFolder.ACCOUNT_ROLE)) {
                folder = new AccountFolder(parent + fd.getName(), fd.getCreated(),
                        msgStore,
                        ((ChunkedString)fd.getExtValues(AccountFolder.VISIBLE_NAME).elementAt(0)).toString(),
                        ((ChunkedString)fd.getExtValues(AccountFolder.EMAIL_ADDRESS).elementAt(0)).toString());
            } else {
                folder = new Folder(parent + Folder.FOLDER_SEPARATOR + fd.getName(), fd.getRole(),
                    fd.getCreated(), msgStore);
            }
            String childId = msgStore.retrieveSubfolderID(luid, parent);
            folder.setRecordId(Integer.parseInt(childId));
            msgStore.updateChild(parent, folder);

            // Invalidate the accounts cache
            accountFolders = null;

            return updateFolder(parent, folder);
        }
    }

    /**
     * Updates an existing <code>Folder</code>.
     *
     * @param folder The <code>Folder</code> to be updated to the Store
     *
     * @return The updated <code>Folder</code>.
     *
     * @throws XmlException  If the byte stream cannot be parsed.
     * @throws MailException If the <code>Folder</code> cannot be updated.
     */
    public Folder updateFolder(Folder folder) throws XmlException, MailException {
        Log.trace(LOG_TAG, "updateFolder(" + folder + ")");
        return updateFolder(folder.getParent().getFullName(), folder);
    }

    /**
     * Updtes the given <code>Folder</code> to the Store
     * @param parent
     * @param folder
     * @return
     * @throws XmlException
     * @throws MailException
     */
    private Folder updateFolder(String parent, Folder folder) throws XmlException, MailException {
        Log.trace(LOG_TAG, "updateFolder(" + parent + ", " + folder + ")");
        synchronized (msgStore) {
            if(folder instanceof AccountFolder) {
                AccountFolder existing = (AccountFolder)getFolder(folder.getFullName());
                UIController.getInboxMessageList().accountUpdated(existing, (AccountFolder)folder);
            }
            msgStore.updateChild(parent, folder);
            return folder;
        }
    }

    /**
     * Get an existing <code>Folder</code> from key.
     *
     * @param key It represents the path of the existing <code>Folder</code>.
     * @return The <code>Folder</code> object.
     */
    public Folder getFolder(String key) throws MailException {
        Log.trace(LOG_TAG, "getFolder(" + key + ")");
        synchronized (msgStore) {
            return msgStore.getFolder(key, false);
        }
    }

    /**
     * Get an existing <code>Folder</code> from key.
     *
     * @param key It represents the path of the existing <code>Folder</code>.
     * @param lightFolder If true return the <code>Folder</code> item without
     * reading its properties (e.g. role, creation date). It doesn't check
     * if it really exists in the Store.
     * @return The <code>Folder</code> object.
     */
    public Folder getFolder(String key, boolean lightFolder) throws MailException {
        Log.trace(LOG_TAG, "getFolder(" + key + ")");
        synchronized (msgStore) {
            return msgStore.getFolder(key, lightFolder);
        }
    }

    /**
     * Get an existing parent of the given <code>Folder</code> from key.
     *
     * @param folder The <code>Folder</code>.
     * @return The <code>Folder</code> parent.
     */
    public Folder getFolderParent(Folder folder) throws MailException {
        Log.trace(LOG_TAG, "getFolderParent(Folder)");
        synchronized (msgStore) {
            return folder.getParent();
        }
    }

    /**
     * Delete an existing <code>Folder</code> from key.
     *
     * @param key It represents the path of the existing <code>Folder</code>.
     *
     * @throws MailException If the <code>Folder</code> deletion fails.
     */
    public void deleteFolder(String key) throws MailException {
        Log.trace(LOG_TAG, "deleteFolder(" + key + ")");
        synchronized (msgStore) {
            Folder folder = msgStore.getFolder(key, false);
            msgStore.removeFolder(folder, true);
            unreadMessagesCountByFolder.remove(key);

            // Remove the foldersByAccount cached information if we remove a
            // child of an account or the account itself
            String accountFullName = (folder.getRole().equals(AccountFolder.ACCOUNT_ROLE)) ?
                folder.getFullName() : folder.getParent().getFullName();
            foldersByAccount.remove(accountFullName);
            // Invalidate the accounts cache
            accountFolders = null;
        }
    }

    /**
     * Get all the <code>AccountFolder</code> objects available in the
     * <code>Store</code>.
     *
     * @return The <code>AccountFolder</code> objects array.
     *
     * @throws MailException If the operation fails.
     */
    public AccountFolder[] getAccounts() throws MailException {
        Log.trace(LOG_TAG, "getAccounts()");

        synchronized (msgStore) {
            if (accountFolders == null) {
                Folder[] list = msgStore.list(Folder.ROOT_FOLDER_PATH);
                Vector accounts = new Vector();
                for(int i=0; i<list.length; i++) {
                    if(list[i] instanceof AccountFolder) {
                        accounts.addElement((AccountFolder)list[i]);
                    }
                }
                accountFolders = toAccountFolderArray(accounts);
            }
        }
        return accountFolders;
    }

    /**
     * Get all the <code>Folder</code> objects contained in a
     * <code>AccountFolder</code>.
     *
     * @param account The <code>AccountFolder</code> reference.
     *
     * @return The <code>Folder</code> objects array.
     *
     * @throws MailException If the operation fails.
     */
    public Folder[] getAccountFolders(AccountFolder account) throws MailException {
        Log.trace(LOG_TAG, "getAccountFolders(" + account.getFullName() + ")");
        return getAccountFolders(account.getFullName());
    }

    /**
     * Get all the <code>Folder</code> objects contained in a
     * <code>AccountFolder</code>.
     *
     * @param account The account name string.
     *
     * @return The <code>Folder</code> objects array.
     *
     * @throws MailException If the operation fails.
     */
    public Folder[] getAccountFolders(String account) throws MailException {
        Log.trace(LOG_TAG, "getAccountFolders(" + account + ")");
        synchronized (msgStore) {

            // Check if the information is already cached and complete
            // (e.g. it contains all the account folders: Inbox, Outbox, Drafts, Sent)
            Object folders = foldersByAccount.get(account);
            if(folders != null) {
                Vector foldersVector = (Vector)folders;
                if(foldersVector.size() == 4) {
                    return toFolderArray((Vector)folders);
                }
            }

            // Fix the account name if it doesn't include the root folder path
            if(!account.startsWith(Folder.ROOT_FOLDER_PATH)) {
                account = Folder.ROOT_FOLDER_PATH + account;
            }

            Folder[] foldersArray = msgStore.list(account);

            // Update the cache
            Vector foldersVector = new Vector();
            for(int i=0; i<foldersArray.length; i++) {
                foldersVector.addElement(foldersArray[i]);
            }
            foldersByAccount.put(account, foldersVector);

            return foldersArray;
        }
    }

    /**
     * Get the <code>Folder</code> belonging to a <code>AccountFolder</code>,
     * which has the specified role.
     *
     * @param account The <code>Folder</code> role, or null if it cannot be
     *        found
     *
     * @return The <code>Folder</code> object.
     *
     * @see com.funambol.mailclient.syncml.FolderData#INBOX_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#OUTBOX_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#SENT_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#DRAFTS_ROLE
     *
     * @throws MailException If the operation fails.
     */
    public Folder getAccountFolderFromRole(AccountFolder account, String role) throws MailException {
        Log.trace(LOG_TAG, "getAccountFolderFromRole(AccountFolder, " + role + ")");
        synchronized(msgStore) {
            Folder[] folders = getAccountFolders(account);
            for(int i=0; i<folders.length; i++) {
                if(folders[i].getRole().equals(role)) {
                    return folders[i];
                }
            }
        }
        Log.debug(LOG_TAG, "Cannot find folders having the specified role: " + role);
        return null;
    }

    /**
     * Get the <code>Folder</code> objects having the specified role.
     *
     * @param role A specific role.
     * 
     * @see com.funambol.mailclient.syncml.FolderData#INBOX_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#OUTBOX_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#SENT_ROLE
     * @see com.funambol.mailclient.syncml.FolderData#DRAFTS_ROLE
     *
     * @return The <code>Folder</code> objects array
     */
    public Folder[] getFoldersFromRole(String role) {
        Log.trace(LOG_TAG, "getFoldersFromRole("+role+")");

        synchronized(msgStore) {
            AccountFolder[] accounts = getAccounts();
            Vector ret = new Vector();
            for(int i=0; i<accounts.length; i++) {
                Folder f = getAccountFolderFromRole(accounts[i], role);
                if(f != null) {
                    ret.addElement(f);
                }
            }
            return toFolderArray(ret);
        }
    }

    /***********************************************************************/
    /************************* UTILITY METHODS *****************************/
    /***********************************************************************/

    /**
     * This method creates a LUID for a given message.
     * In our client a LUID is folderName/messageRecordId
     *
     * @param msg the message
     * @return the LUID
     */
    public static String makeLuid(Message msg) {
        return msg.getParent().getFullName() + "/" + msg.getRecordId();
    }

    /**
     * Get the message Id from the key.
     *
     * @param key is the message key
     * @return is the record id
     */
    public static String msgIdFromKey(String key) {
        int pos = key.lastIndexOf(Folder.FOLDER_SEPARATOR);
        if(pos != -1) {
            return key.substring(pos+1);
        }
        return key;
    }

    /**
     * Get the folder fullname from the key.
     *
     * @param key is the message key
     * @return the folder name
     */
    public String folderFromKey(String key) {
        if(key == null) {
            return null;
        }
        int pos = key.lastIndexOf(Folder.FOLDER_SEPARATOR);
        if(pos != -1) {
            return key.substring(0, pos);
        }
        return null;
    }

    /**
     * Convert Funambol 3.0 folder names into local folder names.
     *
     * @param parent the item parent string
     * @return the converted name, or the original one if does not
     *         starts with "ROOT/"
     */
    public String folderFromParent(String parent) {
        final String root = "ROOT/";
        if(parent == null) {
            return null;
        }

        // Handle backward compatibility when receiving items from an old server
        if (parent.startsWith(root)) {
            return "/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/" +
                    folderFromPrefix(parent.charAt(root.length()));
        } else if(isOldLocalFolderId(parent)) {
            return "/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/" + parent;
        }
        return parent;
    }

    /**
     * Find the folder name from the Funambol key prefix.
     */
    private String folderFromPrefix(char prefix) {
        switch (prefix) {
            case 'I':
                return "Inbox";
            case 'O':
                return "Outbox";
            case 'D':
                return "Drafts";
            case 'S':
                return "Sent";
            default:
                return null;
        }
    }

    private boolean isOldLocalFolderId(String id) {
        return "Inbox".equals(id) || "Outbox".equals(id);
    }

    /**
     * Get the account name from the key.
     *
     * @param key is the message key
     * @return the account name
     */
    public String accountFromKey(String key) {
        if(key == null) {
            return null;
        }
        int pos = key.indexOf(Folder.FOLDER_SEPARATOR, 1);
        if(pos != -1) {
            return key.substring(0, pos);
        }
        return null;
    }

    /**
     * Converts a <code>Vector</code> of <code>Folder</code> items into an
     * array.
     *
     * @param v The vector to be converted
     * @return Folder[] representation of <code>Vector</code> v
     */
    private Folder[] toFolderArray(Vector v) {
        int size = v.size();
        Folder[] ret = new Folder[size];
        v.copyInto(ret);
        return ret;
    }

    /**
     * Converts a <code>Vector</code> of <code>AccountFolder</code> items into
     * an array.
     *
     * @param v The vector to be converted
     * @return AccountFolder[] representation of <code>Vector</code> v
     */
    private AccountFolder[] toAccountFolderArray(Vector v) {
        int size = v.size();
        AccountFolder[] ret = new AccountFolder[size];
        v.copyInto(ret);
        return ret;
    }
}

