/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.syncml;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Date;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import com.funambol.mail.BodyPart;
import com.funambol.util.ChunkedString;
import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;
import com.funambol.mail.MIMEFormatter;
import com.funambol.mail.MIMEProcessor;
import com.funambol.mail.MailException;

import com.funambol.mail.Multipart;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.util.Log;
import com.funambol.util.XmlUtil;
import com.funambol.util.StringUtil;
import com.funambol.util.MailDateFormatter;
import com.funambol.util.QuotedPrintable;
import com.funambol.util.XmlException;
import java.io.UnsupportedEncodingException;

/**
 * This class is able to parse and format the OMA-DS 1.2 Email object.
 */
public class EmailData {

    // ------------------------------------------------------- Private data
    // These fields are SyncML tags for email description

    private static final String EMAIL_TAG = "Email";
    private static final String READ_TAG  = "read";
    private static final String FORW_TAG  = "forwarded" ;
    private static final String REPL_TAG  = "replied" ;
    private static final String TREC_TAG  = "received" ;
    private static final String TCRE_TAG  = "created" ;
    private static final String TMOD_TAG  = "modified" ;
    private static final String DELE_TAG  = "deleted" ;
    private static final String FLAG_TAG  = "flagged" ;
    private static final String ITEM_TAG  = "emailitem" ;
    private static final String EXT_TAG   = "Ext" ;

    private MessageFlags flags;
    
    private Date received;
    private Date created;
    private Date modified;

    private MIMEFormatter mimef;

    private Message emailItem;


    // ------------------------------------------------------- Constructors

    /** 
     * Inizialize an empty instance
     */
    public EmailData() {
        this(null);
    }

    /** 
     * Initialize an instance taking the information from Message.
     */
    public EmailData(Message msg) {
        this(msg, null);
        mimef = new MIMEFormatter();
    }

    public EmailData(Message msg, MIMEFormatter formatter) {
        if (msg != null) {
            this.flags = msg.getFlags();
            this.received = msg.getReceivedDate();
            this.created = msg.getSentDate();
            this.modified = msg.getSentDate();
        }
        else {
            flags = new MessageFlags() ;
            received = created = modified = null ;
        }

        this.emailItem = msg;
        this.mimef = formatter;
    }

    // ---------------------------------------------------------- Accessors  
    
    /** Returns the message flags */
    public MessageFlags getFlags() { return flags; }
    /** Sets the message flags */
    public void setFlags(MessageFlags flags) { this.flags=flags; }

    /** Returns the received time */
    public Date getReceived() { return received; }
    /** Sets the received time */
    public void setReceived(Date v) { received=v; }

    /** Returns the created time */
    public Date getCreated() { return created; }
    /** Sets the created time */
    public void setCreated(Date v) { created=v; }

    /** Returns the modified time */
    public Date getModified() { return modified; }
    /** Sets the modified time */
    public void setModified(Date v) { modified=v; }

    /** 
     * Returns the Message contained in this Email object, if present
     */
    public Message getEmailItem() {
        return emailItem;
    }

    /** 
     * Set the emailItem Message in this Email object, to be sent out
     */
    public void setEmailItem(Message msg) {
        emailItem = msg;
    }

    // ----------------------------------------------------- Public Methods

    /**
     * Parse the Email item content.
     */
    public void parse(byte[] syncmlData)
                            throws XmlException, MailException {

        ChunkedString data = new ChunkedString(new String(syncmlData));

        ChunkedString xml = XmlUtil.getTagValue(data, EMAIL_TAG);
        data = null;
        flags.setFlag(MessageFlags.OPENED,    getXmlFlag(xml, READ_TAG));
        flags.setFlag(MessageFlags.FORWARDED, getXmlFlag(xml, FORW_TAG));
        flags.setFlag(MessageFlags.ANSWERED,  getXmlFlag(xml, REPL_TAG));
        flags.setFlag(MessageFlags.DELETED,   getXmlFlag(xml, DELE_TAG));
        flags.setFlag(MessageFlags.FLAGGED,   getXmlFlag(xml, FLAG_TAG));
        
        received = getXmlDate(xml, TREC_TAG);
        created =  getXmlDate(xml, TCRE_TAG);
        modified = getXmlDate(xml, TMOD_TAG);

        int itemPos = XmlUtil.getTag(xml, ITEM_TAG);
        if(itemPos > -1) {
            int begin = xml.indexOf(">", itemPos);
            if(xml.charAt(begin-1) == '/') {
                Log.debug("EmailData.parse: empty emailitem.");
                return;
            }
            int end = xml.indexOf("</"+ITEM_TAG+">", begin);
            if(begin == -1 || end == -1) {
                Log.error("[EmailData] error parsing EmailItem");
                throw new XmlException("Can't parse email item");
            }
            // FIXME: implement the CDATA handling in the Xml parser or use
            //        kxml
            int pos = xml.indexOf("<![CDATA[", begin);
            if (pos != -1) {
                begin = pos + 8;
                end = xml.indexOf("]]>", begin);
                if (end == -1) {
                    Log.error("[EmailData] can't find CDATA end.");
                    throw new XmlException("Can't find CDATA end.");
                }
            }
            ChunkedString item = xml.substring(begin+1, end);
            Hashtable attr = XmlUtil.getTagAttributes(xml, ITEM_TAG);
            //Log.debug("Retrieving encoding");
            String encoding = (String)attr.get("enc");
            // Try with Uppercase for 3.0 connector bug
            if(encoding == null) {
                encoding = (String)attr.get("ENC");
            }
            if(encoding != null) {
                Log.debug("EmailData.parse: encoding from server= " + encoding);
                if(StringUtil.equalsIgnoreCase(encoding,"quoted-printable")) {
                    // TODO: optimize this. Since the connector now does
                    //       not send QP to the client, we will not fall
                    //       into this case.
                    String qp = item.toString();
                    qp = QuotedPrintable.decode(qp.getBytes(), "UTF-8");
                    item = new ChunkedString(qp);
                }
            }
            // 
            MIMEProcessor mp = new MIMEProcessor();
            emailItem = mp.parseMailMessage(item);
            emailItem.setFlags(flags);

            // Process the Funambol extensions for filtering. Example:
            // <Ext><XNam>x-funambol-body</XNam><XVal>15000</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach-n</XNam><XVal>1</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach</XNam><XVal>att1.txt</XVal><XVal>10000</XVal></Ext>
            // XXX: we are now considering x-funambol-body only
            Vector extList = XmlUtil.getTagValues(xml, EXT_TAG);
            for(int i=0, l=extList.size(); i<l; i++) {
                ChunkedString ext = (ChunkedString)extList.elementAt(i);
                try {
                    ChunkedString xnam = XmlUtil.getTagValue(ext, "XNam");
                    if(xnam.equals("x-funambol-body")) {
                        String xval = XmlUtil.getTagValue(ext, "XVal").toString();
                        if (xval == null) {
                            Log.info("Incomplete header");
                        }
                        emailItem.getFlags().setFlag(MessageFlags.PARTIAL, true);
                        break;
                    }
                } catch (XmlException xe) {
                    xe.printStackTrace();
                    Log.error("[EmailData.parse]Invalid Ext field. Ignoring it");
                }
            }
        }
    }

    public byte[] format() {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            format(baos, false);
            return baos.toByteArray();
        } catch (IOException ioe) {
            Log.error("Cannot format email data: " + ioe.toString());
            return null;
        }
    }

    public void format(OutputStream os) throws IOException {
        format(os, false);
    }


    /**
     * Format the item content.
     */
    public void format(OutputStream os, boolean explodeAtt) throws IOException {
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");
        addXmlFlag(out, READ_TAG, flags.isSet(MessageFlags.OPENED));
        addXmlFlag(out, FORW_TAG, flags.isSet(MessageFlags.FORWARDED));
        addXmlFlag(out, REPL_TAG, flags.isSet(MessageFlags.ANSWERED));
        addXmlFlag(out, DELE_TAG, flags.isSet(MessageFlags.DELETED));
        addXmlFlag(out, FLAG_TAG, flags.isSet(MessageFlags.FLAGGED));


        String utcdate = null;
        
        if(received != null){
            utcdate = MailDateFormatter.dateToUTC(received);
            XmlUtil.addElement(out, TREC_TAG, utcdate);
        }
        if(created != null){
            utcdate = MailDateFormatter.dateToUTC(created);
            XmlUtil.addElement(out, TCRE_TAG, utcdate);
        }
        if(modified != null) {
            utcdate = MailDateFormatter.dateToUTC(modified);
            XmlUtil.addElement(out, TMOD_TAG, utcdate);
        }
        
        print(os, out.toString());

        if (emailItem != null) {
            print(os, "<emailitem>\n<![CDATA[");
            // Append MIME representation of emailItem to out.
            mimef.format(emailItem, os);
            print(os, "]]&gt;\n</emailitem>\n");
        }
        print(os, "</Email>\n");
    }

    //---------------------------------------------------------- Private methods

    

    /**
     * Dumps an email flags into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format
     */
    public byte[] formatFlags(MessageFlags flags) {
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");
        formatFlags(flags, out);
        out.append("</Email>\n");
        return out.toString().getBytes();
    }

    /**
     * Dumps an email into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format
     */
    public byte[] format(Message msg) {
        MessageFlags flags = msg.getFlags();
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");

        formatFlags(flags, out);

        String utcdate = null;
        Date received = msg.getReceivedDate();
        Date created  = msg.getSentDate();
        Date modified = msg.getSentDate();
        
        if(received != null){
            utcdate = MailDateFormatter.dateToUTC(received);
            XmlUtil.addElement(out, TREC_TAG, utcdate);
        }
        if(created != null){
            utcdate = MailDateFormatter.dateToUTC(created);
            XmlUtil.addElement(out, TCRE_TAG, utcdate);
        }
        if(modified != null) {
            utcdate = MailDateFormatter.dateToUTC(modified);
            XmlUtil.addElement(out, TMOD_TAG, utcdate);
        }
        
        out.append("<emailitem>\n<![CDATA[");
        // Append MIME representation of emailItem to out.
        mimef.format(msg, out);
        out.append("]]&gt;\n</emailitem>\n");
        out.append("</Email>\n");

        return out.toString().getBytes();
    }

    
    /**
     * Dumps flags and msgid of an email into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format with only flags and msgid
     */
    public byte[] formatMinimalInfo(Message msg) {
        MessageFlags flags = msg.getFlags();
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");

        // append the the flags
        formatFlags(flags, out);
        
        out.append("<emailitem>\n<![CDATA[");
        
        // Append MIME representation of message id 
        mimef.formatOnlyMsgID(msg, out);
        out.append("]]&gt;\n</emailitem>\n");
        out.append("</Email>\n");

        return out.toString().getBytes();
    }

        /**
     * Utility method to dump the message flags into a string representing a
     * SyncML email item.
     * @param flags message flags
     * @param out output string
     */
    private void formatFlags(MessageFlags flags, StringBuffer out) {
        addXmlFlag(out, READ_TAG, flags.isSet(MessageFlags.OPENED));
        addXmlFlag(out, FORW_TAG, flags.isSet(MessageFlags.FORWARDED));
        addXmlFlag(out, REPL_TAG, flags.isSet(MessageFlags.ANSWERED));
        addXmlFlag(out, DELE_TAG, flags.isSet(MessageFlags.DELETED));
        addXmlFlag(out, FLAG_TAG, flags.isSet(MessageFlags.FLAGGED));
    }
    
    
    //--------------------------------------------------------- Private Fields
   

    //--------------------------------------------------------- Private Methods

    /**
     * Utility function to parse an email message flags (the email is a SyncML
     * byte stream)
     *
     * @param syncmlData is the SyncML byte stream
     * @return the MessageFlags
     */
    public MessageFlags parseFlags(byte[] syncmlData)
    throws XmlException, MailException {
        ChunkedString data = new ChunkedString(new String(syncmlData));
        ChunkedString xml = XmlUtil.getTagValue(data, EMAIL_TAG);
        return parseFlags(xml);
    }

     /**
     * Utility function to parse an email message flags (the email is a SyncML
     * chunked string)
     *
     * @param xml is the SyncML byte stream tokenized
     * @return the MessageFlags
     */
    private MessageFlags parseFlags(ChunkedString xml)
    throws XmlException, MailException {

        MessageFlags flags = new MessageFlags();

        flags.setFlag(MessageFlags.OPENED,    getXmlFlag(xml, READ_TAG));
        flags.setFlag(MessageFlags.FORWARDED, getXmlFlag(xml, FORW_TAG));
        flags.setFlag(MessageFlags.ANSWERED,  getXmlFlag(xml, REPL_TAG));
        flags.setFlag(MessageFlags.DELETED,   getXmlFlag(xml, DELE_TAG));
        flags.setFlag(MessageFlags.FLAGGED,   getXmlFlag(xml, FLAG_TAG));

        return flags;
    }


    /**
     * This method parses a SyncML byte stream describing an email and builds
     * the corresponding Message object.
     *
     * @param syncmlData email in SyncML byte stream format
     * @return the Message
     * @throws XmlException if the byte stream cannot be parsed
     */
    public Message parseMessage(byte[] syncmlData)
    throws XmlException {

        // Dump memory stats
        Log.stats("Beginning of MessageManager.parseMessage");

        ChunkedString data = null;
        try {
            // The encoding must be set to UTF-8 
            // in order to read the symbols like "euro" 
            data = new ChunkedString(new String(syncmlData, "UTF-8"));
        } catch (UnsupportedEncodingException uee) {
            uee.printStackTrace();
            Log.error("[MessageManager] Cannot convert email data");
        }
        
        ChunkedString xml = XmlUtil.getTagValue(data, EMAIL_TAG);
        data = null;

        MessageFlags flags = parseFlags(xml);
       
        Date received = getXmlDate(xml, TREC_TAG);
        Date created =  getXmlDate(xml, TCRE_TAG);
        Date modified = getXmlDate(xml, TMOD_TAG);

        int itemPos = XmlUtil.getTag(xml, ITEM_TAG);
        if(itemPos > -1) {
            int begin = xml.indexOf(">", itemPos);
            if(xml.charAt(begin-1) == '/') {
                Log.debug("EmailData.parse: empty emailitem.");
                return null;
            }
            int end = xml.indexOf("</"+ITEM_TAG+">", begin);
            if(begin == -1 || end == -1) {
                Log.error("[EmailData] error parsing EmailItem");
                throw new XmlException("Can't parse email item");
            }
            // FIXME: implement the CDATA handling in the Xml parser or use
            //        kxml
            int pos = xml.indexOf("<![CDATA[", begin);
            if (pos != -1) {
                begin = pos + 8;
                end = xml.indexOf("]]>", begin);
                
                if (end == -1) {
                    Log.error("[EmailData] can't find CDATA end.");
                    throw new XmlException("Can't find CDATA end.");
                }
            }
            ChunkedString item = xml.substring(begin+1, end);
            
            Hashtable attr = XmlUtil.getTagAttributes(xml, ITEM_TAG);
            String encoding = (String)attr.get("enc");
            // Try with Uppercase for 3.0 connector bug
            if(encoding == null) {
                encoding = (String)attr.get("ENC");
            }
            if(encoding != null) {
                Log.debug("EmailData.parse: encoding from server= " + encoding);
                if(StringUtil.equalsIgnoreCase(encoding,"quoted-printable")) {
                    // TODO: optimize this. Since the connector now does
                    //       not send QP to the client, we will not fall
                    //       into this case.
                    String qp = item.toString();
                    qp = QuotedPrintable.decode(qp.getBytes(), "UTF-8");
                    item = new ChunkedString(qp);
                }
            }
            // 
            MIMEProcessor mp = new MIMEProcessor();
            Message msg = mp.parseMailMessage(item);
            
            msg.setFlags(flags);

            // Process the Funambol extensions for filtering. Example:
            // <Ext><XNam>x-funambol-body</XNam><XVal>15000</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach-n</XNam><XVal>1</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach</XNam><XVal>att1.txt</XVal><XVal>10000</XVal></Ext>
            // XXX: we are now considering x-funambol-body only
            Vector extList = XmlUtil.getTagValues(xml, EXT_TAG);
            
            Multipart multipart = null;
            
            
            
            for(int i=0, l=extList.size(); i<l; i++) {
                ChunkedString ext = (ChunkedString)extList.elementAt(i);
                try {
                    ChunkedString xnam = XmlUtil.getTagValue(ext, "XNam");
                    if(xnam.equals("x-funambol-body")) {
                        String xval = XmlUtil.getTagValue(ext, "XVal").toString();
                        if (xval == null) {
                            Log.info("Incomplete header");
                        }
                        msg.getFlags().setFlag(MessageFlags.PARTIAL, true);
                        //break;
                    }

                  
//ifndef isBlackberry
                    
                    if(xnam.equals("x-funambol-attach-n")){
                        if(msg.isMultipart()){
                            multipart = (Multipart)msg.getContent();
                        }else{
                            multipart = new Multipart();
                            BodyPart bodyPartText = new BodyPart();
                            bodyPartText.addHeader(BodyPart.CONTENT_TRANSFER_ENCODING, "8bit");
                            bodyPartText.setContent(msg.getTextContent());
                            multipart.addBodyPart(bodyPartText);
                        }
                        
                        
                    }
        
                    if(xnam.equals("x-funambol-attach")){
                        Vector vec = XmlUtil.getTagValues(ext, "XVal");
                        BodyPart bodyPartAttach = new BodyPart();
                        bodyPartAttach.setDisposition(BodyPart.CD_ATTACHMENT);
                        bodyPartAttach.addHeader(BodyPart.CONTENT_TRANSFER_ENCODING, "8bit");
                        String name = ((ChunkedString)vec.elementAt(0)).toString();
                        if (name.equals("null")){
                            name = "Att.html";
                        }
                        bodyPartAttach.setFileName(name);
                        bodyPartAttach.setSize(((ChunkedString)vec.elementAt(1)).toString());
                        if (vec.size() == 4){
                            bodyPartAttach.setContentType(((ChunkedString)vec.elementAt(2)).toString());
                            
                            String url = checkUrl(((ChunkedString)vec.elementAt(3)).toString());
                            bodyPartAttach.setAttachUrl(url);
                        }
                        multipart.addBodyPart(bodyPartAttach); 
                    }
                            
//endif   
                } catch (XmlException xe) {
                    xe.printStackTrace();
                    Log.error("[EmailData.parse]Invalid Ext field. Ignoring it");
                }
            }
            
            if(multipart != null){
                msg.setContent(multipart);
                Log.debug("Message is multipart : "+msg.isMultipart());
            }
            // Dump memory stats
            Log.stats("End of MessageManager.parseMessage");

            return msg;
        }
        return null;
    }
    

    /**
     * Utility method to check if a tag is found in a chunked string
     *
     * @param xml a chunked string
     * @param tag to be extracted
     *
     * @return true if the tag is found, false otherwise
     */
    private boolean getXmlFlag(ChunkedString xml, String tag) throws XmlException {
        ChunkedString val = null;

        if(XmlUtil.getTag(xml, tag) > -1) {
            val = XmlUtil.getTagValue(xml, tag);
            if(val.equals("true")) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Utility method to add a boolean xml flag into a String buffer
     * @param out the xml output stream
     * @param tag tag name
     * @param flag tag value (true, false)
     */
    private void addXmlFlag(StringBuffer out, String tag, boolean flag) {
        String val = flag ? "true" : "false" ;
        XmlUtil.addElement(out, tag, val);
    }
    
    /**
     * Utility method to get an xml date out of a chunked string
     * @param xml is the chunked string
     * @param tag the date tag
     * @return the date tag value
     */
    private Date getXmlDate(ChunkedString xml, String tag) throws XmlException {
        String val = null;
        MailDateFormatter df = new MailDateFormatter();

        if(XmlUtil.getTag(xml, tag) > -1) {
            val = XmlUtil.getTagValue(xml, tag).toString();
            return MailDateFormatter.parseUTCDate(val);
        }
        return null;
    }

    private String checkUrl(String url) {
        if(url.startsWith("/")){
            try{
                String syncUrl = ConfigManager.getConfig().getUserAccount().getUrl();
                String prefixUrl = StringUtil.extractAddressFromUrl(syncUrl);
                return prefixUrl + url;                
            } catch (ConfigException cex) {
                cex.printStackTrace();
                Log.error(this,"Extracting attachments url from ext "+cex.toString());
                return null;
            }
            
        }else{
            return url;
        }
    }

    private void print(OutputStream os, String msg) throws IOException {
        os.write(msg.getBytes());
    }
}

