/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.syncml;

import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import com.funambol.util.ChunkedString;
import com.funambol.util.XmlUtil;
import com.funambol.util.MailDateFormatter;
import com.funambol.util.XmlException;

/**
 * This class is able to parse and format the OMA-DS 1.2 Folder object.
 */
public class FolderData {

    // ------------------------------------------------------- Private data

    private static final String FOLDER_TAG = "Folder";
    private static final String NAME_TAG = "name";
    private static final String CREATED_TAG = "created";
    private static final String ROLE_TAG = "role";
    private static final String EXT_TAG = "Ext";
    private static final String EXT_XNAM_TAG = "XNam";
    private static final String EXT_XVAL_TAG = "XVal";

    public static final String INBOX_ROLE = "inbox";
    public static final String OUTBOX_ROLE = "outbox";
    public static final String DRAFTS_ROLE = "drafts";
    public static final String SENT_ROLE = "sent";

    private String name;
    private Date created;
    private String role;

    private Hashtable ext;

    // ------------------------------------------------------- Constructors

    /** 
     * Inizialize an empty instance
     */
    public FolderData() {
        this(null, null, null, null);
    }

    /** 
     * Inizialize an empty instance
     */
    public FolderData(String name, Date created, String role, Hashtable ext) {
        this.name = name;
        this.created = created;
        this.role = role;
        this.ext = ext;
    }

    // ---------------------------------------------------------- Accessors  
    
    /** Returns folder name */
    public String getName() { return name; }

    /** Returns the created time */
    public Date getCreated() { return created; }

    /** Returns the folder role */
    public String getRole() { return role; }

    /** Returns the folder ext */
    public Hashtable getExt() { return ext; }

    /** Returns the folder ext values of the specified field name */
    public Vector getExtValues(String xNam) {
        return (Vector)ext.get(xNam);
    }

    /** Sets folder name */
    public void setName(String name) { this.name = name; }

    /** Sets the created time */
    public void setCreated(Date created) { this.created = created; }

    /** Sets the folder role */
    public void setRole(String role) { this.role = role; }

    /** Sets the folder ext */
    public void setExt(Hashtable ext) { this.ext = ext; }

    /** Add a folder values array */
    public void addExtValues(String xNam, Vector xVal) {
        if(ext == null) {
            ext = new Hashtable();
        }
        ext.put(xNam, xVal);
    }

    // ----------------------------------------------------- Public Methods

    /**
     * Parse the item content.
     */
    public void parse(byte[] syncmlData) throws XmlException {
    	ChunkedString xmlFolder = new ChunkedString(new String(syncmlData));
        if (XmlUtil.getTag(xmlFolder, FOLDER_TAG) >= 0) {
            name = XmlUtil.getTagValue(xmlFolder, NAME_TAG).toString();
            role = XmlUtil.getTagValue(xmlFolder, ROLE_TAG).toString();
            created = MailDateFormatter.parseUTCDate(XmlUtil.getTagValue(xmlFolder,
                    CREATED_TAG).toString());
            // Parse Exts
            Vector xmlExts = XmlUtil.getTagValues(xmlFolder, EXT_TAG);
            for(int i=0; i<xmlExts.size(); i++) {
                ChunkedString extTagValue = (ChunkedString)xmlExts.elementAt(i);
                String xNam = XmlUtil.getTagValue(extTagValue, EXT_XNAM_TAG).toString();
                Vector xVals = XmlUtil.getTagValues(extTagValue, EXT_XVAL_TAG);
                if(ext == null) {
                    ext = new Hashtable();
                }
                this.ext.put(xNam, xVals);
            }
        } else {
            throw new XmlException("Invalid folder from server: " + xmlFolder);
        }
    }

    /**
     * Format the item content.
     */
    public byte[] format() {
        StringBuffer out = new StringBuffer("<Folder>\n");

        XmlUtil.addElement(out, NAME_TAG, name);
        XmlUtil.addElement(out, ROLE_TAG, role);
        XmlUtil.addElement(out, CREATED_TAG, MailDateFormatter.dateToUTC(created));

        // Format exts
        if(ext != null) {
            Enumeration xNames = ext.keys();
            while(xNames.hasMoreElements()) {
                out.append("<Ext>\n");
                String xNam = (String)xNames.nextElement();
                Vector xVals = (Vector)ext.get(xNam);
                XmlUtil.addElement(out, EXT_XNAM_TAG, xNam);
                for(int i=0; i<xVals.size(); i++) {
                    XmlUtil.addElement(out, EXT_XVAL_TAG, ((ChunkedString)xVals.elementAt(i)).toString());
                }
                out.append("</Ext>\n");
            }
        }
        out.append("</Folder>\n");
        return out.toString().getBytes();
    }
}



