/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.syncml;

import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Date;
import java.util.Vector;

import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.MailAccount;
import com.funambol.mailclient.mm.MessageManager;
import com.funambol.mailclient.mm.FormattedEmailStore;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.controller.ServerCapStore;

import com.funambol.syncml.client.BaseSyncSource;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.protocol.SyncMLStatus;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.protocol.DevInf;
import com.funambol.syncml.protocol.Ext;

import com.funambol.mail.Folder;
import com.funambol.mail.AccountFolder;
import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;
import com.funambol.mail.MailException;

import com.funambol.util.ChunkedString;
import com.funambol.util.Log;

/**
 * An implementation of the <i>SyncSource</i> interface, used to send
 * and receive email over the SyncML protocol.
 *
 */
public class MailSyncSource extends BaseSyncSource {

    private static final String TAG_LOG = "MailSyncSource";
    
    public static final String FOLDER_TYPE = "application/vnd.omads-folder+xml";
    public static final String EMAIL_TYPE = "application/vnd.omads-email+xml";
    
    //--------------------------------------------------------------- Attributes
    
    /** Flag used to mark the beginning of the receiving phase */
    private boolean receiving;
    
    private Hashtable updatedMsgFlagsBeforeSync;

    private boolean supportMultipleEmailAccount = false;

    //------------------------------------------------------------- Constructors
    
    /**
     * MailSyncSource constructor: initialize source config and
     * init all the rest to null. The real initialization is done by
     * the beginSync method.
     */
    public MailSyncSource(SourceConfig config) {
        super(config);
        globalStatus = STATUS_SUCCESS;
        receiving = false;
        updatedMsgFlagsBeforeSync = new Hashtable();
    }
    
    //----------------------------------------------------------- Public Methods
    
    /**
     * Called after SyncManager preparation and initialization just before start
     * the synchronization of the SyncSource.
     *
     * @param syncMode the synchronization type: one of the values in
     *                 sync4j.framework.core.AlertCode
     *
     * @throws SyncException in case of error. This will stop the sync process
     */
    public void beginSync(int syncMode) throws SyncException {

        Log.info(TAG_LOG, "Beginning sync");
        // Check if this server supports smart slow sync
        boolean supported = ServerCapStore.getInstance().isExtSupported(Ext.X_FUNAMBOL_SMARTSLOW);
        if (!supported) {
            String msg = "Server does not support smart slow sync, or unknown server capabilities";
            Log.info(TAG_LOG, msg);
            throw new SyncException(SyncException.SMART_SLOW_SYNC_UNSUPPORTED, msg);
        }


        //
        // Backward compatibility handling: sync with servers which don't
        // support the Multiple Email Account features:
        //  - create the default account (from the current config), if it
        //    doesn't already exist;
        //  - create default folders (Inbox, Outbox, Sent, Drafts)
        //
        MessageManager mm = MessageManager.getInstance();
        ServerCapStore store = ServerCapStore.getInstance();
        supportMultipleEmailAccount = store.isExtSupported(Ext.X_FUNAMBOL_MULTIPLE_EMAIL_ACCOUNT);
        if(!supportMultipleEmailAccount) {
            AccountFolder[] exAccounts = mm.getAccounts();
            if(exAccounts == null || exAccounts.length == 0) {
                Log.debug(TAG_LOG, "Accounts not found: create a default one.");
                MailAccount.createDefaultAccount();
            }
        }
        
        super.beginSync(syncMode);      // call BaseSyncSource.beginSync()
        globalStatus = STATUS_SUCCESS;
        receiving = false;
        // if we're performing a refresh from server we need to reset store
        if ((syncMode == SyncML.ALERT_CODE_REFRESH_FROM_SERVER)) {
            Log.debug(TAG_LOG, "Resetting inbox folders...");
            Folder[] inboxes = mm.getFoldersFromRole(FolderData.INBOX_ROLE);
            for(int i=0; i<inboxes.length; i++) {
                mm.emptyFolder(inboxes[i]);
            }
            FlagQueue.clear();
            // The store has been reset: create a default account
            Log.debug(TAG_LOG, "Inbox folders reset");
        }
    }

    /**
     * Called just before committing the synchronization process by the
     * SyncManager. The SyncSource can stop the commit phase raising an
     * exception here.
     *
     * @throws SyncException in case of error, to stop the commit.
     */
    public void endSync() throws SyncException {
        super.endSync();        // call BaseSyncSource.endSync() 
        updatedMsgFlagsBeforeSync = null;
    }

    /**
     * Add a new item from the server to the mail store.
     */
    public int addItem(SyncItem item) throws SyncException {
        Log.info(TAG_LOG, "New item " + item.getKey() + " from server.");
        MessageManager mm = MessageManager.getInstance();
        if(item.getType().equals(EMAIL_TYPE)) {
            Log.debug(TAG_LOG, "Email item");
            int errorCode;
            try {
                String targetParent = item.getParent();

                if (targetParent == null) {
                    String sourceParent = item.getSourceParent();
                    if (sourceParent != null) {
                        Log.debug(TAG_LOG + "Source Parent is null");
                        // The item is expressed as source parent. This should
                        // never happen with a MEA server. We assume this is an
                        // item for the inbox
                        targetParent = "/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/Inbox";
                    }
                }

                Log.debug(TAG_LOG + ".addItem IN");

                targetParent = MailSyncItemHelper.filterDefaultAccount(targetParent, false);
                Message msg = mm.createNewMessage(targetParent,
                                                  item.getContent());
                Log.debug(TAG_LOG, "message created");
                if (msg != null) {
                    item.setKey(msg.getKey());
                    Log.debug(TAG_LOG, "ADD ITEM: key before reset " + item.getKey());
                    item.setClientRepresentation(msg);
                    errorCode = SyncMLStatus.SUCCESS;
                } else {
                    Log.info(TAG_LOG, "Invalid empty email data from item " + item.getKey() );
                    globalStatus = STATUS_SERVER_ERROR;
                    errorCode = SyncMLStatus.GENERIC_ERROR;
                }
                item.setKey(MailSyncItemHelper.filterDefaultAccount(msg.getKey(), true));
                targetParent = MailSyncItemHelper.filterDefaultAccount(targetParent, true);
            } catch (Exception e) {
                Log.error(TAG_LOG, "Error saving message: " + item.getKey() 
                            + " " + e.toString() );
                
                e.printStackTrace();
                // TODO: implement a finer handling of the exception?
                globalStatus = STATUS_RECV_ERROR;
                errorCode = SyncMLStatus.GENERIC_ERROR;
            }
            return errorCode;
            
        } else if(item.getType().equals(FOLDER_TYPE)) {
            Log.debug(TAG_LOG, "Folder item");
            int errorCode;
            try {
                Log.debug(TAG_LOG, "THIS ITEM IS A FOLDER: " + item.getKey());
                String folderKey = item.getKey();
                // Handle backward compatibility
                if(isOldFolderKey(folderKey)) {
                    errorCode = SyncMLStatus.SUCCESS;
                    // Doesn't create folders from server, they hve been
                    // automatically created by the client.
                    if(folderKey.equals("ROOT/I") || folderKey.equals("Inbox")) {
                        item.setParent("/");
                        if (supportMultipleEmailAccount) {
                            item.setKey("/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/Inbox");
                        } else {
                            item.setKey("/Inbox");
                        }
                        item.setClientRepresentation(mm.getFolder("/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/Inbox"));
                    } else if(folderKey.equals("ROOT/O") || folderKey.equals("Outbox")) {
                        item.setParent("/");
                        if (supportMultipleEmailAccount) {
                            item.setKey("/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/Outbox");
                        } else {
                            item.setKey("/Outbox");
                        }
                        item.setClientRepresentation(mm.getFolder("/" + MailAccount.DEFAULT_ACCOUNT_NAME + "/Outbox"));
                    } else {
                        Log.info(TAG_LOG, "Invalid folder data from item " +
                            item.getKey() );
                        globalStatus = STATUS_SERVER_ERROR;
                        errorCode = SyncMLStatus.GENERIC_ERROR;
                    }
                } else {
                    Folder folder = mm.createFolder(item.getParent(),  item.getContent());
                    if (folder != null) {
                        item.setKey(folder.getFullName());
                        item.setClientRepresentation(folder);
                        errorCode = SyncMLStatus.SUCCESS;
                    } else {
                        Log.info(TAG_LOG, "Invalid empty folder data from item " +
                                item.getKey() );
                        globalStatus = STATUS_SERVER_ERROR;
                        errorCode = SyncMLStatus.GENERIC_ERROR;
                    }
                }
            } catch (Exception e) {
                Log.error(TAG_LOG,"Error saving folder: " + item.getKey()
                            + " " + e.toString() );
                e.printStackTrace();
                globalStatus = STATUS_RECV_ERROR;
                errorCode = SyncMLStatus.GENERIC_ERROR;
            }
            return errorCode;
        } else {
            Log.error(TAG_LOG,"Unknown item type: " + item.getType());
            globalStatus |= STATUS_RECV_ERROR;
            return SyncMLStatus.GENERIC_ERROR;
        }
    }

    /** Update a given SyncItem stored on the source backend */
    public int updateItem(SyncItem item) throws SyncException {
        Log.debug(TAG_LOG + ".updateItem IN - key");
        item.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), false));

        Log.debug(TAG_LOG + ".updateItem IN - parent");
        item.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), false));
        Log.info(TAG_LOG, "Updated item " + item.getKey() + " from server.");
        MessageManager mm = MessageManager.getInstance();
        if (item.getType().equals(EMAIL_TYPE)) {
            Log.debug(TAG_LOG, "Email item");
            int errorCode;
            try {
                MessageFlags flags = mm.updateMessageFlags(item.getParent(),
                                                           item.getKey(),
                                                           item.getContent());
                item.setClientRepresentation(flags);
                errorCode = SyncMLStatus.SUCCESS;
            } catch (Exception e) {
                Log.error(TAG_LOG, "Error updating message flags " + e.toString());
                globalStatus = STATUS_RECV_ERROR;
                errorCode = SyncMLStatus.GENERIC_ERROR;
            }
            return errorCode;
        } else if (item.getType().equals(FOLDER_TYPE)) {
            Log.debug(TAG_LOG, "Folder item");
            int errorCode;
            try {
                Folder folder = mm.updateFolder(item.getParent(), item.getContent(), item.getKey());
                if (folder != null) {
                    item.setClientRepresentation(folder);
                    errorCode = SyncMLStatus.SUCCESS;
                } else {
                    Log.info(TAG_LOG, "Invalid empty folder data from item " +
                            item.getKey() );
                    globalStatus = STATUS_SERVER_ERROR;
                    errorCode = SyncMLStatus.GENERIC_ERROR;
                }
            } catch (Exception e) {
                Log.error(TAG_LOG, "[updateItem] Error updating folder: " + item.getKey()
                            + " " + e.toString() );
                e.printStackTrace();
                globalStatus = STATUS_RECV_ERROR;
                errorCode = SyncMLStatus.GENERIC_ERROR;
            }
            return errorCode;
        } else {
            Log.error(TAG_LOG, "Unexpected item type" + item.getType());
            globalStatus |= STATUS_RECV_ERROR;
            return SyncMLStatus.GENERIC_ERROR;
        }
    }

    /** Delete a SyncItem stored on the related Items list */
    public int deleteItem(String key) throws SyncException {
        Log.info(TAG_LOG, "Delete from server for item " + key);

        try {
            MessageManager mm = MessageManager.getInstance();
            Log.debug(TAG_LOG + ".deleteItem IN - key");
            key = MailSyncItemHelper.filterDefaultAccount(key, false);
            if(isFolderKey(key)) {
                Log.debug(TAG_LOG, "Folder item");
                mm.deleteFolder(key);
            } else {
                Log.debug(TAG_LOG, "Email item");
                mm.deleteMessage(key);
                updatedMsgFlagsBeforeSync.remove(key);
            }

            return SyncMLStatus.SUCCESS;
        } catch (MailException me) {
            Log.error(TAG_LOG, "Error removing message: " + key + " " + me.toString());
            globalStatus = STATUS_RECV_ERROR;
            return SyncMLStatus.GENERIC_ERROR;
        }
    }

    /**
     * Tell the SyncSource the status returned by the server
     * for an Item previously sent.
     *
     * @param key the key of the item
     * @param status the status code received for that item
     *
     * @throws SyncException if the SyncSource wants to stop the sync
     */
    public void setItemStatus(String key, int status)
            throws SyncException {

        Log.debug(TAG_LOG + ".setItemStatus OUT - key");
        key = MailSyncItemHelper.filterDefaultAccount(key, false);

        Log.info(TAG_LOG, "Status " + status + " for item " + key + " from server.");
        if(status == SyncMLStatus.CHUNKED_ITEM_ACCEPTED) {
            // This is a returned status code for an item chunk
            // return without changing anything
            return;
        }
        try {
            MessageFlags newMask = new MessageFlags();
            MessageManager mm = MessageManager.getInstance();
            String folderPath = mm.folderFromKey(key);
            if(folderPath == null) {
                return;
            }
            Folder itemFolder = mm.getFolder(folderPath);
            if (FolderData.OUTBOX_ROLE.equals(itemFolder.getRole())) {
                if (SyncMLStatus.isSuccess(status)) {
                    mm.messageEffectivelySent(key);
                    // We can clear the formatted email store now
                    try {
                        FormattedEmailStore fes = FormattedEmailStore.getInstance();
                        fes.removeMessage(key);
                    } catch (Exception e) {
                        Log.error(TAG_LOG, "Cannot clean formatted email store " + e.toString());
                    }
                } else {
                    globalStatus = STATUS_SEND_ERROR;
                    Log.error(TAG_LOG, "Cannot send email, received " + status + " from server");
                    // TODO: send an event to the client?
                }
            }
            if (itemFolder.getRole().equals(FolderData.INBOX_ROLE)) {
                newMask = FlagQueue.get(key);
                if (newMask != null) {
                    Log.debug(TAG_LOG, "setItemStatus " + newMask.toString());
                }
                // Changes the delayed messaflags if a 200 is returned
                if (newMask != null && !newMask.isSet(MessageFlags.DELETED)) {
                    Message msg = mm.getMessage(itemFolder, key);

                    if (msg != null) {
                        Log.debug(TAG_LOG, "Msg is not null: Changing fwd/re flags");
                        // we've received a 200 for our update of the flags, 
                        // we've sent the mask stored in the flagqueue so we
                        // can commit this flags (replied or forwarded)
                        Log.debug(TAG_LOG, "Message Flags before update");
                        Log.debug(msg.getFlags().toString());

                        msg.setFlags(newMask);

                        Log.debug(TAG_LOG, "Message Flags after update");
                        Log.debug(msg.getFlags().toString());

                        mm.updateMessage(msg);
                    }
                }

                //return the messageflags of the updated item before the sync
                //this avoid the changes during sync to be lost
                //they will be sent on the next sync
                MessageFlags uibf = (MessageFlags) updatedMsgFlagsBeforeSync.get(key);

                if (uibf != null) {
                    if (isOnlyDelayedFlagChanged(uibf, newMask)) {
                        FlagQueue.remove(key);
                        Log.debug(TAG_LOG, "Item removed: " + key);
                        Log.debug(TAG_LOG, "setItemStatus Flags before sync were: " + uibf.toString());
                        Log.debug(TAG_LOG, "setItemStatus Flags after sync: " + newMask.toString());
                    } else {
                        /* Leave commented in case of more debug issues
                        MessageFlags mf = mm.getMessage(key).getFlags();
                        //compares the new messageflags with the mask
                        Log.debug("[MSS] Comparison VS FQ");
                        int[] compare2 = mf.compareFlags(FlagQueue.get(key));
                        for (int i=0; i<compare2.length; i++) {
                        Log.debug("[MSS] " + i + " = " + compare2[i]);
                        }*/
                        Log.debug(TAG_LOG, "setItemStatus Message " + key
                                  + " has changed during the sync: must not removed from FlagQueue");
                        Log.debug(TAG_LOG, "setItemStatus Flags before sync were: " + uibf.toString());
                        Log.debug(TAG_LOG, "setItemStatus Flags after sync: " + newMask.toString());
                    }
                }

            }
        } catch (MailException me) {
            me.printStackTrace();
            Log.error(TAG_LOG, "setItemStatus: error setting the status for item:" + key);
        }
    }

    private boolean isOnlyDelayedFlagChanged(MessageFlags mf1, MessageFlags mf2) {

        Log.trace(TAG_LOG, "isOnlyDelayedFlagChanged check");

        boolean result = false;
        /* Leave commented in case of more debug issues
        int[] comp = mf1.compareFlags(mf2);
        for (int i=0; i<comp.length; i++) {
        Log.debug("[MSS] " + i + " = " + comp[i]);
        }
         */
        result = (mf1.isSet(MessageFlags.OPENED) == mf2.isSet(MessageFlags.OPENED) &&
                mf1.isSet(MessageFlags.FLAGGED) == mf2.isSet(MessageFlags.FLAGGED));
        Log.trace(TAG_LOG, "isOnlyDelayedFlagChanged ended");
        return result;
    }

    /**
     * Send incomingMsgNumber event when the engine announce the number of
     * changes: it is the first Sync tag from server.
     */
    public void setServerItemsNumber(int number) {
        super.setServerItemsNumber(number);
        receiving = true;
    }

    //------------------------------------------------------- Protected methods
    /**
     * Init the allItems list used in the slow sync. It collects all the
     * <code>AccountFolder</code>, the Inbox <code>Folder</code> and Outbox
     * <code>Folder</code> for each account, and all the Inbox messages.
     */
    protected void initAllItems() throws SyncException {

        //
        //Sends the inbox messages headers list to the server in order to
        //receive only the new items during the slow sync
        //
        try {

            //
            // This is a collector for all the items that need to be synced:
            // accounts, folders and messages.
            //
            Vector syncItems = new Vector();

            // Get the accounts
            MessageManager mm = MessageManager.getInstance();
            AccountFolder[] accounts = mm.getAccounts();

            String folderParentName = "/";

            for(int i=0; i<accounts.length; i++) {
                FolderData folderData = new FolderData(accounts[i].getName(),
                            new Date(), accounts[i].getRole(), null);
                // Set VisibleName and EmailAddress exts
                Vector vector = new Vector();
                vector.addElement(new ChunkedString(accounts[i].getVisibleName()));
                folderData.addExtValues(AccountFolder.VISIBLE_NAME, vector);
                vector = new Vector();
                vector.addElement(new ChunkedString(accounts[i].getEmailAddress()));
                folderData.addExtValues(AccountFolder.EMAIL_ADDRESS, vector);

                // Add an AccountFolder item if supported
                if(supportMultipleEmailAccount) {
                    syncItems.addElement(new SyncItem(accounts[i].getFullName(),
                            FOLDER_TYPE, SyncItem.STATE_UPDATED,
                            mm.getFolderParent(accounts[i]).getFullName(),
                            folderData.format()));
                }

                // Get the Inbox folder for this account
                Folder inbox = mm.getAccountFolderFromRole(accounts[i], FolderData.INBOX_ROLE);
                folderData = new FolderData(inbox.getName(), new Date(),
                        inbox.getRole(), null);

                // Add the Inbox Folder
                folderParentName = supportMultipleEmailAccount ?
                    mm.getFolderParent(inbox).getFullName() : "/";
                Log.debug(MailSyncItemHelper.filterDefaultAccount(inbox.getFullName(), true));
                syncItems.addElement(new SyncItem(MailSyncItemHelper.filterDefaultAccount(inbox.getFullName(), true),
                        FOLDER_TYPE, SyncItem.STATE_UPDATED,
                        folderParentName,
                        folderData.format()));

                // Get the Outbox folder for this account
                Folder outbox = mm.getAccountFolderFromRole(accounts[i], FolderData.OUTBOX_ROLE);
                folderData = new FolderData(outbox.getName(), new Date(),
                        outbox.getRole(), null);

                // Add the Outbox Folder
                folderParentName = supportMultipleEmailAccount ?
                    mm.getFolderParent(outbox).getFullName() : "/";
                syncItems.addElement(new SyncItem(MailSyncItemHelper.filterDefaultAccount(outbox.getFullName(), true),
                        FOLDER_TYPE, SyncItem.STATE_UPDATED,
                        folderParentName,
                        folderData.format()));

                // Collect the messages only from the Inbox folder
                Vector messages = mm.getMessagesInFolder(inbox, null);
                for(int j=0; j<messages.size(); j++) {
                    // Add a Message item
                    Message msg = (Message)messages.elementAt(j);
                    Log.debug(TAG_LOG + ".initAllItems OUT - key");
                    String key = MailSyncItemHelper.filterDefaultAccount(msg.getKey(), true);

                    Log.debug(TAG_LOG + ".initAllItems OUT - parent");
                    String parent = MailSyncItemHelper.filterDefaultAccount(msg.getParent().getFullName(), true);
                    syncItems.addElement(new SyncItem(key,
                            EMAIL_TYPE, SyncItem.STATE_UPDATED,
                            parent, null));
                }
                 
            }
            allItems = new SyncItem[syncItems.size()];
            syncItems.copyInto(allItems);
        } catch (MailException me) {
            Log.error(TAG_LOG, "initAllItems MailException: " + me.toString());
            me.printStackTrace();
            throw new SyncException(SyncException.CLIENT_ERROR, me.toString());
        }
    }

    /**
     * Init the newItems list used in the fast sync with the messages
     * in the outbox only.
     * TODO: check the state of the message
     */
    protected void initNewItems() throws SyncException {
        try {
            // Get the keys of the messages in the Outbox only.
            MessageManager mm = MessageManager.getInstance();
            Vector syncItems = new Vector();
            
            // Get the Outbox folders
            Folder[] outboxes = mm.getFoldersFromRole(FolderData.OUTBOX_ROLE);
            
            for(int i=0; i<outboxes.length; i++) {
                Vector messages = mm.getMessagesInFolder(outboxes[i], null);
                for(int j=0; j<messages.size(); j++) {
                    Message msg = (Message)messages.elementAt(j);
                    syncItems.addElement(new SyncItem(MailSyncItemHelper.filterDefaultAccount(msg.getKey(), true),
                            EMAIL_TYPE, SyncItem.STATE_NEW,
                            MailSyncItemHelper.filterDefaultAccount(msg.getParent().getFullName(), true), null));
                }
            }
            newItems = new SyncItem[syncItems.size()];
            syncItems.copyInto(newItems);
        } catch (MailException me) {
            Log.error(TAG_LOG, "initNewItems: " + me.toString());
            me.printStackTrace();
            throw new SyncException(SyncException.CLIENT_ERROR, me.toString());
        }
    }

    /**
     * Not used now. It will be implemented if the Draft folder
     * has to be synchronized.
     */
    protected void initUpdItems() throws SyncException {
        Log.debug(TAG_LOG, "Initializating Updated Items of folder INBOX");
        int count = 0;
        updItems = null;

        Hashtable updatedItemsList = FlagQueue.getUpdatedItemsList();
        Log.debug(TAG_LOG, updatedItemsList.toString());
        if (updatedItemsList == null) {
            return;
        }

        count = updatedItemsList.size();

        if (count > 0) {
            int i = 0;

            updItems = new SyncItem[count];
            Enumeration itemKeys = updatedItemsList.keys();
            while (itemKeys.hasMoreElements()) {
                String key = (String) itemKeys.nextElement();
                MessageManager mm = MessageManager.getInstance();
                MessageFlags mf = mm.getMessage(key).getFlags();

                updatedMsgFlagsBeforeSync.put(key, mf);
                Log.debug(TAG_LOG, "initUpdItems Message [" + key
                          + "] flags before being sent: " + mf.toString());

                String folder = mm.folderFromKey(key);
                updItems[i] = new SyncItem(MailSyncItemHelper.filterDefaultAccount(key, true), EMAIL_TYPE,
                        SyncItem.STATE_UPDATED, MailSyncItemHelper.filterDefaultAccount(folder, true), null);
                i++;
            }
        }
    }

    /**
     * Init the delItems list used in the fast sync
     * TODO: check the state of the message
     */
    protected void initDelItems() throws SyncException {
        Log.debug(TAG_LOG, "Initializating Deleted Items");
        int count = 0;
        delItems = null;
        Hashtable deletedItemsList = FlagQueue.getDeletedItemsList();

        if (deletedItemsList == null) {
            return;
        }
        count = deletedItemsList.size();
        if (count > 0) {
            delItems = new SyncItem[count];
            Enumeration itemKeys = deletedItemsList.keys();
            int i = 0;
            while (itemKeys.hasMoreElements()) {

                String key = (String) itemKeys.nextElement();
                String folder = MessageManager.getInstance().folderFromKey(key);
                Log.debug(TAG_LOG, "Folder to find: " + folder);
                delItems[i] = new SyncItem(MailSyncItemHelper.filterDefaultAccount(key,true), EMAIL_TYPE,
                        SyncItem.STATE_DELETED, MailSyncItemHelper.filterDefaultAccount(folder, true), null);
                i++;
            }
        }
    }

    /**
     * This function gets the item content in the backend database and
     * returns a complete item. The parameter item is marked final because
     * should not be used for the filled item: it is a reference to the
     * array entry, and filling it would cause the array to keep all the
     * filled items at the end (the gc will not dispose them). <p>
     * The content of the item depends also from the encoding of this
     * SyncSource:
     * <li> if the encoding is <i>none</i>, it must be a String, converted
     *      with getBytes(), so the engine will send it unchanged.
     * <li> if the encoding is <i>b64</i>, the content can be binary, and the
     *      type should be set accordingly, so that the receiving source
     *      can handle it. In this way, the binary content is transferred
     *      encoded in the SyncML message. This encoding can be applied to
     *      a text item too, to avoid problems with charset or other, like
     *      what is done with the SIF format.
     */
    protected SyncItem getItemContent(final SyncItem item) throws SyncException {
        Log.debug(TAG_LOG, "Current Item State: " + item.getState());
        try {
            SyncItem ret = null;
            MessageManager mm = MessageManager.getInstance();
            switch (item.getState()) {
                case SyncItem.STATE_UPDATED:
                    if (item.getType().equals(EMAIL_TYPE)) {
                        Log.debug(TAG_LOG + ".getItemContent OUT - UPDATED - parent");
                        item.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), false));
                        item.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), false));
                        Folder folder = mm.getFolder(item.getParent());
                        if (folder.getRole().equals(FolderData.INBOX_ROLE)) {
                            byte format[];
                            if (syncMode != SyncML.ALERT_CODE_SLOW) {
                                //format the message flags to propagate on the 
                                //server for fast sync: refers to the flagqueue
                                EmailData ed = new EmailData();
                                format = ed.formatFlags(FlagQueue.get(item.getKey()));
                            } else {
                                // we send the inbox messages only during the 
                                // slow sync, and we need to send only message ids
                                // and flags, no other content is needed.
                                Message storedMsg = mm.getMessage(folder, item.getKey());

                                // to avoid sending the entire message we copy
                                // the msgid and flags into a new empty message.
                                if(storedMsg != null){
                                    EmailData ed =new EmailData();
                                    format = ed.formatMinimalInfo(storedMsg);
                                }else{
                                    format = null;
                                }
                            }

                            item.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), true));
                            item.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), true));
                            ret = new SyncItem(item);
                            ret.setContent(format);
                            return ret;
                        }
                        break;
                    }
                case SyncItem.STATE_NEW:
                    if (item.getType().equals(EMAIL_TYPE)) {
                        // For type email, the content has to be filled,
                        // for folders, the item is already complete.
                        // for deleted items
                        Log.debug(TAG_LOG + ".getItemContent OUT - NEW - parent");
                        item.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), false));
                        item.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), false));
                        Folder f = mm.getFolder(item.getParent());
                        if (f.getRole().equals(FolderData.OUTBOX_ROLE)) {
                            Message msg = mm.getMessage(f, item.getKey());

                            //Sent real sent date
                            Date newSentDate = new Date();
                            mm.updateMessageDates(msg, newSentDate, newSentDate);

                            EmailData ed = new EmailData();
                            byte format[] = ed.format(msg);

                            // Check if for this message we already have a
                            // preformatted version
                            FormattedEmailStore fes = FormattedEmailStore.getInstance();
                            if (fes.contains(msg)) {
                                try {
                                    InputStream is = fes.getMessageInputStream(msg.getKey());
                                    ret = new EmailSyncItem(item.getKey(),
                                            item.getType(),
                                            item.getState(),
                                            item.getParent(),
                                            is);
                                    ret.setObjectSize(is.available());
                                    ret.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), true));
                                    ret.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), true));
                                    return ret;
                                } catch (IOException ioe) {
                                    Log.error(TAG_LOG, "Cannot read preformatted email, trying to format in memory");
                                }
                            }
                            item.setParent(MailSyncItemHelper.filterDefaultAccount(item.getParent(), true));
                            item.setKey(MailSyncItemHelper.filterDefaultAccount(item.getKey(), true));
                            ret = new SyncItem(item);
                            ret.setContent(format);
                            Log.trace(TAG_LOG, "getItemContent: returning filled item.");

                            // Return the filled item
                            return ret;
                        }
                    }
                    break;
                default:
                    break;
            }

        } catch (MailException me) {
            Log.error(TAG_LOG, "getItemContent Error filling item:" + item.getKey());
            me.printStackTrace();
            throw new SyncException(SyncException.CLIENT_ERROR, me.toString());
        }

        // If no changes, return the original item
        Log.debug(TAG_LOG, "getItemContent: returning original item.");
        return item;
    }

    /**
     * get the updated message flags list at the begin of the sync
     * @return the list of updated message flags when the sync started
     */
    public Hashtable getUpdatedMsgFlagsBeforeSync() {
        return updatedMsgFlagsBeforeSync;
    }

    // Check if the given path key represents a folder or not
    private boolean isFolderKey(String key) {
        int firstSlashPos = key.indexOf("/");
        if(firstSlashPos == 0) {
            int secondSlashPos = key.substring(firstSlashPos+1).indexOf("/");
            if(secondSlashPos != -1) {
                int thirdSlashPos = key.substring(secondSlashPos+2).indexOf("/");
                if(thirdSlashPos != -1) {
                    // It represents an email item
                    return false;
                }
            }
            return true;
        }
        return false;
    }

    // Check if the specified key represents folder key sent by an old server
    // which doesn't support multiple email account features
    public static boolean isOldFolderKey(String key) {
        return key.indexOf("ROOT/") != -1 || isOldLocalFolderId(key);
    }

    private static boolean isOldLocalFolderId(String id) {
        return "Inbox".equals(id) || "Outbox".equals(id);
    }

    private class EmailSyncItem extends SyncItem {
        private InputStream is;

        public EmailSyncItem(String key, String type, char state,
                             String parent, InputStream is) {
            super(key, type, state, parent);
            this.is = is;
        }

        public InputStream getInputStream() {
            return is;
        }
    }
}
