/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.mm;

import com.funambol.mail.AccountFolder;
import com.funambol.mail.Address;
import com.funambol.mail.Folder;
import com.funambol.mail.MIMEFormatter;
import com.funambol.mail.MailException;
import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;
import com.funambol.mail.Store;
import com.funambol.mail.StoreFactory;
import com.funambol.mailclient.UserAccount;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.mailclient.syncml.FlagQueue;
import com.funambol.mailclient.syncml.FolderData;
import com.funambol.storage.ObjectStore;
import com.funambol.util.ConsoleAppender;
import com.funambol.util.Log;
import java.io.IOException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.midlet.MIDlet;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreException;
import junit.framework.*;

/**
 * Test case for the MessageManager class.
 * The tests initialize the message store, perform the initial sync to
 * get what is on the server, send a test message to self, get the response
 * and remove the sent items.
 */
public class MessageManagerTest extends TestCase {

    private static final String INBOX_NAME = "Inbox";
    private static final String OUTBOX_NAME = "Outbox";
    private static final String DRAFTS_NAME = "Drafts";
    private static final String SENT_NAME = "Sent";
    private static final String ACCOUNT_NAME = "DefaultAccount";
    private static final String ACCOUNT_PATH = Folder.ROOT_FOLDER_PATH + ACCOUNT_NAME;
    private static final String INBOX_PATH = ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + INBOX_NAME;
    private static final String OUTBOX_PATH = ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + OUTBOX_NAME;
    private static final String SENT_PATH = ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + SENT_NAME;
    private static final String DRAFTS_PATH = ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + DRAFTS_NAME;
    private static final String ACCOUNT_EMAIL_ADDRESS = "account@funambol.com";
    private static final String ACCOUNT_VISIBLE_NAME = "Funambol";
    private static final String TEST_ACCOUNT_PATH = "/TestAccount";
    private static final String TEST_ACCOUNT_NAME = "TestAccount";
    private static final String TEST_ACCOUNT_EMAIL_ADDRESS = "test_account@funambol.com";
    private static final String TEST_ACCOUNT_VISIBLE_NAME = "Funambol_Test";
    private static final String TEST_FOLDER_PATH = "/Folder";
    private static final String TEST_FOLDER_NAME = "Folder";
    private static final String TEST_INBOX_PATH = TEST_ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + "Inbox";
    private static final String TEST_OUTBOX_PATH = TEST_ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + "Outbox";
    private static final String TEST_SENT_PATH = TEST_ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + "Sent";
    private static final String TEST_DRAFTS_PATH = TEST_ACCOUNT_PATH + Folder.FOLDER_SEPARATOR + "Drafts";
    private static final String ADDRESS_FROM = "from@funambol.com";
    private static final String ADDRESS_TO = "to@funambol.com";
    private static final String ADDRESS_CC = "cc@funambol.com";
    private static final String ADDRESS_BCC = "bcc@funambol.com";
    private static final String ADDRESS_REPLY_TO = "replyTo@funambol.com";
    private static final String MESSAGE_CONTENT = "generic content";
    private static final String INBOX_MESSAGE_CONTENT = "Inbox Content";
    private static final String OUTBOX_MESSAGE_CONTENT = "Outbox Content";
    private static final String DRAFTS_MESSAGE_CONTENT = "Drafts Content";
    private static final String SENT_MESSAGE_CONTENT = "Sent Content";
    ObjectStore confStore = new ObjectStore();
    Store store = null;
    MessageManager mm = null;

    TestUnreadMessagesListener unreadMessageListener = new TestUnreadMessagesListener();

    int getMessagesNotificationsCount = 0;
    Hashtable getMessagesNotifications = new Hashtable();

    /**
     * Default constructor. Set the log level for all tests
     */
    public MessageManagerTest(String name) {
        super(name);
        Log.initLog(new ConsoleAppender());
        Log.setLogLevel(Log.INFO);
    }

    /**
     * Set up the tests:
     *  - Create the MessageManager instance 
     *  - Retrieve the Store instance 
     *  - Initializes a default account with default folders
     *  - Use the ConfigManager class to initialize the default configuration 
     *    creating a default account with default folders 
     *    (inbox-outbox-sent-drafts)
     */
    public void setUp() {

        mm = MessageManager.getInstance();
        mm.initStore(true);
        mm.setUnreadMessagesListener(unreadMessageListener);

        store = StoreFactory.getStore();
        initDefaultAccount();
        initDefaultConfiguration();

        unreadMessageListener.reset();

        getMessagesNotifications.clear();
        getMessagesNotificationsCount = 0;
    }

    /**
     * Reset the system folders removing them from the system
     */
    public void tearDown() {
        store.removeFolder("/");
        store = null;
    }

    private class TestUnreadMessagesListener implements UnreadMessagesListener {

        public int notificationsCount = 0;
        public Folder lastFolder = null;
        public int lastCount = 0;
        
        public void updateUnreadMessages(Folder folder, int count) {
            lastFolder = folder;
            lastCount = count;
            notificationsCount++;
        }

        public void reset() {
            notificationsCount = 0;
            lastFolder = null;
            lastCount = 0;
        }
    }

    /**
     * Test the singleton pattern implementation
     * @throws java.lang.Exception
     */
    public void testGetInstance() throws Exception {
        Log.info("[MessageManagerTest.testGetInstance]Started");
        MessageManager m1 = MessageManager.getInstance();
        assertTrue(m1.equals(mm));
        Log.info("[MessageManagerTest.testGetInstance]Successful");
    }

    /**
     * Test the Message creation
     * @throws java.lang.Exception
     */
    public void testCreateMessage() throws Exception {
        Log.info("[MessageManagerTest.testCreateMessage]Started");

        String folderName = "Inbox";
        String folderPath = ACCOUNT_PATH + "/" + folderName;
        
        String folderData = getSampleFolderData(folderName, FolderData.INBOX_ROLE);
        mm.createFolder(ACCOUNT_PATH, folderData.getBytes());

        String emailData = getSampleEmailData(true);
        mm.createNewMessage(folderPath, emailData.getBytes());

        assertEquals(mm.getMessagesCount(mm.getFolder(folderPath)), 1);

        assertEquals(unreadMessageListener.notificationsCount, 0);

        Vector messages = mm.getMessagesInFolder(folderPath);

        assertEquals(messages.size(), 1);

        // Test the unread messages count notification
        String unreadEmailData = getSampleEmailData(false);
        mm.createNewMessage(folderPath, unreadEmailData.getBytes());
        mm.createNewMessage(folderPath, unreadEmailData.getBytes());
        mm.createNewMessage(folderPath, emailData.getBytes());
        assertTrue(checkUnreadMessageNotification(2, 2, mm.getFolder(folderPath)));

        Log.info("[MessageManagerTest.testCreateMessage]Successful");
    }

    /**
     * Test the folders added in the setup method checking their fields
     * @throws java.lang.Exception
     */
    public void testAddfolder() throws Exception {
        Log.info("[MessageManagerTest.testAddFolder]Started");

        AccountFolder[] af = mm.getAccounts();
        Folder[] folders = mm.getAccountFolders(af[0]);

        Folder[] inboxes = mm.getFoldersFromRole(FolderData.INBOX_ROLE);
        Folder[] outboxes = mm.getFoldersFromRole(FolderData.OUTBOX_ROLE);
        Folder[] drafts = mm.getFoldersFromRole(FolderData.DRAFTS_ROLE);
        Folder[] sent = mm.getFoldersFromRole(FolderData.SENT_ROLE);

        assertTrue(
                af.length == 1 &&
                af[0].getRole().equals(AccountFolder.ACCOUNT_ROLE) &&
                af[0].getEmailAddress().equals(ACCOUNT_EMAIL_ADDRESS) &&
                af[0].getVisibleName().equals(ACCOUNT_VISIBLE_NAME) &&
                folders.length == 4 &&
                inboxes.length == 1 &&
                inboxes[0].getName().equals(INBOX_NAME) &&
                inboxes[0].getRole().equals(FolderData.INBOX_ROLE) &&
                inboxes[0].getFullName().equals(INBOX_PATH) &&
                outboxes.length == 1 &&
                outboxes[0].getName().equals(OUTBOX_NAME) &&
                outboxes[0].getRole().equals(FolderData.OUTBOX_ROLE) &&
                outboxes[0].getFullName().equals(OUTBOX_PATH) &&
                drafts.length == 1 &&
                drafts[0].getName().equals(DRAFTS_NAME) &&
                drafts[0].getRole().equals(FolderData.DRAFTS_ROLE) &&
                drafts[0].getFullName().equals(DRAFTS_PATH) &&
                sent.length == 1 &&
                sent[0].getName().equals(SENT_NAME) &&
                sent[0].getRole().equals(FolderData.SENT_ROLE) &&
                sent[0].getFullName().equals(SENT_PATH));
        Log.info("[MessageManagerTest.testAddFolder]Successful");
    }

    /**
     * Test the getFoldersFromRole method for all the default folders
     * @throws java.lang.Exception
     */
    public void testGetFoldersFromRole() throws Exception {
        Log.info("[MessageManagerTest.testGetFoldersFromRole]Started");
        Folder[] inboxes = mm.getFoldersFromRole(FolderData.INBOX_ROLE);
        Folder[] outboxes = mm.getFoldersFromRole(FolderData.OUTBOX_ROLE);
        Folder[] drafts = mm.getFoldersFromRole(FolderData.DRAFTS_ROLE);
        Folder[] sent = mm.getFoldersFromRole(FolderData.SENT_ROLE);

        assertTrue(
                inboxes.length == 1 &&
                inboxes[0].getName().equals(INBOX_NAME) &&
                inboxes[0].getRole().equals(FolderData.INBOX_ROLE) &&
                inboxes[0].getFullName().equals(INBOX_PATH) &&
                outboxes.length == 1 &&
                outboxes[0].getName().equals(OUTBOX_NAME) &&
                outboxes[0].getRole().equals(FolderData.OUTBOX_ROLE) &&
                outboxes[0].getFullName().equals(OUTBOX_PATH) &&
                drafts.length == 1 &&
                drafts[0].getName().equals(DRAFTS_NAME) &&
                drafts[0].getRole().equals(FolderData.DRAFTS_ROLE) &&
                drafts[0].getFullName().equals(DRAFTS_PATH) &&
                sent.length == 1 &&
                sent[0].getName().equals(SENT_NAME) &&
                sent[0].getRole().equals(FolderData.SENT_ROLE) &&
                sent[0].getFullName().equals(SENT_PATH));

        Log.info("[MessageManagerTest.testGetFoldersFromRole]Successful");
    }

    /**
     * Test the getFolder method on the default folders
     */
    public void testGetFolder() throws Exception {
        Log.info("[MessageManagerTest.testGetFolder]Started");
        Folder inbox = mm.getFolder(INBOX_PATH);
        Folder outbox = mm.getFolder(OUTBOX_PATH);
        Folder drafts = mm.getFolder(DRAFTS_PATH);
        Folder sent = mm.getFolder(SENT_PATH);
        assertTrue(
                inbox.getName().equals(INBOX_NAME) &&
                inbox.getRole().equals(FolderData.INBOX_ROLE) &&
                inbox.getFullName().equals(INBOX_PATH) &&
                outbox.getName().equals(OUTBOX_NAME) &&
                outbox.getRole().equals(FolderData.OUTBOX_ROLE) &&
                outbox.getFullName().equals(OUTBOX_PATH) &&
                drafts.getName().equals(DRAFTS_NAME) &&
                drafts.getRole().equals(FolderData.DRAFTS_ROLE) &&
                drafts.getFullName().equals(DRAFTS_PATH) &&
                sent.getName().equals(SENT_NAME) &&
                sent.getRole().equals(FolderData.SENT_ROLE) &&
                sent.getFullName().equals(SENT_PATH));

        Log.info("[MessageManagerTest.testGetFolder]Successful");
    }

    /**
     * Test the getFolder method with the lightFolder paramter set to true
     */
    public void testGetLightFolder() throws Exception {
        Log.info("[MessageManagerTest.testGetLightFolder]Started");
        Folder inbox = mm.getFolder(INBOX_PATH, true);
        Folder outbox = mm.getFolder(OUTBOX_PATH, true);
        Folder drafts = mm.getFolder(DRAFTS_PATH, true);
        Folder sent = mm.getFolder(SENT_PATH, true);
        assertTrue(
                inbox.getName().equals(INBOX_NAME) &&
                inbox.getRole().equals(FolderData.INBOX_ROLE) &&
                inbox.getFullName().equals(INBOX_PATH) &&
                outbox.getName().equals(OUTBOX_NAME) &&
                outbox.getRole().equals(FolderData.OUTBOX_ROLE) &&
                outbox.getFullName().equals(OUTBOX_PATH) &&
                drafts.getName().equals(DRAFTS_NAME) &&
                drafts.getRole().equals(FolderData.DRAFTS_ROLE) &&
                drafts.getFullName().equals(DRAFTS_PATH) &&
                sent.getName().equals(SENT_NAME) &&
                sent.getRole().equals(FolderData.SENT_ROLE) &&
                sent.getFullName().equals(SENT_PATH));

        Log.info("[MessageManagerTest.testGetLightFolder]Successful");
    }

    /**
     * Test the initStore method passing a boolean true as reset value
     * @throws java.lang.Exception
     */
    public void testInitStoreWithReset() throws Exception {
        Log.info("[MessageManagerTest.testInitStoreWithReset]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));
        mm.deleteMessage(m1);

        mm.initStore(true);
        Folder f = null;

        try {
            f = mm.getFolder(INBOX_PATH);
            fail("[MessageManagerTest.testInitStoreWithReset]" + INBOX_PATH + "Folder found after a reset of the store");
        } catch (MailException me) {
            assertTrue((f == null) && FlagQueue.getDeletedItemsList().isEmpty());
        }

        Log.info("[MessageManagerTest.testInitStoreWithReset]Successful");
    }

    /**
     * Test the initStore method passing a boolean false as reset value
     * @throws java.lang.Exception
     */
    public void testInitStoreNoReset() throws Exception {
        Log.info("[MessageManagerTest.testInitStoreNoReset]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));
        mm.deleteMessage(m1);

        mm.initStore(false);
        Folder f = mm.getFolder(INBOX_PATH);

        assertTrue((f != null) && FlagQueue.getDeletedItemsList().containsKey(m1.getKey()));
        Log.info("[MessageManagerTest.testInitStoreNoReset]Successful");
    }

    /**
     * Test the appendMessageToFolder method and check that the message exist.
     * @throws java.lang.Exception
     */
    public void testAppendMessageToFolder() throws Exception {
        Log.info("[MessageManagerTest.testAppendMessageToFolder]Started");

        Folder inbox = mm.getFolder(INBOX_PATH);

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, inbox);
        String expected = m1.getKey();

        int count = mm.getFolder(INBOX_PATH).getMessageCount();
        Message result = mm.getMessage(expected);

        assertTrue(result != null && count == 1);

        // Test the unread messages count notification
        Message unreadMessage = createUnreadMessage();
        mm.appendMessageToFolder(unreadMessage, inbox);
        mm.appendMessageToFolder(m1, inbox);
        mm.appendMessageToFolder(unreadMessage, inbox);
        assertTrue(checkUnreadMessageNotification(2, 2, inbox));
 
        Log.info("[MessageManagerTest.testAppendMessageToFolder]Successful");
    }

    /**
     * Test the emptyFolder method on the default folder "inbox"
     */
    public void testEmptyFolder() throws Exception {
        Log.info("[MessageManagerTest.testEmptyFolder]Started");
        Folder inbox = mm.getFolder(INBOX_PATH);
        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, inbox);
        Message m2 = createReadMessage();
        mm.appendMessageToFolder(m2, inbox);
        mm.emptyFolder(inbox);

        assertTrue(checkUnreadMessageNotification(0, 0, null));
        
        assertTrue(inbox.getMessageCount() == 0);

        // Test the unread messages count notification
        Message unreadMessage = createUnreadMessage();
        mm.appendMessageToFolder(unreadMessage, inbox);
        assertTrue(checkUnreadMessageNotification(1, 1, inbox));
        mm.emptyFolder(inbox);
        assertTrue(checkUnreadMessageNotification(2, 0, inbox));
        
        Log.info("[MessageManagerTest.testEmptyFolder]Successful");
    }

    /**
     * Test the updateMessage method appending a message to the default inbox
     * and then updating it.
     */
    public void testUpdateMessage() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessage]Started");
        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, mm.getFolder(INBOX_PATH)));
        
        m1.getFlags().setFlag(MessageFlags.OPENED, true);

        mm.updateMessage(m1);
        Message result = mm.getMessage(m1.getKey());
        assertTrue(result.getFlags().isSet(MessageFlags.OPENED));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, mm.getFolder(INBOX_PATH)));

        Log.info("[MessageManagerTest.testUpdateMessage]Successful");
    }

    /**
     * Test the updateMessageFlag method on 7 messages appended to the "inbox". 
     * 7 messages are used in order to check the update af all signicative flags
     */
    public void testUpdateMessageFlagsByIntFlag() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageFlagsByIntFlag]Started");

        //Update the OPENED flag
        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, mm.getFolder(INBOX_PATH)));
        
        mm.updateMessageFlag(m1, MessageFlags.OPENED, true);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, mm.getFolder(INBOX_PATH)));

        //Update the FLAGGED flag
        Message m2 = createReadMessage();
        mm.appendMessageToFolder(m2, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m2, MessageFlags.FLAGGED, true);

        //update the DELETED flag
        Message m3 = createReadMessage();
        mm.appendMessageToFolder(m3, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m3, MessageFlags.DELETED, true);

        //update the ANSWERED flag
        Message m4 = createReadMessage();
        mm.appendMessageToFolder(m4, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m4, MessageFlags.DRAFT, true);

        //update the DRAFT flag
        Message m5 = createReadMessage();
        mm.appendMessageToFolder(m5, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m5, MessageFlags.FORWARDED, true);

        //update the FORWARDED flag
        Message m6 = createReadMessage();
        mm.appendMessageToFolder(m6, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m6, MessageFlags.PARTIAL, true);

        //update the PARTIAL flag
        Message m7 = createReadMessage();
        mm.appendMessageToFolder(m7, mm.getFolder(INBOX_PATH));
        mm.updateMessageFlag(m7, MessageFlags.ANSWERED, true);

        boolean result1 = checkUpdatedMessageFlagsByInt(m1, MessageFlags.OPENED, true);
        boolean result2 = checkUpdatedMessageFlagsByInt(m2, MessageFlags.FLAGGED, true);
        boolean result3 = checkUpdatedMessageFlagsByInt(m3, MessageFlags.DELETED, false);
        boolean result4 = checkUpdatedMessageFlagsByInt(m4, MessageFlags.DRAFT, false);
        boolean result5 = checkUpdatedMessageFlagsByInt(m5, MessageFlags.FORWARDED, false);
        boolean result6 = checkUpdatedMessageFlagsByInt(m6, MessageFlags.PARTIAL, false);
        boolean result7 = checkUpdatedMessageFlagsByInt(m7, MessageFlags.ANSWERED, false);

        assertTrue(result1 &&
                result2 &&
                result3 &&
                result4 &&
                result5 &&
                result6 &&
                result7);

        Log.info("[MessageManagerTest.testUpdateMessageFlagsByIntFlag]Successful");
    }

    /**
     * Test the updateMessageFlags with 7 messages created on the default 
     * "inbox". 7 messages are used to check the validity of this methods for 
     * all the significative flags
     */
    public void testUpdateMessageFlagsByMessageFlagsObject() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageFlagsByMessageFlagsObject]Started");

        //Update the OPENED flag
        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        update(m1, MessageFlags.OPENED);

        //Update the FLAGGED flag
        Message m2 = createReadMessage();
        mm.appendMessageToFolder(m2, store.getFolder(INBOX_PATH));
        update(m2, MessageFlags.FLAGGED);

        //update the DELETED flag
        Message m3 = createReadMessage();
        mm.appendMessageToFolder(m3, store.getFolder(INBOX_PATH));
        update(m3, MessageFlags.DELETED);

        //update the ANSWERED flag
        Message m4 = createReadMessage();
        mm.appendMessageToFolder(m4, store.getFolder(INBOX_PATH));
        update(m4, MessageFlags.DRAFT);

        //update the DRAFT flag
        Message m5 = createReadMessage();
        mm.appendMessageToFolder(m5, store.getFolder(INBOX_PATH));
        update(m5, MessageFlags.FORWARDED);

        //update the FORWARDED flag
        Message m6 = createReadMessage();
        mm.appendMessageToFolder(m6, store.getFolder(INBOX_PATH));
        update(m6, MessageFlags.PARTIAL);

        //update the PARTIAL flag
        Message m7 = createReadMessage();
        mm.appendMessageToFolder(m7, store.getFolder(INBOX_PATH));
        update(m7, MessageFlags.ANSWERED);

        boolean result1 = FlagQueue.get(m1.getKey()).isSet(MessageFlags.OPENED);
        boolean result2 = FlagQueue.get(m2.getKey()).isSet(MessageFlags.FLAGGED);
        boolean result3 = FlagQueue.get(m3.getKey()).isSet(MessageFlags.DELETED);
        boolean result4 = FlagQueue.get(m4.getKey()).isSet(MessageFlags.DRAFT);
        boolean result5 = FlagQueue.get(m5.getKey()).isSet(MessageFlags.FORWARDED);
        boolean result6 = FlagQueue.get(m6.getKey()).isSet(MessageFlags.PARTIAL);
        boolean result7 = FlagQueue.get(m7.getKey()).isSet(MessageFlags.ANSWERED);

        assertTrue(result1 &&
                result2 &&
                result3 &&
                result4 &&
                result5 &&
                result6 &&
                result7);

        Log.info("[MessageManagerTest.testUpdateMessageFlagsByMessageFlagsObject]Successful");
    }

    /**
     * Test the updateMessageDates passing 2 null dates and checking that 
     * nothing changed on the previous message dates
     * @throws java.lang.Exception
     */
    public void testUpdateMessageDatesWithNullDates() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageDatesWithNullDates]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        Date sent = null;
        Date received = null;
        mm.updateMessageDates(m1, sent, received);

        Date expectedSent = m1.getSentDate();
        Date expectedReceived = m1.getReceivedDate();

        Message m = mm.getMessage(m1.getKey());

        assertTrue(m.getSentDate().equals(expectedSent) &&
                (m.getReceivedDate().equals(expectedReceived)));

        Log.info("[MessageManagerTest.testUpdateMessageDatesWithNullDates]Successful");
    }

    /**
     * Test the udpateMessageDates method passing a null sent date and updating 
     * the message's received date
     * @throws java.lang.Exception
     */
    public void testUpdateMessageDatesWithNullSent() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageDatesWithNullSent]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        Date sent = null;
        Date received = new Date(0);
        mm.updateMessageDates(m1, sent, received);

        Date expectedSent = m1.getSentDate();
        Date expectedReceived = received;

        Message m = mm.getMessage(m1.getKey());

        assertTrue(m.getSentDate().equals(expectedSent) &&
                (m.getReceivedDate().equals(expectedReceived)));

        Log.info("[MessageManagerTest.testUpdateMessageDatesWithNullSent]Successful");
    }

    /**
     * Test the udpateMessageDates method passing a null received date and 
     * updating the message's sent date
     * @throws java.lang.Exception
     */
    public void testUpdateMessageDatesWithNullReceived() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageDatesNullReceived]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        Date sent = new Date(0);
        Date received = null;
        mm.updateMessageDates(m1, sent, received);

        Date expectedSent = sent;
        Date expectedReceived = m1.getReceivedDate();

        Message m = mm.getMessage(m1.getKey());

        assertTrue(m.getSentDate().equals(expectedSent) &&
                (m.getReceivedDate().equals(expectedReceived)));

        Log.info("[MessageManagerTest.testUpdateMessageDatesNullReceived]Successful");
    }

    /**
     * Test the udpateMessageDates method passing both new received date and 
     * new sent date
     * @throws java.lang.Exception
     */
    public void testUpdateMessageDatesWithSentAndReceived() throws Exception {
        Log.info("[MessageManagerTest.testUpdateMessageDatesWithSentAndReceived]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        Date sent = new Date(0);
        Date received = new Date(0);
        mm.updateMessageDates(m1, sent, received);

        Date expectedSent = sent;
        Date expectedReceived = received;

        Message m = mm.getMessage(m1.getKey());

        assertTrue(m.getSentDate().equals(expectedSent) &&
                (m.getReceivedDate().equals(expectedReceived)));

        Log.info("[MessageManagerTest.testUpdateMessageDatesWithSentAndReceived]Successful");
    }

    /**
     * Test the removeFromFlagQueue method checking the content of the FlagQueue 
     * object after the call. The test is performed on an updated message
     * @throws java.lang.Exception
     */
    public void testRemoveFromFlagQueueUpdatedMsg() throws Exception {
        Log.info("[MessageManagerTest.testRemoveFromFlagQueueUpdatedMsg]Started");

        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, mm.getFolder(INBOX_PATH)));
        
        mm.updateMessageFlag(m1, MessageFlags.OPENED, true);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, mm.getFolder(INBOX_PATH)));

        boolean isAdded = FlagQueue.getUpdatedItemsList().containsKey(m1.getKey());

        mm.removeFromFlagQueue(m1);

        boolean isRemoved = !FlagQueue.getUpdatedItemsList().containsKey(m1.getKey());

        assertTrue(isAdded && isRemoved);

        Log.info("[MessageManagerTest.testRemoveFromFlagQueueUpdatedMsg]Successful");
    }

    /**
     * Test the removeFromFlagQueue method checking the content of the FlagQueue 
     * object after the call. The test is performed on a deleted message
     * @throws java.lang.Exception
     */
     public void testRemoveFromFlagQueueDeletedMsg() throws Exception {
        Log.info("[MessageManagerTest.testRemoveFromFlagQueueDeletedMsg]Started");

        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, mm.getFolder(INBOX_PATH)));

        mm.updateMessageFlag(m1, MessageFlags.OPENED, true);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, mm.getFolder(INBOX_PATH)));

        boolean isUpdatedMessageAdded = FlagQueue.getUpdatedItemsList().containsKey(m1.getKey());

        mm.deleteMessage(m1);

        boolean isUpdatedMessageRemoved = !FlagQueue.getUpdatedItemsList().containsKey(m1.getKey());

        boolean isDeletedMessageAdded = FlagQueue.getDeletedItemsList().containsKey(m1.getKey());


        assertTrue(isUpdatedMessageAdded &&
                isUpdatedMessageRemoved &&
                isDeletedMessageAdded);

        Log.info("[MessageManagerTest.testRemoveFromFlagQueueDeletedMsg]Successful");
    }

    /**
     * Test the deleteMessage methods passing a Message Object to reference the 
     * message that must be deleted 
     * @throws java.lang.Exception
     */
    public void testDeleteMessageWithMessage() throws Exception {
        Log.info("[MessageManagerTest.testDeleteMessageWithMessage]Started");

        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, store.getFolder(INBOX_PATH)));
        
        mm.deleteMessage(m1);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, store.getFolder(INBOX_PATH)));

        boolean isDeletedMessageAdded = FlagQueue.getDeletedItemsList().containsKey(m1.getKey());

        Message m = null;

        try {
            m = mm.getMessage(m1.getKey());
            fail("[MessageManagerTest.testDeleteMessageWithMessage]Deleted message was found on the store");
        } catch (MailException me) {
        }

        assertTrue(isDeletedMessageAdded && m == null);


        Log.info("[MessageManagerTest.testDeleteMessageWithMessage]Successful");
    }

    /**
     * Test the deleteMessage methods passing a The Message key to reference the 
     * message that must be deleted 
     * @throws java.lang.Exception
     */
    public void testDeleteMessageWithLUID() throws Exception {
        Log.info("[MessageManagerTest.testDeleteMessageWithLUID]Started");

        Message m1 = createUnreadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(1, 1, store.getFolder(INBOX_PATH)));

        String luid = mm.makeLuid(m1);
        mm.deleteMessage(luid);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(2, 0, store.getFolder(INBOX_PATH)));
        
        boolean isDeletedMessageAdded = !FlagQueue.getDeletedItemsList().containsKey(m1.getKey());

        Message m = null;

        try {
            m = mm.getMessage(m1.getKey());
            fail("[MessageManagerTest.testDeleteMessageWithLUID]Deleted message was found on the store");
        } catch (MailException me) {
        }

        assertTrue(isDeletedMessageAdded && m == null);

        Log.info("[MessageManagerTest.testDeleteMessageWithLUID]Successful");
    }

    
    /**
     * Test the deleteMessage methods passing an invalid Message key to 
     * reference the message that must be deleted. An exception is excpeted to 
     * be thrown
     * @throws java.lang.Exception
     */
    public void testDeleteMessageInvalidLUID() throws Exception {
        Log.info("[MessageManagerTest.testDeleteMessageInvalidLUID]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));
        String luid = m1.getKey().substring(3);

        boolean messageNotFound = false;
        try {
            mm.deleteMessage(luid);
            fail("[MessageManagerTest.testDeleteMessageInvalidLUID]Deleted a message with invalid LUID");
        } catch (MailException me) {
            messageNotFound = true;
        }

        assertTrue(messageNotFound);

        Log.info("[MessageManagerTest.testDeleteMessageInvalidLUID]Successful");
    }

    /**
     * Test the deleteMessage methods passing a message never stored to 
     * reference the message that must be deleted. An exception is excpeted to 
     * be thrown
     * @throws java.lang.Exception
     */
    public void testDeleteMessageUnexistentMessage() throws Exception {
        Log.info("[MessageManagerTest.testDeleteMessageUnexistentMessage]Started");

        Message notStoredMessage = createReadMessage();
        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(INBOX_PATH));

        boolean messageNotFound = false;

        try {
            mm.deleteMessage(notStoredMessage);
            fail("[MessageManagerTest.testDeleteMessageUnexistentMessage]Deleted a not existent message");
        } catch (MailException me) {
            messageNotFound = true;
        }

        assertTrue(messageNotFound);

        Log.info("[MessageManagerTest.testDeleteMessageUnexistentMessage]Successful");
    }

    /**
     * Test the messageEffectivelySent method and check that a message that was 
     * in "sending" state is deleted from the oubox and moved to the sent folder
     * A chek on the proper message flags is made to endure that the TX_SENT and
     * the TX_SENDING flag are properly updated 
     * @throws java.lang.Exception
     */
    public void testMessageEffectivelySent() throws Exception {
        Log.info("[MessageManagerTest.testMessageEffectivelySent]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, store.getFolder(OUTBOX_PATH));

        m1.getFlags().setFlag(MessageFlags.TX_SENDING, true);

        mm.messageEffectivelySent(m1.getKey());

        Folder sent = mm.getFolder(SENT_PATH);
        Folder outbox = mm.getFolder(OUTBOX_PATH);

        Enumeration children = sent.getChildren();
        Message sentMessage = (Message)children.nextElement();

        boolean isSentFlagUpdated = sentMessage.getFlags().isSet(MessageFlags.TX_SENT);
        boolean isSendingFlagUpdated = !sentMessage.getFlags().isSet(MessageFlags.TX_SENDING);
        boolean isSentFolderUpdated = sent.getMessageCount() == 1;
        boolean isOuboxFolderUpdated = outbox.getMessageCount() == 0;

        assertTrue(isSentFlagUpdated &&
                isSendingFlagUpdated &&
                isSentFolderUpdated &&
                isOuboxFolderUpdated);

        Log.info("[MessageManagerTest.testMessageEffectivelySent]Successful");
    }

    /**
     * Test the getMessagesInFolder method on the default folder. 1 message is 
     * added to the inbox, 2 to the outbox, 3 to the drafts and 4 to the sent.
     * The method call is tested by retrieving the folder using its string path 
     * @throws java.lang.Exception
     */
    public void testGetMessagesInFolderWithStringPath() throws Exception {
        Log.info("[MessageManagerTest.testGetMessagesInFolderWithStringPath]Started");

        Message i1 = createReadMessage();
        mm.appendMessageToFolder(i1, store.getFolder(INBOX_PATH));

        Message o1 = createReadMessage();
        mm.appendMessageToFolder(o1, store.getFolder(OUTBOX_PATH));
        Message o2 = createReadMessage();
        mm.appendMessageToFolder(o2, store.getFolder(OUTBOX_PATH));

        Message d1 = createReadMessage();
        mm.appendMessageToFolder(d1, store.getFolder(DRAFTS_PATH));
        Message d2 = createReadMessage();
        mm.appendMessageToFolder(d2, store.getFolder(DRAFTS_PATH));
        Message d3 = createReadMessage();
        mm.appendMessageToFolder(d3, store.getFolder(DRAFTS_PATH));

        Message s1 = createReadMessage();
        mm.appendMessageToFolder(s1, store.getFolder(SENT_PATH));
        Message s2 = createReadMessage();
        mm.appendMessageToFolder(s2, store.getFolder(SENT_PATH));
        Message s3 = createReadMessage();
        mm.appendMessageToFolder(s3, store.getFolder(SENT_PATH));
        Message s4 = createReadMessage();
        mm.appendMessageToFolder(s4, store.getFolder(SENT_PATH));

        Vector inboxMessages = mm.getMessagesInFolder(INBOX_PATH);
        Vector outboxMessages = mm.getMessagesInFolder(OUTBOX_PATH);
        Vector draftMessages = mm.getMessagesInFolder(DRAFTS_PATH);
        Vector sentMessages = mm.getMessagesInFolder(SENT_PATH);

        assertTrue(inboxMessages.size()  == 1 &&
                   outboxMessages.size() == 2 &&
                   draftMessages.size()  == 3 &&
                   sentMessages.size()   == 4);

        Log.info("[MessageManagerTest.testGetMessagesInFolderWithStringPath]Successful");
    }

    /**
     * Test the getMessagesInFolder method on the default folder. 1 message is 
     * added to the inbox, 2 to the outbox, 3 to the drafts and 4 to the sent.
     * The method call is tested by retrieving the folder using the related 
     * folder object
     * @throws java.lang.Exception
     */
    public void testGetMessagesInFolderWithFolder() throws Exception {
        Log.info("[MessageManagerTest.testGetMessagesInFolderWithFolder]Started");

        Message i1 = createReadMessage();
        mm.appendMessageToFolder(i1, store.getFolder(INBOX_PATH));

        Message o1 = createReadMessage();
        mm.appendMessageToFolder(o1, store.getFolder(OUTBOX_PATH));
        Message o2 = createReadMessage();
        mm.appendMessageToFolder(o2, store.getFolder(OUTBOX_PATH));

        Message d1 = createReadMessage();
        mm.appendMessageToFolder(d1, store.getFolder(DRAFTS_PATH));
        Message d2 = createReadMessage();
        mm.appendMessageToFolder(d2, store.getFolder(DRAFTS_PATH));
        Message d3 = createReadMessage();
        mm.appendMessageToFolder(d3, store.getFolder(DRAFTS_PATH));

        Message s1 = createReadMessage();
        mm.appendMessageToFolder(s1, store.getFolder(SENT_PATH));
        Message s2 = createReadMessage();
        mm.appendMessageToFolder(s2, store.getFolder(SENT_PATH));
        Message s3 = createReadMessage();
        mm.appendMessageToFolder(s3, store.getFolder(SENT_PATH));
        Message s4 = createReadMessage();
        mm.appendMessageToFolder(s4, store.getFolder(SENT_PATH));

        Vector inboxMessages = mm.getMessagesInFolder(mm.getFolder(INBOX_PATH));
        Vector outboxMessages = mm.getMessagesInFolder(mm.getFolder(OUTBOX_PATH));
        Vector draftMessages = mm.getMessagesInFolder(mm.getFolder(DRAFTS_PATH));
        Vector sentMessages = mm.getMessagesInFolder(mm.getFolder(SENT_PATH));

        assertTrue(inboxMessages.size()  == 1 &&
                   outboxMessages.size() == 2 &&
                   draftMessages.size()  == 3 &&
                   sentMessages.size()   == 4);

        Log.info("[MessageManagerTest.testGetMessagesInFolderWithFolder]Successful");
    }

    /**
     * Test the MessagesLoaderListener when calling the getMessagesInFolder
     * method.
     *
     * @throws java.lang.Exception
     */
    public void testMessagesLoaderListener() throws Exception {
        Log.info("[MessageManagerTest.testGetMessagesInFolderWithFolder]Started");

        Message unreadm = createUnreadMessage();
        Message readm   = createReadMessage();

        Folder inbox = store.getFolder(INBOX_PATH);

        mm.appendMessageToFolder(unreadm, inbox);
        mm.appendMessageToFolder(unreadm, inbox);
        mm.appendMessageToFolder(readm,   inbox);
        mm.appendMessageToFolder(unreadm, inbox);
        mm.appendMessageToFolder(unreadm, inbox);
        mm.appendMessageToFolder(readm,   inbox);
        mm.appendMessageToFolder(readm,   inbox);
        mm.appendMessageToFolder(readm,   inbox);
        mm.appendMessageToFolder(unreadm, inbox);
        mm.appendMessageToFolder(unreadm, inbox);

        MessagesLoaderListener listener = new MessagesLoaderListener() {
            public void messageLoaded(int count, int total) {
                getMessagesNotificationsCount++;
                getMessagesNotifications.put(new Integer(count), new Integer(total));
            }
        };

        Vector inboxMessages = mm.getMessagesInFolder(inbox, listener);

        assertTrue(inboxMessages.size()            == 10);
        assertTrue(getMessagesNotificationsCount   == 10);
        assertTrue(getMessagesNotifications.size() == 10);
        for(int count=1; count<=getMessagesNotifications.size(); count++) {
            assertTrue(((Integer)getMessagesNotifications.get(
                    new Integer(count))).intValue() == 10);
        }

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(6, 6, mm.getFolder(INBOX_PATH)));

        Log.info("[MessageManagerTest.testGetMessagesInFolderWithFolder]Successful");
    }

    /**
     * Test the getMessage method to retrieve a stored message using its String 
     * key
     * @throws java.lang.Exception
     */
    public void testGetMessageWithKey() throws Exception {
        Log.info("[MessageManagerTest.testGetMessageWithKey]Started");

        Message i1 = createReadMessage();
        mm.appendMessageToFolder(i1, store.getFolder(INBOX_PATH));

        boolean messageFound = false;

        try {
            mm.getMessage(i1.getKey());
            messageFound = true;
        } catch (MailException me) {
            fail("[MessageManagerTest.testGetMessageWithKey]Message " + i1.getKey() + " not Found");
        }

        assertTrue(messageFound);

        Log.info("[MessageManagerTest.testGetMessageWithKey]Successful");
    }

    /**
     * Test the getMessage method to retrieve a stored message using its String 
     * key and passing also its related folder
     * @throws java.lang.Exception
     */
    public void testGetMessageWithFolderAndKey() throws Exception {
        Log.info("[MessageManagerTest.testGetMessageWithFolderAndKey]Started");

        Message i1 = createReadMessage();
        mm.appendMessageToFolder(i1, store.getFolder(INBOX_PATH));

        boolean messageFound = false;

        try {
            mm.getMessage(mm.getFolder(INBOX_PATH), i1.getKey());
            messageFound = true;
        } catch (MailException me) {
            fail("[MessageManagerTest.testGetMessageWithKey]Message " + i1.getKey() + " not Found");
        }

        assertTrue(messageFound);

        Log.info("[MessageManagerTest.testGetMessageWithFolderAndKey]Successful");
    }

    /**
     * Test the createFolder metod passing a byte[] that represents a folder in 
     * SyncML format. The folder must be added to the store
     * @throws java.lang.Exception
     */
    public void testCreateFolder() throws Exception {
        Log.info("[MessageManagerTest.testCreateFolder]Started");

        String folderData = getSampleFolderData(TEST_FOLDER_NAME,
                FolderData.INBOX_ROLE);

        Folder f = mm.createFolder(
                ACCOUNT_PATH,
                folderData.getBytes());

        Folder result = mm.getFolder(ACCOUNT_PATH + TEST_FOLDER_PATH);

        assertTrue(result != null);

        Log.info("[MessageManagerTest.testCreateFolder]Successful");
    }

    /**
     * Test the updateFolder method passing a SyncML object in byte[] format. 
     * The test is based on changing the role "inbox" of the defaut inbox folder 
     * to "outbox".
     * @throws java.lang.Exception
     */
    public void testUpdateFolder() throws Exception {
        Log.info("[MessageManagerTest.testUpdateFolder]Started");

        String folderData = getSampleFolderData(INBOX_NAME,
                FolderData.OUTBOX_ROLE);

        Folder f = mm.updateFolder(
                ACCOUNT_PATH,
                folderData.getBytes(), INBOX_PATH);

        Folder result = mm.getFolder(INBOX_PATH);

        assertTrue(result.getRole().equals(FolderData.OUTBOX_ROLE));

        Log.info("[MessageManagerTest.testUpdateFolder]Successful");
    }
    
    /**
     * Test the updateFolder method passing a Folder object directly 
     * The test is based on changing the role "inbox" of the defaut inbox folder 
     * to "outbox".
     * @throws java.lang.Exception
     */
    public void testUpdateFolder_Folder() throws Exception {
        Log.info("[MessageManagerTest.testUpdateFolder_Folder]Started");

        Folder newFolder = mm.getFolder(INBOX_PATH);
        
        newFolder.setRole(FolderData.OUTBOX_ROLE);
        mm.updateFolder(newFolder);
        Folder result = mm.getFolder(INBOX_PATH);
        assertTrue(result.getRole().equals(FolderData.OUTBOX_ROLE));
        
        newFolder.setRole(FolderData.INBOX_ROLE);
        mm.updateFolder(newFolder);
        result = mm.getFolder(INBOX_PATH);
        assertTrue(result.getRole().equals(FolderData.INBOX_ROLE));

        Log.info("[MessageManagerTest.testUpdateFolder_Folder]Successful");
    }

    /**
     * Test the updateFolder method passing a AccountFolder object directly
     * The test is based on changing the visible name and email address.
     * 
     * @throws java.lang.Exception
     */
    public void testUpdateFolder_AccountFolder() throws Exception {
        Log.info("[MessageManagerTest.testUpdateFolder_AccountFolder]Started");

        AccountFolder newAccount = mm.getAccounts()[0];
        newAccount.setVisibleName("New Visible Name");
        newAccount.setEmailAddress("New.Email@Address.com");

        AccountFolder oldAccount = mm.getAccounts()[0];
        mm.updateFolder(newAccount);
        
        AccountFolder resultAccount = mm.getAccounts()[0];
        assertTrue(resultAccount.getVisibleName().equals("New Visible Name"));
        assertTrue(resultAccount.getEmailAddress().equals("New.Email@Address.com"));

        newAccount.setVisibleName(oldAccount.getVisibleName());
        newAccount.setEmailAddress(oldAccount.getEmailAddress());
        mm.updateFolder(newAccount);

        resultAccount = mm.getAccounts()[0];
        assertTrue(resultAccount.getVisibleName().equals(oldAccount.getVisibleName()));
        assertTrue(resultAccount.getEmailAddress().equals(oldAccount.getEmailAddress()));

        Log.info("[MessageManagerTest.testUpdateFolder_AccountFolder]Successful");
    }

    /**
     * Test the deleteFolder method passing the folder path to the tested method
     * @throws java.lang.Exception
     */
    public void testDeleteFolder() throws Exception {
        Log.info("[MessageManagerTest.testDeleteFolder]Started");

        Folder inbox = mm.getFolder(INBOX_PATH);
        mm.deleteFolder(INBOX_PATH);

        boolean folderDeleted = false;
        try {
            mm.getUnreadMessagesCount(inbox);
            fail("[MessageManagerTest.testDeleteFolder]got unread msg count of " +
                    "a deleted folder " + INBOX_PATH);
        } catch(MailException ex) {
            folderDeleted = true;
        }
        assertTrue(folderDeleted);

        folderDeleted = false;
        try {
            inbox = mm.getFolder(INBOX_PATH);
            fail("[MessageManagerTest.testDeleteFolder]got a deleted folder " + INBOX_PATH);
        } catch (MailException me) {
            folderDeleted = true;
        }
        assertTrue(folderDeleted);

        Log.info("[MessageManagerTest.testDeleteFolder]Successful");
    }

    /**
     * Test the getAccounts method. Before the method is invoked a new test
     * account is added. The result of the call must return an AccountFolder[]
     * with two accounts.
     * @throws java.lang.Exception
     */
    public void testGetAccounts() throws Exception {
        Log.info("[MessageManagerTest.testGetAccounts]Started");

        initTestAccount();

        AccountFolder[] af = mm.getAccounts();

        assertTrue(af.length == 2 &&
                af[0].getFullName().equals(TEST_ACCOUNT_PATH) &&
                af[1].getFullName().equals(ACCOUNT_PATH));

        Log.info("[MessageManagerTest.testGetAccounts]Successful");
    }

    /**
     * Test the getAccountFolders method passing the AccountFolder object as 
     * parameter. The test is performed on the default account
     * @throws java.lang.Exception
     */
    public void testGetAccountFoldersWithAccountFolder() throws Exception {
        Log.info("[MessageManagerTest.testGetAccountFoldersWithAccountFolder]Started");

        AccountFolder account = new AccountFolder(ACCOUNT_PATH,
                new Date(),
                store,
                ACCOUNT_VISIBLE_NAME,
                ACCOUNT_EMAIL_ADDRESS);

        Folder[] folders = mm.getAccountFolders(account);

        assertTrue(folders[0].getFullName().equals(DRAFTS_PATH) &&
                folders[1].getFullName().equals(SENT_PATH) &&
                folders[2].getFullName().equals(OUTBOX_PATH) &&
                folders[3].getFullName().equals(INBOX_PATH));

        Log.info("[MessageManagerTest.testGetAccountFoldersWithAccountFolder]Successful");
    }

    /**
     * Test the getAccountFolders method passing the Account String name as 
     * parameter. The test is performed on the default account
     * @throws java.lang.Exception
     */
    public void testGetAccountFoldersWithAccountStringName() throws Exception {
        Log.info("[MessageManagerTest.testGetAccountFoldersWithAccountStringName]Started");

        initTestAccount();

        Folder[] defaultFolders = mm.getAccountFolders(ACCOUNT_PATH);
        Folder[] testFolders = mm.getAccountFolders(TEST_ACCOUNT_PATH);

        assertTrue(defaultFolders[0].getFullName().equals(DRAFTS_PATH) &&
                defaultFolders[1].getFullName().equals(SENT_PATH) &&
                defaultFolders[2].getFullName().equals(OUTBOX_PATH) &&
                defaultFolders[3].getFullName().equals(INBOX_PATH) &&
                testFolders[0].getFullName().equals(TEST_DRAFTS_PATH) &&
                testFolders[1].getFullName().equals(TEST_SENT_PATH) &&
                testFolders[2].getFullName().equals(TEST_OUTBOX_PATH) &&
                testFolders[3].getFullName().equals(TEST_INBOX_PATH));

        Log.info("[MessageManagerTest.testGetAccountFoldersWithAccountStringName]Successful");
    }

    /**
     * Test the getAccountfolderFromRole on the default and an additional test 
     * account. Check that the folders are retrieved by their role
     * @throws java.lang.Exception
     */
    public void testGetAccountFolderFromRole() throws Exception {
        Log.info("[MessageManagerTest.testGetAccountFolderFromRole]Started");

        initTestAccount();

        AccountFolder defaultAccount = new AccountFolder(ACCOUNT_PATH,
                new Date(),
                store,
                ACCOUNT_VISIBLE_NAME,
                ACCOUNT_EMAIL_ADDRESS);

        Folder af = mm.getAccountFolderFromRole(defaultAccount, FolderData.INBOX_ROLE);

        assertTrue(af.getFullName().equals(INBOX_PATH));

        Log.info("[MessageManagerTest.testGetAccountFolderFromRole]Successful");
    }

    /**
     * Test the makeLUID utility method that return the Message LUID given the 
     * message 
     * @throws java.lang.Exception
     */
    public void testMakeLUID() throws Exception {
        Log.info("[MessageManagerTest.testMakeLUID]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        String expected = m1.getParent().getFullName() + "/" + m1.getRecordId();

        assertEquals(expected, mm.makeLuid(m1));
        Log.info("[MessageManagerTest.testMakeLUID]Successful");
    }

    /**
     * Test the msgIdFromKey utility method that returns the MessageId given the 
     * message key 
     * @throws java.lang.Exception
     */
    public void testMsgIdFromKey() throws Exception {
        Log.info("[MessageManagerTest.testMsgIdFromKey]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        int index = m1.getKey().lastIndexOf(Folder.FOLDER_SEPARATOR);
        String expected = m1.getKey().substring(index + 1);

        assertEquals(expected, mm.msgIdFromKey(m1.getKey()));

        Log.info("[MessageManagerTest.testMsgIdFromKey]Successful");
    }

    /**
     * Test the folderFromKey utility method that returns the Message folder 
     * given the message key 
     * @throws java.lang.Exception
     */
    public void testFolderFromKey() throws Exception {
        Log.info("[MessageManagerTest.testFolderFromKey]Started");

        Message m1 = createReadMessage();
        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        int pos = m1.getKey().lastIndexOf(Folder.FOLDER_SEPARATOR);

        String expected = m1.getKey().substring(0, pos);

        assertEquals(expected, mm.folderFromKey(m1.getKey()));

        Log.info("[MessageManagerTest.testFolderFromKey]Successful");
    }

    /**
     * Test the accountFromKey utility method that returns the Message account
     * given the message key 
     * @throws java.lang.Exception
     */
    public void testAccountFromKey() throws Exception {
        Log.info("[MessageManagerTest.testAccountFromKey]Started");

        Message m1 = createReadMessage();

        mm.appendMessageToFolder(m1, mm.getFolder(INBOX_PATH));

        int pos = m1.getKey().indexOf(Folder.FOLDER_SEPARATOR, 1);
        String expected = m1.getKey().substring(0, pos);

        assertEquals(expected, mm.accountFromKey(m1.getKey()));

        Log.info("[MessageManagerTest.testAccountFromKey]Successful");
    }

    /**
     * Test the createFolder method passing and Account SyncML data object. The 
     * account must be added to the store after it has been parsed. 
     * @throws java.lang.Exception
     */
    public void testCreateAccount() throws Exception {
        Log.info("[MessageManagerTest.testCreateAccount]Started");

        String accountData = getSampleAccountFolderData(TEST_ACCOUNT_NAME,
                TEST_ACCOUNT_VISIBLE_NAME, TEST_ACCOUNT_EMAIL_ADDRESS);

        Folder f = mm.createFolder(
                Folder.ROOT_FOLDER_PATH,
                accountData.getBytes());

        Folder result = mm.getFolder(TEST_ACCOUNT_PATH);

        assertTrue(result != null);

        Log.info("[MessageManagerTest.testCreateAccount]Successful");
    }

    /**
     * Test the updateFolder method passing and Account SyncML data object to 
     * update the default account. The default account after the update must 
     * have the email address and visible name changed into the new ones, passed
     * into the SyncML byte[] data.
     */
    public void testUpdateAccount() throws Exception {
        Log.info("[MessageManagerTest.testUpdateAccount]Started");

        String accountData = getSampleAccountFolderData(ACCOUNT_NAME,
                TEST_ACCOUNT_VISIBLE_NAME, TEST_ACCOUNT_EMAIL_ADDRESS);

        Folder f = mm.updateFolder(
                Folder.ROOT_FOLDER_PATH,
                accountData.getBytes(),
                ACCOUNT_PATH);

        AccountFolder[] af = mm.getAccounts();

        assertTrue(af[0].getFullName().equals(ACCOUNT_PATH) &&
                af[0].getEmailAddress().equals(TEST_ACCOUNT_EMAIL_ADDRESS) &&
                af[0].getVisibleName().equals(TEST_ACCOUNT_VISIBLE_NAME));

        Log.info("[MessageManagerTest.testUpdateAccount]Successful");
    }

    /**
     * Test the getUnreadMessagesCount method with a Folder as parameter
     */
    public void testGetUnreadMessagesCountFromFolder() throws Exception {

        Log.info("[MessageManagerTest.testGetUnreadMessagesCountFromFolder]Started");
        
        Folder inbox = mm.getFolder(INBOX_PATH);

        Message readm1 =   createReadMessage();
        Message readm2 =   createReadMessage();
        Message unreadm1 = createUnreadMessage();
        Message unreadm2 = createUnreadMessage();
        Message unreadm3 = createUnreadMessage();
        Message unreadm4 = createUnreadMessage();

        mm.appendMessageToFolder(unreadm1, inbox);
        mm.appendMessageToFolder(readm1,   inbox);
        mm.appendMessageToFolder(unreadm2, inbox);
        mm.appendMessageToFolder(unreadm3, inbox);
        mm.appendMessageToFolder(unreadm4, inbox);
        mm.appendMessageToFolder(readm2,   inbox);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(4, 4, inbox));

        assertTrue(mm.getUnreadMessagesCount(inbox) == 4);

        unreadm1.getFlags().setFlag(MessageFlags.OPENED, true);
        mm.updateMessage(unreadm1);
        assertTrue(mm.getUnreadMessagesCount(inbox) == 3);

        mm.deleteMessage(unreadm1);
        assertTrue(mm.getUnreadMessagesCount(inbox) == 3);

        mm.deleteMessage(unreadm2);
        assertTrue(mm.getUnreadMessagesCount(inbox) == 2);

        mm.deleteMessage(unreadm4.getKey());
        assertTrue(mm.getUnreadMessagesCount(inbox) == 1);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(7, 1, inbox));

        Log.info("[MessageManagerTest.testGetUnreadMessagesCountFromFolder]Successful");
    }

    /**
     * Test the getUnreadMessagesCount method with a AccountFolder as parameter
     */
    public void testGetUnreadMessagesCountFromAccountFolder() throws Exception {

        Log.info("[MessageManagerTest.testGetUnreadMessagesCountFromAccountFolder]Started");

        Folder inbox = mm.getFolder(INBOX_PATH);
        AccountFolder account = (AccountFolder)inbox.getParent();

        Message readm1 =   createReadMessage();
        Message readm2 =   createReadMessage();
        Message unreadm1 = createUnreadMessage();
        Message unreadm2 = createUnreadMessage();
        Message unreadm3 = createUnreadMessage();
        Message unreadm4 = createUnreadMessage();

        mm.appendMessageToFolder(unreadm1, inbox);
        mm.appendMessageToFolder(readm1,   inbox);
        mm.appendMessageToFolder(unreadm2, inbox);
        mm.appendMessageToFolder(unreadm3, inbox);
        mm.appendMessageToFolder(unreadm4, inbox);
        mm.appendMessageToFolder(readm2,   inbox);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(4, 4, inbox));

        assertTrue(mm.getUnreadMessagesCount(account) == 4);

        unreadm1.getFlags().setFlag(MessageFlags.OPENED, true);
        mm.updateMessage(unreadm1);
        assertTrue(mm.getUnreadMessagesCount(account) == 3);

        mm.deleteMessage(unreadm1);
        assertTrue(mm.getUnreadMessagesCount(account) == 3);

        mm.deleteMessage(unreadm2);
        assertTrue(mm.getUnreadMessagesCount(account) == 2);

        mm.deleteMessage(unreadm4.getKey());
        assertTrue(mm.getUnreadMessagesCount(account) == 1);

        // Test the unread messages count notification
        assertTrue(checkUnreadMessageNotification(7, 1, inbox));

        Log.info("[MessageManagerTest.testGetUnreadMessagesCountFromAccountFolder]Successful");
    }

    private void initDefaultAccount() {
        // Add the default ACCOUNT
        mm.addFolder(new AccountFolder(ACCOUNT_PATH, new Date(), store, ACCOUNT_VISIBLE_NAME, ACCOUNT_EMAIL_ADDRESS));

        // Add the default account FOLDERS
        mm.addFolder(new Folder(INBOX_PATH, FolderData.INBOX_ROLE, new Date(), store));
        mm.addFolder(new Folder(OUTBOX_PATH, FolderData.OUTBOX_ROLE, new Date(), store));
        mm.addFolder(new Folder(SENT_PATH, FolderData.SENT_ROLE, new Date(), store));
        mm.addFolder(new Folder(DRAFTS_PATH, FolderData.DRAFTS_ROLE, new Date(), store));
    }

    private void initTestAccount() {
        // Add the default ACCOUNT
        mm.addFolder(new AccountFolder(TEST_ACCOUNT_PATH, new Date(), store, TEST_ACCOUNT_VISIBLE_NAME, TEST_ACCOUNT_EMAIL_ADDRESS));

        // Add the default account FOLDERS
        mm.addFolder(new Folder(TEST_INBOX_PATH, FolderData.INBOX_ROLE, new Date(), store));
        mm.addFolder(new Folder(TEST_OUTBOX_PATH, FolderData.OUTBOX_ROLE, new Date(), store));
        mm.addFolder(new Folder(TEST_SENT_PATH, FolderData.SENT_ROLE, new Date(), store));
        mm.addFolder(new Folder(TEST_DRAFTS_PATH, FolderData.DRAFTS_ROLE, new Date(), store));
    }

    private void update(Message m, int flag) {
        MessageFlags mf = m.getFlags();
        mf.setFlag(flag, true);
        m.setFlags(mf);
        mm.updateMessageFlags(m, m.getFlags());
    }

    private boolean checkUpdatedMessageFlagsByInt(Message m, int flag, boolean expected) {
        boolean isUpdated = m.getFlags().isSet(flag);

        boolean isUpdatesavedOnFlagQueue = FlagQueue.get(m.getKey()).isSet(flag);

        boolean isUpdateSavedOnStorage = mm.getMessage(m.getKey()).getFlags().isSet(flag);


        boolean updated = isUpdated;

        if (flag == MessageFlags.OPENED || flag == MessageFlags.FLAGGED) {
            updated = isUpdated == isUpdateSavedOnStorage;
        }

        if (isUpdatesavedOnFlagQueue) {
            return updated == expected;
        }

        return false;
    }

    private boolean checkUnreadMessageNotification(int notificationsCount,
            int lastCount, Folder lastFolder) {
        boolean ret = (unreadMessageListener.notificationsCount == notificationsCount) &&
                      (unreadMessageListener.lastCount == lastCount);
        if(lastFolder == null) {
            return ret && (unreadMessageListener.lastFolder == null);
        } else {
            return ret && (unreadMessageListener.lastFolder.getFullName().equals(
                lastFolder.getFullName()));
        }
    }

    private Message createReadMessage() {
        Message m = new Message();
        Address from = new Address(Address.FROM, ADDRESS_FROM);
        Address[] to = new Address[1];
        to[0] = new Address(Address.TO, ADDRESS_TO);
        Address[] cc = new Address[1];
        cc[0] = new Address(Address.CC, ADDRESS_CC);
        Address[] bcc = new Address[1];
        bcc[0] = new Address(Address.BCC, ADDRESS_BCC);
        Address[] replyTo = new Address[1];
        replyTo[0] = new Address(Address.REPLYTO, ADDRESS_REPLY_TO);

        m.getFlags().setFlag(MessageFlags.OPENED, true);
        
        m.setFrom(from);
        m.setTo(to);
        m.setCc(cc);
        m.setBcc(bcc);
        m.setReplyTo(replyTo);

        String content = MESSAGE_CONTENT;
        m.setContent(content);

        return m;
    }

    private Message createUnreadMessage() {
        Message m = createReadMessage();
        m.getFlags().setFlag(MessageFlags.OPENED, false);
        return m;
    }

    public void initDefaultConfiguration() {
        MailClientConfig conf = new MailClientConfig();
        conf.enableDeletePropagation(true);
        ConfigManager.setMailClientConfig(conf);
    }

    private String getSampleFolderData(String name, String role) {
        return "<Folder>" +
                "<name>" + name + "</name>" +
                "<created>20090428T162654Z</created>" +
                "<role>" + role + "</role>" +
                "</Folder>";
    }

    private String getSampleEmailData(boolean read) {
        return    "<Email>" +
                  "<read>" + read + "</read>" +
                  "<forwarded>false</forwarded>" +
                  "<replied>false</replied>" +
                  "<deleted>false</deleted>" +
                  "<flagged>false</flagged>" +
                  "<emailitem>" +
                  "<![CDATA[ " +
                  "\r\nReceived: by 10.210.18.13 with SMTP id 13cs322923ebr; Fri, 2 Oct 2009 09:46:03 -0700 (PDT)" +
                  "\r\nDate: Fri, 2 Oct 2009 09:37:40 -0700 (PDT)" +
                  "\r\nFrom: Name <email@address1.com>" +
                  "\r\nTo: \"email@address2.com\"" +
                  "\r\nMessage-ID: <17339599.21701254501460790.JavaMail.root@a.mail.sonomait.com>" +
                  "\r\nSubject: Subject" +
                  "\r\nMIME-Version: 1.0" +
                  "\r\nContent-Type: text/plain; charset=UTF-8" +
                  "\r\nContent-Transfer-Encoding: 8bit" +
                  "\r\n\r\n Generic Content \r\n\r\n" +
                  "]]>" +
                  "</emailitem>" +
                  "</Email>";
    }

    private String getSampleAccountFolderData(String name, String vName, String email) {
        return "<Folder>" +
                "<name>" + name + "</name>" +
                "<created>20090428T162654Z</created>" +
                "<role>account</role>" +
                "<Ext>" +
                "<XNam>VisibleName</XNam>" +
                "<XVal>" + vName + "</XVal>" +
                "</Ext>" +
                "<Ext>" +
                "<XNam>EmailAddress</XNam>" +
                "<XVal>" + email + "</XVal>" +
                "</Ext>" +
                "</Folder>";
    }
}

