/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.syncml;

import com.funambol.util.ConsoleAppender;
import com.funambol.util.Log;
import com.funambol.util.MailDateFormatter;
import com.funambol.util.XmlException;
import com.funambol.util.ChunkedString;

import java.util.Hashtable;
import java.util.Vector;

import junit.framework.*;

public class FolderDataTest extends TestCase {

    private final String XML_FOLDER_1 = "<Folder>" +
                                        "<name>Gmail Home</name>" +
                                        "<created>20090428T162654Z</created>" +
                                        "<role>account</role>" +
                                        "<Ext>" +
                                        "<XNam>VisibleName</XNam>" +
                                        "<XVal>Name Surname</XVal>" +
                                        "</Ext>" +
                                        "<Ext>" +
                                        "<XNam>EmailAddress</XNam>" +
                                        "<XVal>Name.Surname@gmail.com</XVal>" +
                                        "</Ext>" +
                                        "</Folder>";

    /**
     * Creates a new instance of this testcase
     */
    public FolderDataTest(String name) {
        super(name);
        Log.initLog(new ConsoleAppender());
        Log.setLogLevel(Log.DEBUG);
    }
    
    /**
     * TestCase setUp default method
     */
    public void setUp() { }
    
    /**
     * TestCase tearDown default method
     */
    public void tearDown() { }
    
    /**
     * Test a simple FolderData (without Ext tags)
     */
    public void testBasicFolderData() throws Exception {

        // Set up and fill a FolderData object
        FolderData fdata = new FolderData();
        fdata.setName("folder name");
        fdata.setRole(FolderData.OUTBOX_ROLE);
        fdata.setCreated(MailDateFormatter.parseUTCDate("20081208T093000Z"));

        // Retrieve each FolderData attribute
        assertTrue(fdata.getName().equals("folder name"));
        assertTrue(fdata.getRole().equals(FolderData.OUTBOX_ROLE));
        assertTrue(MailDateFormatter.dateToUTC(fdata.getCreated()).equals("20081208T093000Z"));
        assertTrue(fdata.getExt() == null);
    }
    
    /**
     * Test the FolderData accessor methods
     */
    public void testFolderData() throws Exception {

        // Set up and fill a FolderData object
        FolderData fdata = new FolderData();
        fdata.setName("folder name");
        fdata.setRole(FolderData.INBOX_ROLE);
        fdata.setCreated(MailDateFormatter.parseUTCDate("20091208T093000Z"));
        fdata.setExt(getTestExt());

        // Retrieve each FolderData attribute
        assertTrue(fdata.getName().equals("folder name"));
        assertTrue(fdata.getRole().equals(FolderData.INBOX_ROLE));
        assertTrue(MailDateFormatter.dateToUTC(fdata.getCreated()).equals("20091208T093000Z"));

        // Check folder exts
        Hashtable ext = fdata.getExt();
        Vector val = (Vector)ext.get("extName1");
        assertTrue(val.size() == 1);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("extValue1"));
        val = (Vector)ext.get("extName2");
        assertTrue(val.size() == 2);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("extValue1"));
        assertTrue(((ChunkedString)val.elementAt(1)).equals("extValue2"));
    }

    /**
     * Test the FolderData parser
     * @throws XmlException
     */
    public void testFolderDataParser() throws Exception {

        FolderData fdata = new FolderData();
        fdata.parse(XML_FOLDER_1.getBytes());

        assertTrue(fdata.getName().equals("Gmail Home"));
        assertTrue(fdata.getRole().equals("account"));
        assertTrue(MailDateFormatter.dateToUTC(fdata.getCreated()).equals("20090428T162654Z"));

        // Check folder exts
        Hashtable ext = fdata.getExt();
        Vector val = (Vector)ext.get("VisibleName");
        assertTrue(val.size() == 1);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("Name Surname"));
        val = (Vector)ext.get("EmailAddress");
        assertTrue(val.size() == 1);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("Name.Surname@gmail.com"));
    }

    /**
     * Test the FolderData Formatter:
     *  1) populate a FolderData object
     *  2) format the FolderData
     *  3) parse the FolderData
     *  4) compare the parsed FolderData and the original one
     * @throws XmlException
     */
    public void testFolderDataFormatter() throws Exception {

        FolderData fdata = new FolderData();
        fdata.setName("Funambol Home");
        fdata.setRole("account");
        fdata.setCreated(MailDateFormatter.parseUTCDate("20090428T162654Z"));
        fdata.setExt(getTestExt());

        String xmlFolder = new String(fdata.format());

        fdata = new FolderData();
        fdata.parse(xmlFolder.getBytes());

        assertTrue(fdata.getName().equals("Funambol Home"));
        assertTrue(fdata.getRole().equals("account"));
        assertTrue(MailDateFormatter.dateToUTC(fdata.getCreated()).equals("20090428T162654Z"));

        // Check folder exts
        Hashtable ext = fdata.getExt();
        Vector val = (Vector)ext.get("extName1");
        assertTrue(val.size() == 1);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("extValue1"));
        val = (Vector)ext.get("extName2");
        assertTrue(val.size() == 2);
        assertTrue(((ChunkedString)val.elementAt(0)).equals("extValue1"));
        assertTrue(((ChunkedString)val.elementAt(1)).equals("extValue2"));
    }

    /**
     * Get a Ext HashTable for test
     * @return
     */
    private Hashtable getTestExt() throws Exception {
        Hashtable ext = new Hashtable();

        Vector extValues1 = new Vector();
        extValues1.addElement(new ChunkedString("extValue1"));
        ext.put("extName1", extValues1);

        Vector extValues2 = new Vector();
        extValues2.addElement(new ChunkedString("extValue1"));
        extValues2.addElement(new ChunkedString("extValue2"));
        ext.put("extName2", extValues2);
        
        return ext;
    }
}
