/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient;

import com.funambol.util.Log;
import com.funambol.util.ConsoleAppender;

/**
 * General Testcase container representation for high level usage. Extend this
 * this class to create real test case.
 */
public abstract class TestCase {

    public static final int STATUS_IDLE = 0;

    public static final int RESULT_OK = 1;
    
    public static final int RESULT_KO = 2;
    
    private String testName = null;

    protected int result = STATUS_IDLE;

    //protected Robot robot = null;

    protected final int WAIT_FOR_START_SYNC_MAX_DELAY = 60000;
    protected final int WAIT_FOR_END_SYNC_MAX_DELAY   = 300000;
    protected final int SHOW_SCREEN_MAX_DELAY         = 20000;

    private int initialInboxMessages = 0;
    private int initialOutboxMessages = 0;
    private int initialDraftsMessages = 0;
    private int initialSentMessages = 0;
    private int initialAccounts = 0;

    private boolean populateContactStore = false;

    /**
     * Constructor
     * @param testName is the test name
     */
    public TestCase(String testName) {
        this.testName = testName;
        //robot = Robot.getInstance();
    }
    
    /**
     * Set up, execute, and tear down the test case
     */
    public void runTest() {

        setUp();
        try {
            Thread.sleep(2000);
        } catch (InterruptedException ex) {
            Log.error("[TestCase.runTest]Exception occurred: " + ex);
        }
        try {
            Log.info("[TestCase.runTest] " + getTestName() + " Started");
            test();
            result = RESULT_OK;
            Log.info("[TestCase.runTest] " + getTestName() + " Successful");
        } catch (ClientTestException cte) {
            result = RESULT_KO;
            Log.error("[TestCase.runTest] " + getTestName() + " Failed: " + cte + cte.getMessage());
            cte.printStackTrace();
        } catch (Throwable t) {
            result = RESULT_KO;
            Log.error("[TestCase.runTest] " + getTestName() + " Failed: " + t);
            t.printStackTrace();
        } finally {
            tearDown();
        }
    }

    /**
     * Accessor method
     * @return String the test name
     */
    public String getTestName() {
        return this.testName;
    }

    /**
     * Get the test result
     * @return int the test result
     */
    protected int getTestResult() {
        return this.result;
    }
    
    /**
     * set the test result for this specific test
     * @param result the test result to be set
     */
    protected void setTestResult(int result) {
        this.result = result;
    }
    
    /**
     * Inbox initial messages number
     * @return the requested messages number
     */
    public int getInitialInboxMessages() {
        return initialInboxMessages;
    }

    /**
     * Outbox initial messages number
     * @return the requested messages number
     */
    public int getInitialOutboxMessages() {
        return initialOutboxMessages;
    }

    /**
     * Drafts initial messages number
     * @return the requested messages number
     */
    public int getInitialDraftsMessages() {
        return initialDraftsMessages;
    }

    /**
     * Sent initial messages number
     * @return the requested messages number
     */
    public int getInitialSentMessages() {
        return initialSentMessages;
    }

    /**
     * Sent initial accounts number
     * @return the requested accounts number
     */
    public int getInitialAccounts() {
        return initialAccounts;
    }

    /**
     * Check if the Contact Store must be populated.
     * @return true if it must be populated
     */
    public boolean getPopulateContactStore() {
        return populateContactStore;
    }

    /**
     * Set Inbox initial messages number
     * @param initialInboxMessages the messages number
     */
    public void setInitialInboxMessages(int initialInboxMessages) {
        this.initialInboxMessages=initialInboxMessages;
    }

    /**
     * Set Outbox initial messages number
     * @param initialOutboxMessages the messages number
     */
    public void setInitialOutboxMessages(int initialOutboxMessages) {
        this.initialOutboxMessages=initialOutboxMessages;
    }

    /**
     * Set Drafts initial messages number
     * @param initialDraftsMessages the messages number
     */
    public void setInitialDraftsMessages(int initialDraftsMessages) {
        this.initialDraftsMessages=initialDraftsMessages;
    }

    /**
     * Set Sent initial messages number
     * @param initialSentMessages the messages number
     */
    public void setInitialSentMessages(int initialSentMessages) {
        this.initialSentMessages=initialSentMessages;
    }

    /**
     * Set initial accounts number
     * @param initialAccounts the accounts number
     */
    public void setInitialAccounts(int initialAccounts) {
        this.initialAccounts=initialAccounts;
    }

    /**
     * Set if the Contact Store must be populated.
     * @param populate if it must be populated
     */
    public void setPopulateContactStore(boolean populate) {
        this.populateContactStore=populate;
    }
    
    /**
     * Integer Equality Assertion
     * @param expected int expected value
     * @param result int given result
     * @throws com.funambol.mailclient.ClientTestException if the two integer 
     * are not equals
     */
    public void assertEquals(int expected, int result) throws ClientTestException {
        if (expected != result) {
            throw new ClientTestException("Expected: " + expected + " - Result: " + result);
        }
    }
    
    /**
     * Object Equality Assertion
     * @param expected int expected value
     * @param result int given result
     * @throws com.funambol.mailclient.ClientTestException if the two integer 
     * are not equals
     */
    public void assertEquals(Object expected, Object result) throws ClientTestException {
        if (!expected.equals(result)) {
            throw new ClientTestException("Expected: " + expected + " - Result: " + result);
        }
    }
    
    public void assertEquals(String expected, String result) throws ClientTestException {
        if (!expected.equals(result)) {
            throw new ClientTestException("Expected: " + expected + " - Result: " + result);
        }
    }
    
    public void assertTrue(boolean isAssertionTrue) throws ClientTestException {
        if (!isAssertionTrue) {
            throw new ClientTestException("Assert True Failed");
        }
    }
    
    /**
     * Force the test failure
     * @throws com.funambol.mailclient.ClientTestException
     */
    public void fail() throws ClientTestException {
        throw new ClientTestException("Failure force by test");
    }
    
    /**
     * Abstract method to setup the environment for a specific test
     */
    public abstract void setUp();

    /**
     * Abstract method to tear down the environment for a specific test
     */
    public abstract void tearDown();

    /**
     * Abstract method to execute a specific test. Represents the test body.
     */
    public abstract void test() throws Throwable;
}
