/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient;

/**
 * This component lists all the commands available in the automatic test.
 * scripting language.
 */
public interface UserCommands {

    /**
     * This command simulates a user clicking on a menu item. This is done regardless
     * of where the menu item is actually positioned (soft button or menu).
     * The command does not cause the menu to show up, but it just simulates
     * the action associated to the menu.
     *
     * @param item the string shown on the menu item 
     */
    public void ClickMenuItem(String item);

    /**
     * This command waits for a certain screen to be displayed and blocks the
     * script execution until such a screen is visible or the given timeout expires.
     *
     * @param name the screen name is a string that can take one of the values
     * listed here below.
     * <ul>
     *   <li> InboxMessageList is the inbox </li>
     *   <li> OutboxMessageList is the outbox </li>
     *   <li> DraftsMessageList is the draft </li>
     *   <li> SentMessageList is the the sent </li>
     *   <li> AlertScreen is any alert or prompt </li>
     *   <li> AccountSettingsForm is the account screen </li>
     *   <li> GeneralSettingsForm is the general settings screen </li>
     *   <li> ContactList is the contact list </li>
     *   <li> ComposeMessageScreen is the compose message screen </li>
     *   <li> ViewMessageScreen is the view message screen </li>
     *   <li> EditField$EditForm is any edit screen where the user can enter text </li>
     *   <li> FileExplorerScreen is the screen to browse for attachments </li>
     *   <li> AddContactForm is the screen to add a new contact </li>
     * </ul>
     * @param timeout the timeout is an integer value expressed in milliseconds.
     */
    public void WaitForScreen(String name, int timeout);

    /**
     * This command simulates a user entering a value into a field.
     * The value is entered one shot and not char by char.
     * The label name is the label associated to the field.
     * For example in the screen for adding a new recipient we have “First Name" and this value
     * must be used as label name.
     *
     * The application has several types of input fields. For example:
     * <ul>
     * <li> TextField like username and password fields</li>
     * <li>ChoiceGroup like the sync mode</li>
     * <li>TextBox like the body and the subject in the compose screen</li>
     * </ul>
     *
     * For TextBox and TextField the value is the string to be entered,
     * while for ChoiceGroup the value is the index in the list of possible values (starting from 0).
     *
     *
     * @param labelName the label name (must be identical to the one displayed)
     * @param value an arbitrary string to be set in the field. If the field
     *              accepts only one type of characters then the value shall
     *              contain only valid characters. If invalid characters are
     *              supplied the test behavior is unpredictable.
     *              For choice groups, the value represent the index (an integer
     *              starting from 0) of the selected item
     */
    public void SetInputField(String labelName, String value);

    /**
     * This command suspends the test execution for the given amount of time.
     * @param delay is an integer value expressed in milliseconds. If the delay is <= 0
     *              then the command wait forever and the script gets interrupted.
     *              This is useful if the script shall give the user the possibility to use the application.
     */
    public void Wait(int delay);

    /**
     * This instruction must be the first one at the beginning of a test.
     * For a single test script there is only one of such instruction, while a multi test
     * script may contain several instances.
     * @param name is the name of the test as it will be reported in the final summary.
     */
    public void BeginTest(String name);

    /**
     * Terminates the commands of a test.
     */
    public void EndTest();

    /**
     * This command clears the client configuration. After this command is executed,
     * the client is configuration-less.
     */
    public void ResetConfig();

    /**
     * This command starts the main application. When the test starts the main
     * application (Funambol Email Client) is not started until the script triggers this command.
     */
    public void StartMainApp();

    /**
     * This command clears the remote inbox associated to an account (a Funambol account).
     * For this operation to succeed it is necessary to have a sync account associated
     * to the mailbox to clean.
     * If there are network errors the operation may fail. (TODO Shall we stop
     * execution???)
     *
     * @param syncUrl is the syncrhonization URL (such as: http://my.funambol.com/sync)
     * @param userName is the sync user name
     * @param pwd is the sync password
     */
    public void ResetRemoteInbox(String syncUrl, String userName, String pwd);

    /**
     * This command simulates the user pressing a key (single press, not repeated).
     * @param keyName can take the following values:
     *              <ul>
     *                <li>ArrowDown</li>
     *                <li>ArrowUp</li>
     *                <li>ArrowLeft</li>
     *                <li>ArrowRight</li>
     *                <li>Fire</li>
     *                <li>KeyChar</li>
     *              </ul>
     * @param keyChar this parameter is optional and it is only required when
     * the keyName is KeyChar. In this case it represents the character to
     * simulate (e.g. "0" to simulate the 0 key, or "*" to simulate the star
     * key)
     */
    public void KeySinglePress(String keyName, String keyChar);

    /**
     * This command creates an email account on the device.
     * @param accountName the account name (e.g. test@funambol.com)
     * @param emailAddress the email address (e.g. test@funambol.com)
     * @param displayName the display name for this account (e.g. Test)
     */
    public void CreateAccount(String accountName, String emailAddress, String displayName);

    /**
     * This command creates a folder for a given account
     * @param accountName the account name (e.g. test@funambol.com). If the account
     *                    does not exist, then the test will fail.
     * @param folderPath the folder path (e.g. Inbox). If the provided path is not
     *                   at first level, such as inbox/work, then the parent folder
     *                   must exist otherwise the test fails
     * @param role the folder role. This can be: 
     *             <ul>
     *               <li>inbox</li>
     *               <li>outbox</li>
     *               <li>sent</li>
     *               <li>drafts</li>
     *             </ul>
     */
    public void CreateFolder(String accountName, String folderPath, String role);

    /**
     * This command creates a message in a given folder. It supports various 
     * fields of a message, but not all of them.  At the moment it is not possible
     * to specify attachments, recipients other than the TO, flags other than READ.
     * These features will be added in the future, as they are needed.
     *
     * @param accountName the account name (e.g. test@funambol.com). If the account does not exist, then the test will fail.
     * @param folderPath the folder path (e.g. Inbox). If the folder does not exist the test fails
     * @param from the from address (e.g. test@funambol.com)
     * @param subject the subject (any string)
     * @param body the body (any string)
     * @param read the read flag. This can be true to indicate that the message is read, false otherwise.
     */
    public void CreateMessage(String accountName, String folderPath, String from, String subject,
                              String body, String read);

    /**
     * Asserts on the number of items in a list. Various lists are present in the client.
     * They can be referred by a symbolic name. See section ?? for the valid names.
     * If the count does not match the expected value, the script is aborted and
     * an error is reported.
     *
     * @param listName is a symbolic name for the list.
     * @param count is an integer value (greater or equal to zero)
     */
    public void CheckListCount(String listName, int count);


    /**
     * Asserts that the active contact (currently selected) has a certain property value.
     * For example it is possible to check that the currently selected contact has a given first or last name.
     * @param propertyName is the name of a property. Valid values are:
     *                     <ul>
     *                       <li>FirstName</li>
     *                       <li>LastName</li>
     *                       <li>VisibleName</li>
     *                       <li>value is the value of the field</li>
     *                     </ul>
     * @param value the expected value of the given property
     */
    public void CheckActiveContactProperty(String propertyName, String value);


    /**
     * This command wait for a sync to start and finish within a maximum amount
     * of time. As soon as the sync terminates, the script continues its
     * execution
     *
     * @param minStart is the time the script waits for the sync to start. If
     *                 the sync does not start withing this time, the test fails
     * @param maxTime  is the maximum time the script is willing for the sync to
     *                 terminate. If it does not terminate within this limit,
     *                 the test fails.
     */
    public void WaitForSyncToComplete(String minStart, String maxTime);

    /**
     * This command includes another script file.
     *
     * @param name is the script name. If this is an URL, then the script is
     *             fetched from that location, otherwise the base url if the main script is
     *             added as prefix to this name
     */
    public void Include(String name);

    /**
     * Asserts that the active message (currently selected) has a certain property value.
     * For example it is possible to check that the currently selected message has a given sender.
     * @param name is the name of a property. Valid values are:
     *                     <ul>
     *                       <li>From</li>
     *                       <li>DisplayName</li>
     *                     </ul>
     * @param value the expected value of the given property
     */
    public void CheckActiveMessageProperty(String name, String value);

    /**
     * This command can be used to populate the user account before the main
     * application is started. It SHOULD not be used after the main application
     * has been started.
     *
     * @param url the syn url
     * @param user the user name
     * @param pwd the password
     */
    public void SaveAccount(String url, String user, String pwd);

    /**
     * This command deletes (resets) all the file in a given directory.
     *
     * @param dir the directory to be reset. Possible values are:
     *            <ul>
     *              <li>Pictures</li>
     *            </ul>
     */
    public void ResetDir(String dir);

    /**
     * This command copies one file to the specified directory.
     *
     * @param file the source file name. Possible values are:
     *            <ul>
     *              <li>Photo0.jpg</li>
     *            <ul>
     *
     * @param dir the directory to be reset. Possible values are:
     *            <ul>
     *              <li>Pictures</li>
     *            </ul>
     */
    public void CopyFileToDir(String file, String dir);
}
    

